/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Coral Binary Tree */



#ifndef __CORAL_BINTREE_H
# define __CORAL_BINTREE_H

#include <coral/def.h>



/* node directions */
#define BINTREE_LEFT   0
#define BINTREE_RIGHT  1


/* flags for _BINTREEITERPATH */
#define BINTREE_FLAG_ALLOW_REALLOC 0x0002 /* reallocation for iteration path buffer is allowed */
#define BINTREE_FLAG_ALLOCATED     0x0004 /* set if the buffer was allocated ( using memAlloc ) */
#define BINTREE_FLAG_BALANCED      0x0010 /* avl balancing */
#define BINTREE_FLAG_REDBLACK      0x0020 /* red-black balancing ( currently unsupported ) */


/* default iteration path buffer */
#define BEGIN_BINTREE_ITERPATH(iterpath,flag) ( (iterpath).path=NULL, (iterpath).last=-1, (iterpath).flag=flag|BINTREE_FLAG_ALLOW_REALLOC )
#define END_BINTREE_ITERPATH(iterpath) memFree((iterpath).path)


/* bintreeIterate, error codes */
/* ERROR_BUFFER, iteration was failed by stack error */
#define BINTREE_INVALID_KEY 3 /* the given key is not found and invalid */


/* bintreeCreate, flags */
/* FLAG_CLEAR, to clear created node */



/* bintreeCreate, error codes
 ERROR_MEM, node allocation was failed
 ERROR_BUFFER,
 SUCCESS_EXIST */



/* bintreeRemove, error codes
 BINTREE_INVALID_KEY
 ERROR_BUFFER */



/* bintreeAdd
 ERROR_BUFFER
 BINTREE_KEY_PRESENT */



/* bintreeCut
 BINTREE_INVALID_KEY
 ERROR_BUFFER */



/* bintreeUnlink
 ERROR_BUFFER */



/* bintreeTraverse, traversing methods */
#define BINTREE_PREORDER  0 /* preorder traversing */
#define BINTREE_INORDER   1 /* inorder traversing */
#define BINTREE_POSTORDER 2 /* postorder traversing */
#define BINTREE_SORT  BINTREE_INORDER   /* traversing for sorted order */
#define BINTREE_CLEAR BINTREE_POSTORDER /* traversing for clearing */



/* bintreeTraverse, error codes
0 : no problem
positive value : traversing function returns positive value
ERROR_BUFFER : stack error was occurred */

/* bintreeSearchStr */
/*
 FLAG_STRING_POINTER
 FLAG_CASE_INSENSITIVE
 FLAG_PATH_NAME */



/* recursive preorder traversing */

#define BEGIN_BINTREE_PREORDER(name,nodeclass,nodename) \
void _API name( nodeclass nodename ) \
{ \
  if( nodename ) { \

#define END_BINTREE_PREORDER(name,nodename) \
    name( nodename->left ); \
    name( nodename->right ); \
  } \
}

#define BEGIN_BINTREE_PREORDER_EXT(name,nodeclass,nodename,dataclass,dataname) \
void _API name( nodeclass nodename, dataclass dataname ) \
{ \
  if( nodename ) { \

#define END_BINTREE_PREORDER_EXT(name,nodename,dataname) \
    name( nodename->left, dataname ); \
    name( nodename->right, dataname ); \
  } \
}



/* recursive inorder traversing */

#define BEGIN_BINTREE_INORDER(name,nodeclass,nodename) \
void _API name( nodeclass nodename ) \
{ \
  if( nodename ) { \
    name( nodename->left ); \

#define END_BINTREE_INORDER(name,nodename) \
    name( nodename->right ); \
  } \
}

#define BEGIN_BINTREE_INORDER_EXT(name,nodeclass,nodename,dataclass,dataname) \
void _API name( nodeclass nodename, dataclass dataname ) \
{ \
  if( nodename ) { \
    name( nodename->left, dataname ); \

#define END_BINTREE_INORDER_EXT(name,nodename,dataname) \
    name( nodename->right, dataname ); \
  } \
}



/* recursive postorder traversing */

#define BEGIN_BINTREE_POSTORDER(name,nodeclass,nodename) \
void _API name( nodeclass nodename ) \
{ \
  if( nodename ) { \
    name( nodename->left ); \
    name( nodename->right );

#define END_BINTREE_POSTORDER(name,nodename) \
  } \
}

#define BEGIN_BINTREE_POSTORDER_EXT(name,nodeclass,nodename,dataclass,dataname) \
void _API name( nodeclass nodename, dataclass dataname ) \
{ \
  if( nodename ) { \
    name( nodename->left, dataname ); \
    name( nodename->right, dataname );

#define END_BINTREE_POSTORDER_EXT(name,nodename,dataname) \
  } \
}



/* open binary tree base class */

typedef DECLARE_BALANCED_BINTREE( _BINTREE )
DECLARE_END BINTREE;



/* binary tree iteration path */

struct _BINTREEITERITEM {
  int direction; /* next iteration direction */
  union {
    struct _BINTREE __F* ptr;       /* the pointer of a node */
    struct _BINTREE __F*__F* child; /* child reference of a node */
  } node; /* node on the path */
};


struct _BINTREEITERPATH {
  /* externally initialized */
  struct _BINTREEITERITEM __F* path; /* iteration path item buffer */
  int last;      /* the last index of path buffer, -1 if 'path' is NULL */
  unsigned flag; /* flag, high byte can be used to specify the buffer size to be expanded when the buffer lack exception is occurred ( default==2 )  */

  /* internally used */
  int stored; /* the stored path item index */
  int pivot;  /* the pivot node index */
};



/* binary tree traversing stack */

struct _BINTREESTACK {
  int last; /* the stored path item index */
  struct _BINTREE __F* stack[1]; /* stack */
};


/*??? temp code : old function which do some different things sholud not be called */
#ifndef __INTERNAL__
# define bintreeLink bintreeLink_
# define bintreeUnlink bintreeUnlink_
#endif



__CBEGIN__



/* memAlloc based binary tree node allocation and deallocation */
_API( int, bintreeCreate, ( struct _BINTREE __F*__F* newnode, struct _BINTREE __F*__F* root, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), struct _BINTREEITERPATH __F* iterpath, long size, unsigned flag ) );
_API( int, bintreeRemove, ( struct _BINTREE __F*__F* root , void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), struct _BINTREEITERPATH __F* iterpath ) );
_API( int, bintreeCreateExt, ( struct _BINTREE __F*__F* newnode, struct _BINTREE __F*__F* root, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), long size, unsigned flag ) );
_API( int, bintreeRemoveExt, ( struct _BINTREE __F*__F* root , void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), unsigned flag ) );
_API( void, bintreeClear, ( struct _BINTREE __F*__F* root ) );

/* add & cut operations */

_API( int, bintreeLink, ( struct _BINTREE __F*__F* root, struct _BINTREE __F* node, struct _BINTREEITERPATH __F* iterpath ) );
_API( int, bintreeUnlink, ( struct _BINTREE __F*__F* root, struct _BINTREEITERPATH __F* iterpath ) );

_API( int, bintreeAdd, ( struct _BINTREE __F*__F* root, struct _BINTREE __F* node, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), struct _BINTREEITERPATH __F* iterpath ) );
_API( int, bintreeCut, ( struct _BINTREE __F*__F* cut, struct _BINTREE __F*__F* root, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), struct _BINTREEITERPATH __F* iterpath ) );
_API( int, bintreeAddExt, ( struct _BINTREE __F*__F* root, struct _BINTREE __F* node, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), unsigned flag ) );
_API( int, bintreeCutExt, ( struct _BINTREE __F*__F* cut, struct _BINTREE __F*__F* root, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), unsigned flag ) );
_API( void, bintreeLinkAVL, ( struct _BINTREE __F*__F* root, struct _BINTREE __F* node, struct _BINTREEITERPATH __F* iterpath ) );
_API( int, bintreeUnlinkAVL, ( struct _BINTREE __F*__F* root, struct _BINTREEITERPATH __F* iterpath ) );

/* iterators */
_API( long, bintreeCount, ( struct _BINTREE __F* root ) );
_API( struct _BINTREE __F*, bintreeSearch, ( struct _BINTREE __F* root, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ) ) );
_API( int, bintreeIterate, ( struct _BINTREE __F*__F* foundnode, struct _BINTREE __F* root, void __F* key, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), struct _BINTREEITERPATH __F* iterpath ) );
_API( int, bintreeTraverse, ( struct _BINTREE __F* root, int ( __API* traverse )( struct _BINTREE __F* node, void __F* param ), void __F* param, unsigned method ) );
_API( struct _BINTREE __F*, bintreeFindInt1, ( struct _BINTREE __F* root, _integer1 value, int offset ) );
_API( struct _BINTREE __F*, bintreeFindInt2, ( struct _BINTREE __F* root, _integer2 value, int offset ) );
_API( struct _BINTREE __F*, bintreeFindInt4, ( struct _BINTREE __F* root, _integer4 value, int offset ) );
_API( struct _BINTREE __F*, bintreeFindInt, ( struct _BINTREE __F* root, int value, int offset ) );
_API( struct _BINTREE __F*, bintreeFindLong, ( struct _BINTREE __F* root, long value, int offset ) );
_API( struct _BINTREE __F*, bintreeFindStr, ( struct _BINTREE __F* root, const char __F* s, int offset, unsigned flag ) );


_API( long, bintreeGetHeight, ( struct _BINTREE __F* child ) );
_API( long, bintreeCheckBalance, ( struct _BINTREE __F* root, void (__API* report)( struct _BINTREE __F* invalid, void __F* data ), void __F* data ) );
_API( _bool, bintreeBalance, ( struct _BINTREE __F*__F* root, int (__API* compare)( struct _BINTREE __F* node, void __F* key ), void __F* (__API* getkey)( struct _BINTREE __F* node ) ) ); //????



/* resolve bintreeFindInt, bintreeFindLong, bintreeMaxInt, and bintreeMaxLong */

#if __ARCH__<=32

# define bintreeFindLong(root,value,offset) bintreeFindInt4(root,(_integer4)(value),offset)
# define bintreeMaxLong(root,value,offset) bintreeMaxInt4(root,(_integer4 __F*)(value),offset)

# if __ARCH__<=16
#  define bintreeFindInt(root,value,offset) bintreeFindInt2(root,(_integer2)(value),offset)
#  define bintreeMaxInt(root,value,offset) bintreeMaxInt2(root,(_integer2 __F*)(value),offset)
# else /* __ARCH__>16 */
#  define bintreeFindInt bintreeFindLong
#  define bintreeMaxInt bintreeMaxLong
# endif /* __ARCH__<=16 */

#else /* __ARCH__>32 */

#if __ARCH__==64 /* for lp64 model */
# define bintreeFindInt(root,value,offset) bintreeFindInt4(root,(_integer4)(value),offset)
# define bintreeMaxInt(root,value,offset) bintreeMaxInt4(root,(_integer4 __F*)(value),offset)
#else
... /* unknown architecture */
#endif

#endif /* __ARCH__<=32 */



#define bintreeFindChar(root,value,offset) bintreeFindInt1(root,(_integer1)(value),offset)
#define bintreeFindShort(root,value,offset) bintreeFindInt2(root,(_integer2)(value),offset)
#define bintreeFindPtr(root,ptr,offset) bintreeFindLong(root,(long)ptr,offset)
#define bintreeMaxChar(root,value,offset) bintreeMaxInt1(root,(_integer1 __F*)(value),offset)
#define bintreeMaxShort(root,value,offset) bintreeMaxInt2(root,(_integer2 __F*)(value),offset)



__CEND__



#endif /* __CORAL_BINTREE_H */
