/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Inline Macros */



#ifndef __CORAL_INLINE_H
#	define __CORAL_INLINE_H



#include <coral/def.h>



/* Integer value combinations */

#define OP_HIGH4(val)    ((_unsigned2)((_unsigned4)(val)>>16))
#define OP_LOW4(val)     ((_unsigned2)(val))
#define OP_COMBINE4(a,b) ((_unsigned4)(((_unsigned4)(_unsigned2)(a)<<16)|((_unsigned2)(b))))
#define OP_HIGH2(val)    ((_unsigned1)((_unsigned2)(val)>>8))
#define OP_LOW2(val)     ((_unsigned1)(val))
#define OP_COMBINE2(a,b) ((_unsigned2)(((_unsigned2)(_unsigned1)(a)<<8)|((_unsigned1)(b))))
#define OP_HIGH1(val)    ((_unsigned1)(val)>>4)  /* high nibble */
#define OP_LOW1(val)     ((_unsigned1)(val)&0xf) /* low nibble */
#define OP_COMBINE1(a,b) ((_unsigned1)(((_unsigned1)(a)<<4)|(b))) /* nibble combination */



/* Integer value limits */

#define OP_MAXINT(type)  ((type)OP_MAXBITINT_EXT((type)1,(int)sizeof(type)*8))
#define OP_MININT(type)  ((type)OP_MINBITINT_EXT((type)1,(int)sizeof(type)*8))
#define OP_MAXUINT(type) ((type)-1)



/* Alignement */

#define OP_BYTE2(val)  (((val)+1)>>1)
#define OP_ALIGN2(val) (OP_BYTE2(val)<<1)
#define OP_BYTE4(val)  (((val)+3)>>2)
#define OP_ALIGN4(val) (OP_BYTE4(val)<<2)
#define OP_BYTE8(val)  (((val)+7)>>3)
#define OP_ALIGN8(val) (OP_BYTE8(val)<<3)

/* byte-1 will be constant if byte is constant and division will
 be optimized by the compiler if byte can be represented by 2^n,
 n is integer */
#define OP_BYTE(val,byte) (((val)+((byte)-1))/(byte))
#define OP_ALIGN(val,byte) (OP_BYTE(val,byte)*(byte))

#define OP_BYTE_TRUNC(val,byte) ((val)/(byte))
#define OP_ALIGN_TRUNC(val,byte) (OP_BYTE_ALIGN(val,byte)*(byte))

#define OP_BYTE_PAD(val,byte) (((val)+(byte))/(byte))
#define OP_ALIGN_PAD(val,byte) (OP_BYTE_PAD(val,byte)*(byte))


/* Bit operators */

#define OP_ENABLE(a,flag)  (a|=(flag))
#define OP_DISABLE(a,flag) (a&=~(flag)) /* if flag is constant, only & is needed in runtime */
#define OP_TOGGLE(a,flag)  (a^=(flag))
#define OP_ENABLED(a,flag) ((a)|(flag)) //??? naming
#define OP_DISABLED(a,flag) ((a)&~(flag)) ///???naming
#define OP_TOGGLED(a,flag) ((a)^(flag)) ///??? naming
#define OP_ALLFLAG(a,flag) (((a)&(flag))==(flag))
#define OP_ODDINT(val) ((val)&0x1) /* ((val)%2) */
#define OP_EVENINT(val) (!OP_ODDINT(val))

/* following macros are sometimes needed for a 16 bit application, which
 cannot define 4 byte bit fields. Results are unsigned long and valid
 cast must be needed. */
#define OP_BIT_SINGLE(v) ((v)&0x1UL)
#define OP_BIT_MULTIPLE(v,nbit) ((v)&((1UL<<(nbit))-1))
#define OP_BIT_SINGLE_BASED(v,basebit) OP_BIT_SINGLE((v)>>(basebit))
#define OP_BIT_MULTIPLE_BASED(v,basebit,nbit) OP_BIT_MULTIPLE((v)>>(basebit),nbit)



/* Other operators */

#define OP_SWAPINT(a,b)   (a^=b,b^=a,a^=b) /* just enjoy, not so fast */
#define OP_MAX(a,b) (((a)>(b))?(a):(b))
#define OP_MIN(a,b) (((a)<(b))?(a):(b))
#define OP_OFFSET(c,m) ((int)&((c*)0)->m)
#define OP_LIST_OFFSET(c,m,l) (OP_OFFSET(c,m)-OP_OFFSET(c,l))
#define OP_RESERVE(object) (object=object)
#define OP_INITIALIZE(count) ((count)++==0)
#define OP_FINALIZE(count) ((count)>=1?--(count)==0:0)
#define OP_ELEM_COUNT(array) (sizeof(array)/sizeof(array[0]))
#define OP_ARRAY_PARAM(array) array, OP_ELEM_COUNT(array) /* array pointer and element count combination */
#define OP_ARRAY_PARAM_EXT(array) OP_ARRAY_PARAM(array), sizeof(array[0])

#if !defined(__STDC__) && defined(__UNIXC__)
#	include <coral/mem.h>
#	define OP_COPY(a,b) mem_short_copy(&a,&b,sizeof(a))
#else
#	define OP_COPY(a,b) a = b
#endif



/* Debug */

#ifdef __DEBUG__
#	include <coral/compat/stdio.h>
#	define DEBUG_EMIT(code) { code } /* ',' must not be used outside of parentheses */
#	define DEBUG_PRINTF(s) printf s;
#	define DEBUG_PUTS(s)  puts s;
#	define DEBUG_BEGIN_ERRCODE(code) { int __errcode=code; switch( __errcode ) {
#	define DEBUG_ERRCODE(code) case code: printf( #code ); break
#	define DEBUG_END_ERRCODE default: printf( "%d : unknown error code", __errcode ); } }
#else /* !__DEBUG__ */
#	define DEBUG_EMIT(code)
#	define DEBUG_PRINTF(s)
#	define DEBUG_PUTS(s)
#	define DEBUG_BEGIN_ERROR_CODE(code)
#	define DEBUG_EVAL_ERROR_CODE(code)
#	define DEBUG_END_ERROR_CODE
#endif /* __DEBUG__ */



#endif /* __CORAL_INLINE_H */

