/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Coral Binary Tree Implementation

 1998.07.24  Started
 1998.08.21
 1999.03.12  Reviewed
 2001.03.24  AVL rotation bug fixed

 Definitions :
 pivot node : the root node or the node which has non-zero balancing factor
 agent node : the node takes the place of the deleted node */



#undef __DEBUG__
#include <coralsrc/memory.h>
#include <coralsrc/list.h>
#include <coralsrc/bintree.h>
#include <coralsrc/io.h>
#include <coralsrc/string.h>
#include <coral/inline.h>
__MODULEEND__



#ifdef __TEMPLATE__



/*------------------------------------------------------------------------*

	 Templates

 *------------------------------------------------------------------------*/



#ifdef F_bintreeFindInt
/* bintreeFindInt
 Template parameters
 F_bintreeFindInt : the function name
 F_bintree_find_int : the traversing function name
 T_bintree_find_int : the parameter class for the traversing function
 T_value : the value type */


struct T_bintree_find_int {
	T_value              value;  /* the value to be compared */
	int                  offset; /* offset of the member */
	struct _BINTREE __F* node;   /* the found node */
};

static int __API F_bintree_find_int
P_B(( node, bfi ))
P_1  struct _BINTREE __F* node
P_N  struct T_bintree_find_int __F* bfi
P_E
{
	if( *(T_value __F*)( (_integer1 __F*)node + bfi->offset )==bfi->value ) {
		bfi->node = node;
		return 1; /* positive value to terminate traversing */
	}
	return 0;
}

struct _BINTREE __F* __API
F_bintreeFindInt
P_B(( root, value, offset ))
P_1  struct _BINTREE __F* root
P_N  T_value value
P_N  int offset
P_E
{
	struct T_bintree_find_int bfi;
	bfi.value = value;
	bfi.offset = offset;

	return
		bintreeTraverse(
			root,
			(int (__APICAST*)(struct _BINTREE __F*,void __F*))F_bintree_find_int,
			&bfi,
			BINTREE_SORT )>0 ?
		bfi.node : NULL;
}

#undef F_bintreeFindInt
#undef F_bintree_find_int
#undef T_bintree_find_int
#undef T_value
#endif /* F_bintreeFindInt */



#ifdef F_bintreeMaxInt
/* bintreeMaxInt
#define F_bintreeMaxInt bintreeMaxLong
#define F_bintree_max_int bintree_max_long
#define T_bintree_max_int _bintree_max_long
#define T_value long */



#endif /* F_bintreeMaxInt */



#undef __TEMPLATE__



#else /* !__TEMPLATE__ */



/*------------------------------------------------------------------------*

	 Internal Method

 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "brallocp.c" )
/* bintree_realloc_path */

_bool
bintree_realloc_path
P_B(( iterpath, pathcache, lastcache ))
P_1  struct _BINTREEITERPATH __F* iterpath
P_N  struct _BINTREEITERITEM __F*__F* pathcache
P_N  int __F* lastcache
P_E
{
	unsigned flag = iterpath->flag;

	/* reallocation of buffer was allowed */
	if( flag&BINTREE_FLAG_ALLOW_REALLOC ) {
		struct _BINTREEITERITEM __F* path = iterpath->path;
		int storedcount = iterpath->last+1;

		/* get the new last index
		 : the default size for the buffer expansion is 2 and the default
			buffer size is 32 ( 1 was already added in 'storedcount' ) */
		register int
		 inc = (int)OP_HIGH2(flag),
		 newlast = storedcount + ( path? inc==0? 1 : inc-1 : 31 );

		/* reallocate buffer */
		struct _BINTREEITERITEM __F* newpath = (struct _BINTREEITERITEM __F*)
			memAlloc( (newlast+1)*sizeof(struct _BINTREEITERITEM) );

		if( newpath ) {
			/* copy contents in the old buffer to the new buffer */
			if( storedcount>0 )
				mem_short_copy( newpath, path,
					storedcount*sizeof(struct _BINTREEITERITEM) );

			/* the old buffer was allocated */
			if( flag&BINTREE_FLAG_ALLOCATED ) {
				memFree( path );
			}
			else {
				/* the buffer was allocated */
				OP_ENABLE( iterpath->flag, BINTREE_FLAG_ALLOCATED );
			}

			/* update iteration path informations */
			iterpath->last = *lastcache = newlast;
			iterpath->path = newpath;
			*pathcache = &newpath[storedcount];

			return 1;
		} /* end if( newpath ) */

	} /* end if( flag&BINTREE_FLAG_ALLOW_REALLOC ) */

	return 0;

}
__MODULEEND__



__MODULEBEGIN__( "bavlrot.c" )
/* bintree_avl_rotate */


_bool bintree_avl_rotate
P_B(( root, iterpath, del ))
P_1  struct _BINTREE __F*__F* root
P_N  struct _BINTREEITERPATH __F* iterpath
P_N  _bool del
P_E
{
	struct _BINTREE __F* subtreenode, __F* iter;
	int pivotbf;

	/* non-zero to continue balancing ( for deletion ) */
	int result = 1;

	/* cache */
	int pivot=iterpath->pivot;
	struct _BINTREEITERITEM __F* pivotpath = &iterpath->path[pivot];

	/* pivot node */
	struct _BINTREE __F* pivotnode = pivotpath->node.ptr;

	/* subtreebf is initially the subtree direction */
	int subtreebf = pivotpath->direction;

	if( del ) {
		/* for deletion */
		subtreebf= !subtreebf;
	}
	else {
		/* for addition :
			update balance factors from pivot node to the parent node of the
		 inserted node */
		int stored = iterpath->stored, i = pivot;
		struct _BINTREEITERITEM __F* path = pivotpath;
		while( ++i<=stored ) {
			++path;
			path->node.ptr->bf = path->direction==BINTREE_LEFT ? 1 : -1;
		}
	} /* end if( del ) */


	/* get the subtree node */
	iter = subtreenode = ((struct _BINTREE __F*__F*)pivotnode)[subtreebf];

	/* direction index to balance factor */
	subtreebf = subtreebf==BINTREE_LEFT ? 1 : -1;

	pivotbf = pivotnode->bf;

	/* tree is still balanced */
	if( !pivotbf || !(pivotbf+subtreebf ) ) {
		DEBUG_PUTS(( "AVL rotation is not needed." ))
		pivotnode->bf += subtreebf;
		return pivotbf;
	}
	else {
		struct _BINTREE __F* subtreechild;

		/* left unbalanced, bf is 1 for addition and 0 for deletion so >=0 is used */
		if( subtreebf>=0 ) {
			if( subtreenode->bf>=0 ) {
				DEBUG_PUTS(( "LL rotation" ))

				/* LL rotation

					 pivotnode(+2)            subtreenode(0)
					/            \              /    \
				  subtreenode(+1)  ...    ->    |     pivotnode(0)
				  /     \                       *      /  \
				 |       R                     node   R   ...
				 *
				node                                                    */

				pivotnode->left = subtreenode->right;
				subtreenode->right = pivotnode;

				if( --subtreenode->bf==-1 ) {
					/* for deletion, subtreenode->bf==0 before -- is applied */
					result = 0;
					goto l_parent;
				}

				pivotnode->bf--;
			}
			else {
				DEBUG_PUTS(( "LR rotation" ))
				/* LR rotation

					pivotnode(+1)                  *node(0)
					/                                /   \
				  subtreenode(-1)     -> subtreenode(0)  pivotnode(0)
					\
					 *node(==subtreechild)

				  generally,

					pivotnode(+1)              subtreechild(0)
					/         \                  /        \
				subtreenode(-1)  ...   -> subtreenode(0)   pivotnode(-1)
				   /    \                  /    \          /  \
				 ...   subtreechild(+1)  ...     |         R   ...
						/  \                     * node
					   |    R
					   *
					   node
				*/

				/* balance tree */
				subtreechild = subtreenode->right;
				subtreenode->right = subtreechild->left;
				pivotnode->left = subtreechild->right;
				subtreechild->left = subtreenode;
				subtreechild->right = pivotnode;

				/* set bf */

				switch( subtreechild->bf ) {
					case 0 :
					pivotnode->bf = 0;
					subtreenode->bf = 0;
					break;

					case 1 :
					pivotnode->bf = -1;
					subtreenode->bf = 0;
					break;

					default : /* -1 */
					pivotnode->bf = 0;
					subtreenode->bf = 1;
				}

				/* post process */
				subtreechild->bf = 0;
				subtreenode = subtreechild;
			}
		}
		/* right unbalanced, bf is -1 for addtion and 0 for deletion */
		else {
			if( subtreenode->bf<=0 ) {
				DEBUG_PUTS(( "RR rotation" ))

				/* RR rotation

					pivotnode(-2)                       subtreenode(0)
					 /       \                          /         \
				   ...      subtreenode(-1)   ->   pivotnode(0)    |
							 /    \                 /   \          *
							L      |               ...   L        node
								   *
								   node
				*/
				pivotnode->right = subtreenode->left;
				subtreenode->left = pivotnode;

				subtreenode->bf++;
				if( subtreenode->bf==1 ) {
					/* for deletion */
					result = 0;
					goto l_parent;
				}

				pivotnode->bf++;

			}
			else {
				DEBUG_PUTS(( "RL rotation" ))
				/* RL rotation

					 pivotnode(-1)                      subtreechlid(0)
					  /         \                         /          \
					...         subtreenode(+1)   ->  pivotnode(+1)  subtreenode(0)
				                   /       \           /    \       /    \
						  subtreechild(-1)  ...      ...     L     |     ...
							/   \                                  *
						   L     |                                node
								 *
								node
				*/

				/* balance tree */
				subtreechild = subtreenode->left;
				subtreenode->left = subtreechild->right;
				pivotnode->right = subtreechild->left;
				subtreechild->left = pivotnode;
				subtreechild->right = subtreenode;

				/* set bf */
				switch( subtreechild->bf ) {
					case 0 :
					pivotnode->bf = 0;
					subtreenode->bf = 0;
					break;

					case 1:
					pivotnode->bf = 0;
					subtreenode->bf = -1;
					break;

					default: /* -1 */
					pivotnode->bf = 1;
					subtreenode->bf = 0;
				}

				/* post process */
				subtreechild->bf = 0;
				subtreenode = subtreechild;
			} /* end if right rotation type ) */
		} /* end if ( avl rotation type ) */

		/* manage parent */
		l_parent:
		if( pivot-1>=0 ) {
			pivotpath--;
			pivotpath->node.child[pivotpath->direction] = subtreenode;
		}
		else {
			*root = subtreenode;
		}
		
	} /* end if( tree still balanced ) */

	return result;
}
__MODULEEND__



/*-------------------------------------------------------------------------*

	memAlloc based binary tree node allocation and deallocation

 *-------------------------------------------------------------------------*/



__MODULEBEGIN__( "bcreate.c" )
/* bintreeCreate

 flag&BINTREE_FLAG_INTERNAL_CREATION, size will be node size
 flag&BINTREE_FLAG_EXTERNAL_CREATION, size will be externally created
	 node pointer using memAlloc
	 The reason why the core code for bintreeAdd is contained here is
 to share the code for the case when the duplicated key is found and
 no allocation is needed for that case. */

int __API bintreeCreate
P_B(( newnode, root, key, compare, iterpath, size, flag ))
P_1  struct _BINTREE __F*__F* newnode
P_N  struct _BINTREE __F*__F* root
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  struct _BINTREEITERPATH __F* iterpath
P_N  long size
P_N  unsigned flag
P_E
{
	struct _BINTREE __F* node;

	switch( bintreeIterate( &node, *root, key, compare, iterpath  ) ) {
		case 0:

		/* specified key was present, just return it */
		*newnode = node;
		return SUCCESS_EXIST;

		case FAIL_BUFFER:
		return FAIL_BUFFER;
	}

	/* evaluate option flag and create a leaf */
	if( flag&_BINTREE_FLAG_EXTERNAL_CREATION ) {
		node = (struct _BINTREE __F*)size;
	}
	else {
		node = (struct _BINTREE __F*)
			( flag&FLAG_CLEAR? memCAlloc : memAlloc )( size );
		if( !node ) return FAIL_MEMORY;
	} /* end of( flag&_BINTREE_FLAG_EXTERNAL_CREATION ) */

	*newnode = node;

	bintreeLinkAVL( root, node, iterpath );

	/* success */
	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "bremove.c" )
/* bintreeRemove

	This method prototype does not use external destructor defined by a user
 and a binary tree node must be destructed only by memFree. It's better
 to use freely defined destructor which calls bintreeCut and implements
 node destruction than to supply destruction procedure to bintreeRemove
 - anyway a destruction method need to be defined for two cases. */

int __API bintreeRemove
P_B(( root, key, compare, iterpath ))
P_1  struct _BINTREE __F*__F* root
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  struct _BINTREEITERPATH __F* iterpath
P_E
{
	struct _BINTREE __F* cut;

	/* cut the node which has the given key */
	int result = bintreeCut( &cut, root, key, compare, iterpath );

	/* bintreeCut was successful */
	if( result==0 )
		memFree( cut );

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "bcreatex.c" )
/* bintreeCreateExt

	Iteration path is internally managed. Iteration path flags and bintreeCreate
 flags can be combinated ( because there is no collision ) and high byte can
 be increment for the path buffer expansion. other 'Ext' functions follows
 this rules. */

int __API bintreeCreateExt
P_B(( newnode, root, key, compare, size, flag ))
P_1  struct _BINTREE __F*__F* newnode
P_N  struct _BINTREE __F*__F* root
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  long size
P_N  unsigned flag
P_E
{
	int result;
	struct _BINTREEITERPATH iterpath;

	BEGIN_BINTREE_ITERPATH( iterpath, flag );

	/* low byte is masked not to send high byte because bintreeCreate uses
	 internally defined flags in high byte : see comment in bintreeCreate */
	result = bintreeCreate
	 ( newnode, root, key, compare, &iterpath, size, flag&0xff );

	END_BINTREE_ITERPATH( iterpath );

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "brmex.c" )
/* bintreeRemoveExt */

int __API bintreeRemoveExt
P_B(( root, key, compare, flag ))
P_1  struct _BINTREE __F*__F* root
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  unsigned flag
P_E
{
	int result;
	struct _BINTREEITERPATH iterpath;

	BEGIN_BINTREE_ITERPATH( iterpath, flag );

	result = bintreeRemove( root, key, compare, &iterpath );

	END_BINTREE_ITERPATH(iterpath);

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "bclear.c" )
/* bintreeClear

 The reason of this method prototype, see bintreeRemove */

#ifdef __TURBOC__
# pragma argsused
#endif

static int __API bintree_clear_proc
P_B(( root, data ))
P_1  struct _BINTREE __F* root
P_N  void __F* data
P_E
{

# ifdef __MACOS__
#  pragma unused(data)
# endif

	memFree( root );
	return 0;
}

void __API bintreeClear
P_B(( root ))
P_1  struct _BINTREE __F*__F* root
P_E
{
	bintreeTraverse( *root, bintree_clear_proc, NULL, BINTREE_CLEAR );
	*root = NULL;
}
__MODULEEND__



/*-------------------------------------------------------------------------*

	Add & cut operations

 *-------------------------------------------------------------------------*/



__MODULEBEGIN__( "badd.c" )
/* bintreeAdd */

int __API bintreeAdd
P_B(( root, node, key, compare, iterpath, flag ))
P_1  struct _BINTREE __F*__F* root
P_N  struct _BINTREE __F* node
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  struct _BINTREEITERPATH __F* iterpath
P_E
{
	struct _BINTREE __F* newnode;

	/* add externally created node */
	return bintreeCreate
	 ( &newnode, root, key, compare, iterpath,
		(long)node, _BINTREE_FLAG_EXTERNAL_CREATION );
}
__MODULEEND__



__MODULEBEGIN__( "bcut.c" )
/* bintreeCut */

int __API bintreeCut
P_B(( cut, root, key, compare, iterpath ))
P_1  struct _BINTREE __F*__F* cut
P_N  struct _BINTREE __F*__F* root
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  struct _BINTREEITERPATH __F* iterpath
P_E
{
	/* search appropriate node */
	int result = bintreeIterate( cut, *root, key, compare, iterpath );

	/* the node which has the given key was found */
	if( result==0 )
		bintreeUnlink( root, iterpath );

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "baddex.c" )
/* bintreeAddExt */

int __API bintreeAddExt
P_B(( root, node, key, compare, flag ))
P_1  struct _BINTREE __F*__F* root
P_N  struct _BINTREE __F* node
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  unsigned flag
P_E
{
	int result;
	struct _BINTREEITERPATH iterpath;

	BEGIN_BINTREE_ITERPATH( iterpath, flag );

	result = bintreeAdd( root, node, key, compare, &iterpath );

	END_BINTREE_ITERPATH( iterpath );

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "bcutex.c" )
/* bintreeCutExt */

int __API bintreeCutExt
P_B(( cut, root, key, compare, flag ))
P_1  struct _BINTREE __F*__F* cut
P_N  struct _BINTREE __F*__F* root
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  unsigned flag
P_E
{
	int result;
	struct _BINTREEITERPATH iterpath;

	BEGIN_BINTREE_ITERPATH( iterpath, flag );

	result = bintreeCut( cut, root, key, compare, &iterpath );

	END_BINTREE_ITERPATH( iterpath );

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "blink.c" )
/* bintreeLink */
int __API
bintreeLink
P_B(( root, node, iterpath ))
P_1  struct _BINTREE __F*__F* root
P_N  struct _BINTREE __F* node
P_N  struct _BINTREEITERPATH __F* iterpath
P_E
{
	/* initialize leafs */
	node->left = node->right = (struct _BINTREE __F*)NULL;

	/* add the given node */
	if( *root ) {
		/* link the given node to the child of the last-found path item */
		register struct _BINTREEITERITEM __F* path = &iterpath->path[iterpath->stored];
		path->node.child[path->direction] = node;
		return 1;
	}

	*root = node;
	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "blinkavl.c" )
/* bintreeLinkAVL

 separated from bintreeCreate on 1999.12.25, see comments in bintreeUnlink */

void __API
bintreeLinkAVL
P_B(( root, node, iterpath ))
P_1  struct _BINTREE __F*__F* root
P_N  struct _BINTREE __F* node
P_N  struct _BINTREEITERPATH __F* iterpath
P_E
{
	node->bf = 0;

	if( bintreeLink( root, node, iterpath ) )
		bintree_avl_rotate( root, iterpath, 0 );
}
__MODULEEND__



__MODULEBEGIN__( "bunlink.c" )
/* bintreeUnlink

	This method is separated from bintreeCut on 1999.03.13 to reduce
 overhead caused by duplicated calls of bintreeIterate and bintreeCut
 or its families. */

int __API bintreeUnlink
P_B(( root, iterpath ))
P_1  struct _BINTREE __F*__F* root
P_N  struct _BINTREEITERPATH __F* iterpath
P_E
{
	struct _BINTREE __F* agent;

	/* cache */
	unsigned flag = iterpath->flag;
	int stored = iterpath->stored, last=iterpath->last;
	struct _BINTREEITERITEM __F* path = &iterpath->path[stored];

	/* get the node to be unlinked */
	struct _BINTREE __F* del = path->node.ptr;

	/* find the agent node */
	if( !del->left ) {
		agent = del->right;

//    if( (flag&BINTREE_FLAG_BALANCED) && agent )
//      agent->bf = 0; /*///??? needed ???*/
	}
	else if( !del->right ) {
		agent = del->left;

//    if( (flag&BINTREE_FLAG_BALANCED) && agent )
//      agent->bf = 0;
	}
	else {
		/* left and right nodes are present, find minnode in right sub tree and
		 make it the agent. there is evidently the parent node of agent because
		 this section is activated when del->right and del->left are valid
		 del - left
				 - right - ... - pagent - agent(left of pagent) */

		struct _BINTREE __F* pagent=NULL, __F* right=del->right; /* in right sub tree */

		/* find minnode for agent, which means leftmost node in right sub tree */
		int appended=stored;
		struct _BINTREEITERITEM __F* appendedpath = path+1;
		while( 1 ) {
			/* append an item to the path buffer */
			if( appended++>=last ) {
				if( !bintree_realloc_path( iterpath, &appendedpath, &last ) )
					return FAIL_BUFFER;
				/* buffer pointer is changed, update 'path' to be used later */
				path = &iterpath->path[stored];
			}

			appendedpath->direction = 0;
			appendedpath->node.ptr = right;
			appendedpath++;

			/* get the needed nodes */
			agent = right;
			right = right->left;
			if( !right ) break;

			pagent = agent;
		} /* end while( 1 ) */

		iterpath->stored = appended;

		/* replace the last stack entry registered by bintreeIterate to the found
		 agent node */
		path->node.ptr = agent;
		path->direction = 1;

		/* it was mentioned that agent node must have the parent node but
		 following evaluation of the parent node checks pagent is not NULL
		 which really means pagent==del.
			... - del - left
					      - right - left==NULL
		 in this case, above iteration to find the agent node cannot execute
		 'pagent=agent' because there is nothing to iterate. it is known
		 that del->right==agent and agent->left=del->left is sufficient
		 for cutting operation. */

		if( pagent ) {
			/* agent node need to be removed from the left node of the parent node.
			 need not to consider agent->left subtree because agent is the node
			 which is returned by the method to terminate a bintree in left
			 direction - evidently agent->left==NULL. */
			pagent->left = agent->right;

			/* rearrange left nodes for agent */
			agent->right = del->right;

		} /* end if pagent */

		agent->left = del->left;

		if( flag&BINTREE_FLAG_BALANCED )
			agent->bf = del->bf;

	} /* end if( left and right node are present ) */

	/* check the root has the given key, if so replace the root */
	if( stored<=0 ) { /*/iterpath->stored<=0 )*/
		*root = agent;
	}
	else {
		path--;
		path->node.child[path->direction] = agent;
	}


	if( flag&BINTREE_FLAG_BALANCED ) {
		stored = iterpath->stored;
		path = &iterpath->path[stored];

		while( --stored>=0 ) {
			/* make the pivot node to the last stored node */
			iterpath->pivot = stored;

			/* perform avl rotation */
			if( !bintree_avl_rotate
					  ( stored-1>=0? &path->node.child[(--path)->direction] : root,
					    iterpath, 1 ) )
				break;
		} /* end while( --stored>=0 ) */

	} /* end if( flag&BINTREE_FLAG_BALANCED ) */

	return 0;
}
__MODULEEND__



/*------------------------------------------------------------------------*

	Iterators

 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "bcount.c" )
/* bintreeCount */

#ifdef __TURBOC__
# pragma argsused
#endif

static int __API bintree_count_proc
P_B(( node, counter ))
P_1  struct _BINTREE __F* node
P_N  void __F* counter
P_E
{
# ifdef __MACOS__
#  pragma unused(node)
# endif

	(*(long*)counter)++;
	return 0;
}

long __API bintreeCount
P_B(( root ))
P_1  struct _BINTREE __F* root
P_E
{
	long count = 0;
	bintreeTraverse( root, bintree_count_proc, &count, BINTREE_PREORDER );
	return count;
}

__MODULEEND__



__MODULEBEGIN__( "bsrch.c" )
/* bintreeSearch */

struct _BINTREE __F* __API bintreeSearch
P_B(( root, key, compare ))
P_1  struct _BINTREE __F* root
P_N  void __F* key
P_N  int (__API* compare)  P_A(( struct _BINTREE __F* node, void __F* key ))
P_E
{
	while( root ) {

		int code = compare(root,key);

		/* root == key, matching key found */
		if( code==0 )
			break;

		if( code<0 ) {
			/* root < key */
			root = root->right;
		}
		else if( code>0 ) {
			/* root > key */
			root = root->left;
		}

	} /* end while( root ) */

	return root;
}
__MODULEEND__



__MODULEBEGIN__( "biter.c" )
/* bintreeIterate

	This method is similar to bintreeSearch but iteration paths are
 stored in the given buffer.

 //???? following comments are reserved
 This method can report the parent node of the found node :
	if returned pointer is NULL, parent node has the most nearest key.
	if the root has the given key, parent node will be NULL, this case
 can be evaluated by the following two typical codes
	if( result.found.node==root ) or
	if( result.found.parent.node==NULL ) */

int __API
bintreeIterate
P_B(( foundnode, root, key, compare, iterpath ))
P_1  struct _BINTREE __F*__F* foundnode
P_N  struct _BINTREE __F* root
P_N  void __F* key
P_N  int (__API* compare) P_A(( struct _BINTREE __F* node, void __F* key ))
P_N  struct _BINTREEITERPATH __F* iterpath
P_E
{
	/* ready for the iteration path */
	struct _BINTREEITERITEM __F* path=iterpath->path;
	int pivot=-1, last=iterpath->last, stored=-1;
///???  unsigned flag = iterpath->flag;

	/* iterate nodes */
	struct _BINTREE __F* iter = root;

	*foundnode = NULL;

	while( iter ) {
		/* do comparison */
		int code = compare( iter, key );

		/* store iteration path */

		if( stored++>=last ) {
			/* the value in 'stored' is stored count now because ++ was applied */
			if( !bintree_realloc_path( iterpath, &path, &last ) )
				return FAIL_BUFFER;
		} /* end if( stored++>=last ) */

		path->direction = code<0;
		path->node.ptr = iter;
		path++;

		/* bf evaluation,
		 default pivot node is the root node so iter==root is o.k. */
		if( iter->bf || iter==root )
			pivot = stored;

		/* iter == key, matching key found */
		if( code==0 ) {
			*foundnode = iter;
			break;
		}

		if( code<0 ) {
			/* iter < key */
			iter = iter->right;
		}
		else if( code>0 ) {
			/* iter > key */
			iter = iter->left;
		}

	} /* end while( iter ) */

	iterpath->stored = stored;
	iterpath->pivot = pivot;

	/* the given key is invalid */
	return iter? 0 : BINTREE_INVALID_KEY;

}
__MODULEEND__



__MODULEBEGIN__( "btrav.c" )
/* bintreeTraverse

 Universal binary tree traversing method.
 Return 0 if successful, positive value if 'traverse' returns positive value, -1 if there is an internal error
????? */

int __API
bintreeTraverse
P_B(( root, traverse, data, method ))
P_1  struct _BINTREE __F* root
P_N  int ( __API* traverse ) P_A(( struct _BINTREE __F* node, void __F* param ))
P_N  void __F* param
P_N  unsigned method
P_E
{
	int result = 0;

	if( root ) {

		/* stack for binary tree to implement non-recursive traverse */
		DECLARE_STACK( BINTREESTACK )
			struct _BINTREE __F* node;
		DECLARE_END __F* btstack=NULL;

		while( 1 ) {

			/*----- find left terminal node -----*/
			while( root ) {
				struct BINTREESTACK __F* newstack;

				/* preorder traverse is placed here */
				if( method==BINTREE_PREORDER && (result=traverse(root,param))!=0 )
					goto l_terminate;

				/* register new stack entry to emulate recursion */
				newstack = (struct BINTREESTACK __F*)
					listPush( (struct _LIST __F*__F*)&btstack, sizeof(struct BINTREESTACK) );

				if( !newstack ) {
					result = FAIL_BUFFER;
					goto l_terminate;
				}

				newstack->node = root;

				/* left node iteration */
				root = root->left;

			} /* end while( root ) */

			/*----- emulate stack sequence for traverse -----*/
			if( btstack ) {
				root = btstack->node;

				/* special iteration loop is needed for post order traverse */
				if( method==BINTREE_POSTORDER ) {

					struct _BINTREE __F* old = NULL;

					/* if root->right is not NULL at the first time, root->right==old
					  is evaluated but that is evidently false because old is NULL.
					  this loop criterion means the node which has visited right node
					  must not be used in left node iteration again */

					while( !root->right || root->right==old ) {
					  if( (result=traverse(root,param))!=0 ) goto l_terminate;
					  listPop( (struct _LIST __F*__F*)&btstack );
					  if( !btstack ) return 0;
					  old = root;
					  root = btstack->node;
					}
				}
				else {
					/* inorder traverse is placed here */
					if( method==BINTREE_INORDER && (result=traverse(root,param))!=0 )
					  goto l_terminate;
					listPop( (struct _LIST __F*__F*)&btstack );
				} /* end if( method==BINTREE_POSTORDER ) */

				root = root->right;
			}
			else {
				return 0;
			} /* end if( btstack ) */

		} /* end while( 1 ) */

		l_terminate:
		listClear( (struct _LIST __F*__F*)&btstack );
	} /* end if( root ) */

	return result;
}

__MODULEEND__



__MODULEBEGIN__( "bfndi1.c" )
/* bintreeFindInt1 */

#define __TEMPLATE__
#define F_bintreeFindInt bintreeFindInt1
#define F_bintree_find_int bintree_find_int1
#define T_bintree_find_int _bintree_find_int1
#define T_value _integer1
#include <coralsrc/bintree.c>

__MODULEEND__



__MODULEBEGIN__( "bfndi2.c" )
/* bintreeFindInt2 */

#define __TEMPLATE__
#define F_bintreeFindInt bintreeFindInt2
#define F_bintree_find_int bintree_find_int2
#define T_bintree_find_int _bintree_find_int2
#define T_value _integer2
#include <coralsrc/bintree.c>

__MODULEEND__



__MODULEBEGIN__( "bfndi4.c" )
/* bintreeFindInt4 */

#define __TEMPLATE__
#define F_bintreeFindInt bintreeFindInt4
#define F_bintree_find_int bintree_find_int4
#define T_bintree_find_int _bintree_find_int4
#define T_value _integer4
#include <coralsrc/bintree.c>

__MODULEEND__



__MODULEBEGIN__( "bfindstr.c" )
/* bintreeFindStr */

struct _BINTREE __F* __API bintreeFindStr
P_B(( root, s, offset, ptr ))
P_1  struct _BINTREE __F* root
P_N  const char __F* s
P_N  int offset
P_N  unsigned flag
P_E
{
	struct _str_compare_class_str str;

	str.s = s;
	str.offset = offset;
	str.flag = flag;

	return (struct _BINTREE __F*) (
		bintreeTraverse(
			root,
			(int(__APICAST*)P_A((struct _BINTREE __F*,void __F*))) str_compare_class_str,
			&str,
			BINTREE_SORT )>0 ? str.object : NULL );
}

__MODULEEND__



__MODULEBEGIN__( "bgethgt.c" )
/* bintreeGetHeight */

long __API bintreeGetHeight
P_B(( child ))
P_1  struct _BINTREE __F* child
P_E
{
	if( child ) {
		DECLARE_STACK( _bintree_get_height )
			struct _BINTREE __F* node; /* a pushed node */
		DECLARE_END __F* stack = NULL;

		unsigned long current=1, height=1;

		while( 1 ) {
			/* push the current child */
			if( !listPush( (struct _LIST __F*__F*)&stack,
					   sizeof(struct _bintree_get_height) ) ) {
				/* allocation failed so return */
				listClear( (struct _LIST __F*__F*)&stack );
				return -1;
			}

			stack->node = child;

			/* get the left node */
			child = child->left;

			/* increase the current height and update maximum height */
			if( ++current>height )
				height = current;

			/* reached to the terminal node */
			while( !child ) {
				/* the right node was not visited */
				while( !stack->node ) {
					current--;
					/* left and right childs are visited so pop stack */
					listPop( (struct _LIST __F*__F*)&stack );
					if( !stack ) return height;
				}

				child = stack->node->right;
				stack->node = NULL;

			} /* end if( !child ) */

		} /* end while( 1 ) */

	} /* end if( child ) */

	/* because we received the child of a node */
	return 1;
}
__MODULEEND__



__MODULEBEGIN__( "bchkbal.c" )
/* bintreeCheckBalance

 Used only for balanced binary tree. */

struct _bintree_check_balance {
	void (__API* report) P_A(( struct _BINTREE __F* invalid, void __F* data ));
	void __F* data;
	long count;
};

static int __API bintree_check_balance
P_B(( node, data ))
P_1  struct _BINTREE __F* node
P_N  void __F* data
P_E
{
	int bf = node->bf;

	if( bf != bintreeGetHeight(node->left)-bintreeGetHeight(node->right) ||
			bf<-1 || bf>1 )  {
		/* invalid node found, so send callback */
		((struct _bintree_check_balance __F*)data)->report
		 ( node, ((struct _bintree_check_balance __F*)data)->data );

		/* increase invalid node counter */
		((struct _bintree_check_balance __F*)data)->count++;
	}

	return 0;
}

long __API bintreeCheckBalance
P_B(( root, report, data ))
P_1  struct _BINTREE __F* root
P_N  void (__API* report) P_A(( struct _BINTREE __F* invalid, void __F* data ))
P_N  void __F* data
P_E
{
	struct _bintree_check_balance bcb;
	bcb.report = report;
	bcb.data = data;
	bcb.count = 0;

	return bintreeTraverse
		( root, bintree_check_balance, &bcb, BINTREE_SORT )>=0 ?  bcb.count : -1;
}

__MODULEEND__



#ifdef XXX
__MODULEBEGIN__( "bbalance.c" )
/* bintreeBalance

	Recovery operation for bintree linkage destructed by unexpected exception
 occurred in balancing loop cannot be done. This exception can be generated
 by low on memory for the balancing stack.
	Actually following code has limitation on the 16 bit platform because
 huge pointer is not used for node list array. */

static int __API bintree_balance_proc
P_B(( node, nodeiter ))
P_1  struct _BINTREE __F* node
P_N  struct _BINTREE __F*__F* __F* nodeiter
P_E
{
	/* register current node */
	*(*nodeiter)++ = node;
	return 0;
}

_bool __API bintreeBalance
P_B(( root, iterproc
P_1 ( struct _BINTREE __F*__F* root, BINTREEITERPROC iterproc, BINTREEKEYPROC keyproc ) )
{
	struct _BINTREE __F* rootnode = *root;
	long end = bintreeCount( rootnode );
	if( end ) {
		/* create node list */
		struct _BINTREE __F*__F* nodelist =
		 (struct _BINTREE __F*__F*) memAlloc( (long)sizeof(struct _BINTREE __F*)*end );
		if( nodelist ) {
			long start;
			DECLARE_STACK( BINTREECLUSTER )
				long start, end; /* start and end index for a cluster */
			DECLARE_END;
			struct BINTREECLUSTER __F* bchead=NULL;

			/* initialize node list in sorted order */
			struct _BINTREE __F*__F* nodeiter = nodelist;
			bintreeTraverse( rootnode, (BINTREEITERPROC)bintree_balance_proc,
				&nodeiter, BINTREE_SORT );

			/* add nodes using binary subdivision of node list */
			*root = NULL;
			start = 0;
			end--;
			while( 1 ) {

				long half = (start+end)/2;

				/* implicit cast from long to int is occurred to ignore warning on
					16 bit platform */
				struct _BINTREE __F* halfnode = nodelist[(int)half];

				/* balancing must not be specified because this function is only
					needed for pure bintree and evidently following call will not
					return false because duplicated key was excluded by bintree methods */
				bintreeAdd( root, halfnode, keyproc(halfnode), iterproc, 0 );

				/* pop a cluster if requested */
				if( half==start && half==end ) {
					if( !bchead ) break;
					start = bchead->start;
					end = bchead->end;
					listPop( (struct _LIST __F*__F*)&bchead );
				}
				else {
					if( start==half )
					  start++;
					else if( end==half )
					  end--;
					else {
					  /* push right cluster */
					  if( !listPush( (struct _LIST __F*__F*)&bchead,
					         sizeof(struct BINTREECLUSTER) ) )  {
					    /* this exception cannot recover node linkage in the given bintree */
					    return 0;
					  }
					  bchead->start = half+1;
					  bchead->end = end;
					  /* reconfigure current cluster */
					  end = half-1;
					}
				}
			}

			memFree( nodelist );
			return 1;
		}
		return 0;
	}
	return 1;
}

__MODULEEND__

#endif



/*------------------------------------------------------------------------*

	Architecture Specific Routines

 *------------------------------------------------------------------------*/



#if __ARCH__==64


__MODULEBEGIN__( "bfindl.c" )
/* bintreeFindLong */

#define __TEMPLATE__
#define F_bintreeFindInt bintreeFindLong
#define F_bintree_find_int bintree_find_long
#define T_bintree_find_int _bintree_find_long
#define T_value long
#include <coralsrc/bintree.c>

__MODULEEND__



__MODULEBEGIN__( "bmaxl.c" )
/* bintreeMaxLong */

#define __TEMPLATE__
#define F_bintreeMaxInt bintreeMaxLong
#define F_bintree_max_int bintree_max_long
#define T_bintree_max_int _bintree_max_long
#define T_value long
#include <coralsrc/bintree.c>

__MODULEEND__



#endif /* __ARCH__==64 */



#endif /* __TEMPLATE__ */



/* end of bintree.c */
