/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* stdio.h compatibility and extensions */



#include <stdarg.h>
#include <malloc.h>
#include <coral/inline.h>
#include <coral/io.h>
#include <coral/compat/string.h>
#include <coral/compat/stdio.h>



/* internal buffer size for allocated printf functions */

#define _PRINTF_BUFSIZ  128
#define _PRINTF_BUFINC  64

#if _PRINTF_BUFSIZ==_PRINTF_BUFINC
#define _PRINTF_ALIGN_INIT_BUFSIZ(i) OP_ALIGN_PAD(i,_PRINTF_BUFSIZ);
#define _PRINTF_ALIGN_BUFSIZ _PRINTF_ALIGN_INITIAL_BUFSIZ
#else
#define _PRINTF_ALIGN_INIT_BUFSIZ(i) ((i)<_PRINTF_BUFSIZ?_PRINTF_BUFSIZ:_PRINTF_ALIGN_BUFSIZ(i))
#define _PRINTF_ALIGN_BUFSIZ(i) (_PRINTF_BUFSIZ+OP_ALIGN_PAD((i)-_PRINTF_BUFSIZ,_PRINTF_BUFINC))
#endif



#ifdef _MSC_VER

#define MSC_VSNPRINTF_BUG /* maybe this bug is not fixed yet by MS XXX */
#include <memory.h>



/* vsnprintf */

int
vsnprintf( char __F* buf, int len, const char __F* fmt, va_list arg )
{
	int result = _vsnprintf( buf, len-1, fmt, arg );
	if( result>=0 ) {
		buf[result] = 0;
		return result;
	}
	buf[len-1]=0;
	/* return value is not fully compatible with standard vsnprintf
	  but buffer-lack test can be the same :
	  if( return_value>=len ) buffer_lack();

	  the reason why add extra bytes is to consider reallocation occurred
	  in vbnprintf */
	return len+32;
}



/* snprintf */

int
snprintf( char __F* buf, int len, const char __F* fmt, ... )
{
	int result;
	va_list arg;
	va_start(arg,fmt);
	result = vsnprintf( buf, len, fmt, arg );
	va_end(arg);
	return result;
}



/* vasprintf */

int
vasprintf( char __F*__F* buf, const char __F* fmt, va_list arg )
{
	int bufpos = 0;
	*buf = NULL;
	return vabprintf( buf, &bufpos, fmt, arg );
}



/* asprintf */

int
asprintf( char __F*__F* buf, const char __F* fmt, ... )
{
	va_list arg;
	int result, bufpos = 0;
	*buf = NULL;
	va_start(arg,fmt);
	result = vabprintf( buf, &bufpos, fmt, arg );
	va_end(arg);
	return result;
}



#endif /* _MSC_VER */



/* vbnprintf

	Return value : total bytes written includeing previous call of this
		function so that the buffer lack condition can be detected similiar to
		vsnprintf. */

int
vbnprintf
P_B(( buf, len, bufpos, fmt, arg ))
P_1  char __F* buf
P_N  int len
P_N  int __F* bufpos
P_N  const char __F* fmt
P_N  va_list arg
P_E
{
	int result, i = *bufpos;
	buf += i;
	len -= i;

	result = vsnprintf( buf, len, fmt, arg );
	if( result<len ) *bufpos += result;
	return result+i;
}



/* bnprintf */

int
bnprintf
P_B(( buf, len, bufpos, fmt ))
P_1  char __F* buf
P_N  int len
P_N  int __F* bufpos
P_N  const char __F* fmt
P_X
{
	int result;
	va_list arg;
	va_start(arg,fmt);
	result = vbnprintf( buf, len, bufpos, fmt, arg );
	va_end(arg);
	return result;
}



/* vabprintf */

int
vabprintf
P_B(( buf, bufpos, fmt, arg ))
P_1  char __F*__F* buf
P_N  int __F* bufpos
P_N  const char __F* fmt
P_N  va_list arg
P_E
{
	int result, i, len;
	char __F* p = *buf;

	if(  p==NULL ) {
		p = (char __F*) malloc( _PRINTF_BUFSIZ );
		if( p==NULL ) return FAIL_MEMORY;
		*buf = p;
	}

	i = *bufpos;
	len = _PRINTF_ALIGN_INIT_BUFSIZ(i);

	while( 1 ) {
		result=vbnprintf( p, len, bufpos, fmt, arg );
		if( result<len ) break; /* failure or sufficient buffer*/
		len += _PRINTF_BUFINC;
		len = _PRINTF_ALIGN_BUFSIZ(len);
		*buf = p = realloc( p, len );
		if( p==NULL ) return FAIL_MEMORY;
	}

	return result;
}



/* abprintf */

int
abprintf
P_B(( buf, len, bufpos, fmt ))
P_1  char __F*__F* buf
P_N  int* bufpos
P_N  const char* fmt
P_X
{
	int result;
	va_list arg;
	va_start(arg,fmt);
	result = vabprintf( buf, bufpos, fmt, arg );
	va_end(arg);
	return result;
}



__MODULEBEGIN__( "stdlock.c" )



/* stdioLockBuffer */

_integer1 __F* __API
stdioLockBuffer
	P_B(( stdiolock, locked, size, flag ))
	P_1		struct _STDIOLOCKBUFFER __F* stdiolock
	P_N		int __F* locked
	P_N		int size
	P_N		unsigned flag
	P_E
{
	int bufpos, buflen;
	_integer1 __F* buf;

	if( flag&IO_FLAG_WRITE ) return NULL; /* write lock is unsupported now */
	if( size>0 ) return NULL; /* explicit size is unsupported now */

	buf = stdiolock->buf;
	bufpos = stdiolock->bufpos;
	buflen = stdiolock->buflen;

	if( bufpos<buflen ) {
		*locked = buflen - bufpos;
	}
	else {
		if( stdiolock->state==STDIO_NORMAL && bufpos==0 ) {
			if( stdiolock->prompt ) stdiolock->prompt( stdiolock );
			if( fgets(buf,stdiolock->bufsize-1,stdiolock->in) ) {
				stdiolock->state = STDIO_NORMAL;
				stdiolock->buflen = *locked = strlen(buf);
			}
			else {
				stdiolock->state = IO_END;
				stdiolock->buflen = *locked = 0;
			}
		}
		else {
			stdiolock->state = STDIO_BREAK;
			stdiolock->buflen = *locked = 0;
			*locked = 0;
		}

		bufpos = stdiolock->bufpos = 0;
	}
	return buf + bufpos;
}



/* stdioUnlockBuffer */

void __API
stdioUnlockBuffer
	P_B(( stdiolock, current ))
	P_1		struct _STDIOLOCKBUFFER* stdiolock
	P_N		_integer1 __F* current
	P_E
{
	stdiolock->bufpos = current - stdiolock->buf;
}


__MODULEEND__



__MODULEBEGIN__( "frdstd.c" )
/* fileReadStdIO */

#if __ARCH__<=16


static _unsigned2 __API stdio_read
P_B(( io, p, size ))
P_1  struct _IO __F* io
P_N  void __F* p
P_N  _unsigned2 size
P_E
{
 return fread( p, 1, size, (FILE __F*)io->handle );
}


long __API stdioRead
P_B(( io, p, size ))
P_1  struct _IO __F* io
P_N  void __F* p
P_N  long size
P_E
{
 return ioEnhance( stdio_read, io, p, size );
}


#else /* __ARCH__>16 */


long __API
fileReadStdIO( io, p, size )
	struct _IO __F* io;
	void __F* p;
	long size;
{
	return fread( p, 1, size, (FILE __F*)io->handle );
}


#endif /* __ARCH__<=16 */



__MODULEEND__



__MODULEBEGIN__( "fwrstd.c" )
/* fileWriteStdIO */

#if __ARCH__<=16


static _unsigned2 __API stdio_write
P_B(( io, p, size ))
P_1  struct _IO __F* io
P_N  void __F* p
P_N  _unsigned2 size
P_E
{
 return fwrite( p, 1, size, (FILE __F*)io->handle );
}


long __API stdioWrite
P_B(( io, p, size ))
P_1  struct _IO __F* io
P_N  const void __F* p
P_N  long size )
P_E
{
 return ioEnhance( stdio_write, io, (void __F*)p, size );
}


#else


long __API
fileWriteStdIO( io, p, size )
	struct _IO __F* io;
	void __F* p;
	long size;
{
 return fwrite( p, 1, size, (FILE*)io->handle );
}


#endif /* __ARCH__<=16 */
__MODULEEND__



__MODULEBEGIN__( "fseekstd.c" )
/* fileSeekStdIO */

long __API
fileSeekStdIO( io, offset, offset_high, pos )
	struct _IO __F* io;
	long offset;
	long __F* offset_high;
	int pos;
{
 register FILE __F* fp = (FILE __F*)io->handle;

 /* current position optimization */
 if( pos==IO_SEEK_CURRENT && offset==0 ) return ftell(fp);

 /* seek and return the current position */
 return fseek( fp, offset, pos )==0 ? ftell(fp) : FAIL_IO;
}

__MODULEEND__


/* EOF */

