/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Directory Operations */



#include <coralsrc/memory.h>
#include <coralsrc/list.h>
#include <coralsrc/string.h>
#include <coralsrc/dir.h>
#include <coral/compat/string.h>
#include <coral/compat/dirent.h>
#include <coral/compat/stat.h>



__MODULEBEGIN__( "dcreate.c" )
/* dirCreate */

int __API
dirCreate( path, mode, flag, pos )
	const char __F* path;
	unsigned mode;
	unsigned flag;
	int __F* pos;
{
	const char __F* delimiter = strGetDirectorySeparator();
	char __F* s = strCopy( NULL, path, 0 );
	int len = strlen(s);
	if( s ) {
		struct stat st;
		int i, depth;

		if( flag&DIR_FLAG_FILE_NAME )
			strSplitPath( NULL, NULL, NULL, s );

		/* reduce the last directory separator if it is present */
		while( (i=strBackwardTokenIndex(s,len,delimiter,0))>=0 ) {
			if( i<len-1 ) break; /* XXX test is needed for MacOS */
			len = i;
		}

		/* get non-exsiting directory name */
		s[len] = 0;
		depth = 0;
		while( stat(s,&st)==-1 ) {
			char __F* current = strBackwardToken(s,len,delimiter,0);
			if( current==NULL ) break;
			len = current-s-1;
			depth++;
		}

		if( pos ) *pos = len;

		/* create directories */
		if( depth>0 ) {
			i = depth;
			while( i-->0 ) {
				strStoreDirectorySeparator(s,len);
				if( mkdir( s, mode )==-1 ) {
					depth = FAIL_UNAVAILABLE;
					break;
				}
				len+=strlen(s+len);
			}
		}
		memFree( s );
		return depth; /* depth */
	}
	return FAIL_MEMORY;
}
__MODULEEND__



__MODULEBEGIN__( "dtrace.c" )
/* dirTrace */


/*
#define _SUPPORT_SYMLINK
#define _SUPPORT_CORAL_SYMLINK


DECLARE_LIST


static DIR* _API file_state
 ( char __F* dirname, unsigned flag )
{
 struct stat buf;
 if( lstat(&buf)==0 )
 {
  if( S_ISDIR(buf.st_mode) )
   return opendir(dirname);
  else if( S_ISLNK(buf.st_mode) )
  {
   if( flag&FILE_FLAG_SYMLINK_TRACE )
   {
   }
  }
 }
 return NULL;
}*/


int __API
dirTrace
	P_B(( dirname, trace, dt, flag ))
	P_1		const char __F* dirname
	P_N		int (__API* trace) P_A(( const char __F*name,struct _DIRTRACE __F* dt ))
	P_N		struct _DIRTRACE __F* dt
	P_N		unsigned flag
	P_E
{
	DIR* dir;

	DECLARE_STACK( _dir_trace_stack )
		DIR* dir;                /* posix directory object pointer */
		const char __F* dirname; /* old entry name pointer */
		char newname[1];         /* new entry name buffer */
	DECLARE_END __F* ftstk=NULL;

	if( !dirname ) dirname=".";
	dir = opendir( dirname );

	if( dir ) {

		if( dt ) dt->flag = 0;

		while( 1 ) {
			struct dirent* entry;
			int result;
			while( (entry=readdir(dir))!=NULL ) {

				if( entry->d_name[0]!='.' ) {
					char __F* newname;
					DIR* newdir;

					/* push current directory object : though directory trace
						flag was not specified, this operation is needed to
						generate current file name. */
					if( !listPush( (struct _LIST __F*__F*)&ftstk,
							sizeof(struct _dir_trace_stack) +
							strlen(dirname)+strlen(entry->d_name)+2 ) ) {
						closedir( dir );
						while( ftstk ) {
							closedir( ftstk->dir );
							listPop( (struct _LIST __F*__F*)&ftstk );
						}
						return FAIL_MEMORY;
					}

					/* initialize current stack */
					ftstk->dir = dir;
					ftstk->dirname = dirname;

					/* generate new name */
					strcpy( ftstk->newname, dirname );
					newname = ftstk->newname;
					strAppendDirectorySeparator( newname );
					strcat( newname, entry->d_name );

					/* extra information to direct file name only */
					if( dt ) dt->name = entry->d_name;

					/* try recursion */
					newdir = opendir( newname );
					if( newdir ) {
						if( flag&DIR_FLAG_NOTIFY ) {
							/* assume dt was evidently supplied */
							dt->flag = DIR_FLAG_NOTIFY;
							result = trace( newname, dt );
							if( result!=0 ) {
								closedir( newdir );
								goto l_terminate;
							}
							/* clear current attribute for next notification */
							dt->flag = 0;
						}

						if( flag&DIR_FLAG_RECURSIVE ) {
							dir = newdir;
							dirname = newname;
							continue;
						}

						closedir( newdir );
						/* pop to recover previous state */
						goto l_popstack;
					}
					else { /* if( the found file is not a directory ) */
						result = trace( newname, dt );
						if( result!=0 ) goto l_terminate;
						listPop( (struct _LIST __F*__F*)&ftstk );
					} /* end if( dir ) */

				} /* end if( found directory name is not . or .. ) */
			} /* end while( readdir successful ) */

			closedir( dir );

			/* check stack */
			if( ftstk )	{

				if( flag&DIR_FLAG_LEAVE_NOTIFY ) {
					dt->flag = DIR_FLAG_LEAVE_NOTIFY;
					result = trace( ftstk->newname, dt );
					if( result!=0 ) {
						l_terminate:
						/* clear all instances,
							check currently stack-pushed and this means all opened
							directory can be closed by the following iteration */
						while( ftstk ) {
							closedir( ftstk->dir );
							listPop( (struct _LIST __F*__F*)&ftstk );
						}
						return result;
					}
					/* clear current attribute for next notification */
					dt->flag = 0;
				}


				l_popstack:
				dir = ftstk->dir;
				dirname = ftstk->dirname;
				listPop( (struct _LIST __F*__F*)&ftstk );
			}
			else {
				return 0;
			}
		} /* end while( 1 ) */
	} /* end if( dir ) */

	return FAIL_IO;
}

__MODULEEND__


/* EOF */

