/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA */



/* Coral Extension : Process and thread ( light weight process : LWP )

 Restrictions to simplify portable implementations

 1. Thread validity check is not implemented.

 2. Thread cancelation and forced termination are not supported because
	they can make a thread to be blocked when a mutex is locked by the thread.

 3. If threadWait is executed at the same time, execution will be
	unpredictable. ( No error code is returned )

 4. In addition, threadWait and threadDetach must not be used at the same
	time. All threads created by following implementation are initially
	detached and when waiting is not needed, threadDetach must be called
	for the thread which can be waited but actually wasn't wated. */


#define _GNU_SOURCE


#include <coralsrc/memory.h>
#include <coralsrc/list.h>


#ifdef __WINDOWS__

# include <windows.h>

/* because heavy window.h need not to be included in the header, critical
 section context is defined here */

typedef struct _thread_mutex {
	HANDLE					 handle;
	CRITICAL_SECTION cs;
} __F* THREADMUTEX;


#define THREAD_MUTEX_DEFINED

#endif /* __WINDOWS__ */


#include <coralsrc/ext/process.h>
#include <coral/dl.h>
#include <coral/inline.h>

__MODULEEND__ /* end of common part */



#ifdef __UNIX__




__MODULEBEGIN__( "tinit.c" )
/* threadInitialize */

_bool __API
threadInitialize P_A(( void ))
{
	dlInitialize();
	return 1;
}
__MODULEEND__



__MODULEBEGIN__( "tfinal" )
/* threadFinalize */

void __API
threadFinalize P_A(( void ))
{
	dlFinalize();
}
__MODULEEND__



__MODULEBEGIN__( "tcreatm.c" )
/* threadCreateMutex */

int __API
threadCreateMutex( mutex, flag )
	THREADMUTEX __F* mutex;
	unsigned flag;
{
	pthread_mutex_t __F* m;
	pthread_mutexattr_t attr;
	int result;

	pthread_mutexattr_init( &attr );

	if( flag&MUTEX_FLAG_RECURSIVE )
		pthread_mutexattr_settype( &attr, PTHREAD_MUTEX_RECURSIVE );

	m = (pthread_mutex_t __F*) memAlloc( sizeof(pthread_mutex_t) );
	if( m ) {
		pthread_mutex_init( m, &attr );
		*mutex = m;
		result = 0;
	}
	else {
		result = FAIL_MEMORY;
	}

	pthread_mutexattr_destroy( &attr );

	return result;
}



/* threadRemoveMutex */

void __API
threadRemoveMutex( THREADMUTEX mutex )
{
	if( mutex ) {
		pthread_mutex_destroy( mutex );
		memFree( mutex );
	}
}

__MODULEEND__


#else
#ifdef __WIN32__



THREADCONTEXT thread_context;
int thread_init_count;



#ifdef XXX
__MODULEBEGIN__( "pfork.c" )
/* processFork */

int __API processFork
P_B(( pid, cmd, arg, env ))
P_1	PROCESSCONTEXT __F* pid
P_N	const char __F* cmd
P_N	const char __F* const __F* arg
P_N	const char __F* const __F* env
P_E
{

}
__MODULEEND__
#endif



__MODULEBEGIN__( "tglobal.c" )
THREADKEY thread_context_key;
__MODULEEND__



__MODULEBEGIN__( "tfree.c" )
/* thread_free */

void
thread_free
P_B(( thread ))
P_1	THREADCONTEXT thread
P_E
{
	CloseHandle( (HANDLE)thread->handle );
	dlBegin();
	listDNRemove( (struct _LIST __F*__F*)&thread_context, (struct _LIST __F*)thread );
	dlEnd();
}

__MODULEEND__



__MODULEBEGIN__( "tinit.c" )
/* threadInitialize

 Thread API initializer.
	threadInitialize and threadFinalize are not thread-safe functions */

_bool __API threadInitialize
P_V
{
	int result=0;

	if( dlInitialize() ) {

		dlBegin();
		if( OP_INITIALIZE(thread_init_count) ) {
			if( threadCreateKey(&thread_context_key)==0 )
				result = 1;
		}
		else {
			result = 1;
		}
		dlEnd();
	}

	return result;
}

__MODULEEND__



__MODULEBEGIN__( "tfinal.c" )
/* threadFinalize
 See comment in threadInitialize. */

void __API threadFinalize
P_V
{
	dlBegin();

	if( OP_FINALIZE(thread_init_count) ) {

		while( thread_context )
			thread_free( thread_context );

		threadRemoveKey( thread_context_key );
	}
	dlEnd();
	dlFinalize();
}

__MODULEEND__



__MODULEBEGIN__( "tfork.c" )
/* threadFork */

static DWORD WINAPI thread_proc
P_B(( param ))
P_1	THREADCONTEXT thread
P_E
{
	void __F* result;

	threadSetSpecific( thread_context_key, thread );

	result = thread->start( thread->param );

	/* it is possible to use DllMain and receive DLL_THREAD_ATTACH or DLL_THREAD_DETACH
	 to allocate and free some memories for a thread, but DLL_THREAD_DETACH is not
	 posted when
	 1. a process is terminated before a thread in the process is terminated
	 2. a thread is terminated using TerminateThread
	 - eventually the same effect can be obtained by the following routine
	 See @cmt0008 */

	/* if thread must be waited, thread will be freed in threadWait */
	if( !(thread->flag&THREAD_FLAG_WAIT) )
		thread_free( thread );

	return (DWORD)result;
}

int __API threadFork
P_B(( thread, proc, param, stacksize, priority, flag ))
P_1	THREADCONTEXT __F* thread
P_N	void __F* (__API* start) P_A(( void __F* param ))
P_N	void __F* param
P_N	unsigned long stacksize
P_N	int priority
P_N	unsigned flag
P_E
{

	THREADCONTEXT newthread;

	dlBegin();
	newthread = (THREADCONTEXT)
		listDNCreate( (struct _LIST __F*__F*)&thread_context, NULL,
			sizeof(struct _thread_context), LIST_FLAG_ADD_PREV );
	dlEnd();

	if( newthread ) {

		DWORD tid;

		/* create suspended thread */
		void __F* handle = (void __F*) CreateThread( NULL, stacksize,
		 (LPTHREAD_START_ROUTINE)thread_proc, newthread, CREATE_SUSPENDED, &tid );

		if( handle!=NULL ) {

			/* initialize thread paramters */
			newthread->handle = handle;
			newthread->start = start;
			newthread->param = param;
			newthread->flag = flag;
			threadPriority( newthread, priority );

			/* resume thread */
			ResumeThread( (HANDLE)handle );

			if( thread )
				*thread = newthread;

			return 0;
		} /* end if( handle!=NULL ) */

	} /* end if( threadparam ) */

	return 1;
}
__MODULEEND__



__MODULEBEGIN__( "tdetach.c" )
/* threadDetach */

void __API
threadDetach
P_B(( thread ))
P_1	THREADCONTEXT thread
P_E
{
	if( SuspendThread( thread->handle )==0xffffffffL ) {
		/* thread was already terminated but not waited */
		thread_free( thread );
	}
	else {
		/* thread is still running so disable waiting and thread will clear
		 its resources when it is terminated - if threadDetach and threadWait
		 can be called at the same time, more complex implementation is needed
		 but that case is restricted ( see restrictions at the beginning of
		 this source ) */
		OP_DISABLE( thread->flag, THREAD_FLAG_WAIT );
		ResumeThread( thread->handle );
	}
}

__MODULEEND__



__MODULEBEGIN__( "tself.c" )
/* threadSelf */

void __F* __API threadSelf
P_V
{
	return threadGetSpecific( thread_context_key );
}

__MODULEEND__



__MODULEBEGIN__( "twait.c" )
/* threadWait */

int __API
threadWait
P_B(( thread, result ))
P_1	THREADCONTEXT thread
P_N	void __F*__F* result
P_E
{
	void __F* value;
	int errorcode;

	if( threadEqual(threadSelf(),thread) ) return THREAD_DEADLOCK;

	errorcode = THREAD_DETACHED;

	if( thread->flag&THREAD_FLAG_WAIT ) {

		HANDLE handle = thread->handle;

		if( WaitForSingleObject( handle, INFINITE )!=WAIT_FAILED &&
				GetExitCodeThread( handle, (LPDWORD)&value ) && result ) {
			*result = value;
			errorcode = 0;
		}

		thread_free( thread );

	} /* end if( thread need be waited ) */

	return errorcode;
}

__MODULEEND__



__MODULEBEGIN__( "tcancel.c" )
/* threadCancel */

int __API
threadCancel
P_B(( thread ))
P_1	THREADCONTEXT thread
P_E
{
	TerminateThread( thread->handle, -1 );
	/* if thread must be waited, thread will be freed in threadWait */
	if( !(thread->flag&THREAD_FLAG_WAIT) )
		thread_free( thread );

	return 0;
}


__MODULEEND__



__MODULEBEGIN__( "tprio.c" )
/* threadPriority */

void __API
threadPriority
P_B(( thread, priority ))
P_1	THREADCONTEXT thread
P_N	int priority
P_E
{
	static int prioritymap[] = {
	 THREAD_PRIORITY_IDLE,
	 THREAD_PRIORITY_LOWEST,
	 THREAD_PRIORITY_BELOW_NORMAL,
	 THREAD_PRIORITY_NORMAL,
	 THREAD_PRIORITY_ABOVE_NORMAL,
	 THREAD_PRIORITY_HIGHEST,
	 THREAD_PRIORITY_TIME_CRITICAL,
	};

	/* set thread priority can be applied for the pseudo handle obtained by
	 GetCurrentThread */
	SetThreadPriority( (HANDLE)thread->handle, prioritymap[priority] );
}
__MODULEEND__



__MODULEBEGIN__( "texit.c" )
/* threadExit */

void __API threadExit
P_B(( void __F* result ))
P_1	void __F* result
P_E
{
	THREADCONTEXT thread = (THREADCONTEXT)threadGetSpecific(thread_context_key);
	if( !(thread->flag&THREAD_FLAG_WAIT) )
		thread_free( thread );

	ExitThread((unsigned long)result);
}
__MODULEEND__



__MODULEBEGIN__( "tcreatk.c" )
/* threadCreateKey */

int __API
threadCreateKey
P_B(( key ))
P_1	THREADKEY __F* key
P_E
{
	return (unsigned long)( *key=(THREADKEY)TlsAlloc() )!=0xffffffffL? 0 : -1;
}
__MODULEEND__



__MODULEBEGIN__( "tcreatm.c" )
/* threadCreateMutex */

BOOL (__API* mutex_try_enter_critsect) P_A(( LPCRITICAL_SECTION cs ));
_bool mutex_try_enter_tested=1;

int __API	threadCreateMutex
P_B(( mutex, flag ))
P_1	THREADMUTEX __F* mutex
P_N	unsigned flag
P_E
{
	THREADMUTEX m;

	if( flag&MUTEX_FLAG_TRYLOCK ) {

		if( !mutex_try_enter_tested ) {
			void __F* dl = dlOpen( "kernel32", 0 );
			if( !dl ) return MUTEX_ERROR;

			mutex_try_enter_critsect = (BOOL(__API*)P_A((LPCRITICAL_SECTION)))
				dlSymbol( dl, "TryEnterCriticalSection" );

			dlClose( dl );

			mutex_try_enter_tested = 1;
		}

		/* TryEnterCriticalSection is unsupported by the kernel */
		if( !mutex_try_enter_critsect ) {

			m = (THREADMUTEX)memAlloc( sizeof(HANDLE) );
			if( !m ) return MUTEX_ERROR;

			if( (m->handle=CreateMutex(NULL,FALSE,NULL))==NULL ) {
				memFree( m );
				return MUTEX_ERROR;
			}

		}
		else goto l_critsect;

	}
	else {
		l_critsect:
		m = (THREADMUTEX) memCAlloc( sizeof(struct _thread_mutex) );
		if( !m ) return MUTEX_ERROR;
		InitializeCriticalSection( &m->cs );
	} /* end if( flag&MUTEX_FLAG_TRYLOCK ) */

	*mutex = m;

	return 0;
}



/* threadRemoveMutex */

void __API threadRemoveMutex
P_B(( mutex ))
P_1	THREADMUTEX mutex
P_E
{
	if( mutex ) {
		if( mutex->handle )
			CloseHandle( mutex->handle );
		else
			DeleteCriticalSection( &mutex->cs );
		memFree( mutex );
	}
}
__MODULEEND__



__MODULEBEGIN__( "mttry.c" )
/* threadTryLockMutex */

int __API threadTryLockMutex
P_B(( mutex ))
P_1	THREADMUTEX mutex
P_E
{
	if( mutex_try_enter_critsect )
		return mutex_try_enter_critsect( &mutex->cs )==TRUE ? 0 : MUTEX_BUSY;

	switch( WaitForSingleObject( mutex->handle, 0 ) )
	{
		case WAIT_OBJECT_0 : break;
		case WAIT_TIMEOUT : return MUTEX_BUSY;
		default: return MUTEX_ERROR;
	}
	return 0;
}

__MODULEEND__



__MODULEBEGIN__( "mtlock.c" )
/* threadLockMutex */

void __API threadLockMutex
P_B(( mutex ))
P_1	THREADMUTEX mutex
P_E
{
	if( mutex ) {
		if( mutex->handle )
			WaitForSingleObject( mutex->handle, INFINITE );
		else
			EnterCriticalSection( &mutex->cs );
	}
}



/* threadUnlockMutex */

void __API threadUnlockMutex
P_B(( mutex ))
P_1	THREADMUTEX mutex
P_E
{
	if( mutex ) {
		if( mutex->handle )
			ReleaseMutex( mutex->handle );
		else
			LeaveCriticalSection( &mutex->cs );
	}
}

__MODULEEND__



__MODULEBEGIN__( "concreat.c" )
/* threadCreateCond */

int __API
threadCreateCond
P_B(( cond ))
P_1	THREADCOND __F* cond
P_E
{
	THREADCOND c;

	c = (THREADCOND)CreateSemaphore(NULL,0,OP_MAXINT(LONG),NULL);
	if( c ) {
		*cond = c;
		return 0;
	}
	return -1;
}


/* threadRemoveCond */

void __API
threadRemoveCond
P_B(( cond ))
P_1	THREADCOND __F* cond
P_E
{
	CloseHandle( (HANDLE)cond );
}

__MODULEEND__



__MODULEBEGIN__( "consig.c" )
/* threadSignalCond */

void __API
threadSignalCond
P_B(( cond ))
P_1	THREADCOND* cond
P_E
{
//	if( WaitForSingleObject((HANDLE)cond)
//	ReleaseSemaphore( (HANDLE)cond );
}
__MODULEEND__



__MODULEBEGIN__( "conbroad.c" )
/* threadBroadcastCond */
void __API
threadBroadcastCond
P_B(( cond ))
P_1	THREADCOND __F* cond
P_E
{
	
}
__MODULEEND__

#endif /* __WINDOWS__ */
#endif /* __UNIX__ */



/* end of thread.c */

