/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Coral File Operations */

/* 1997.10.  Started
   1999.08.14  Modified for open I/O buffer API */


#ifdef __hpux
# define _HPUX_SOURCE
#endif

#include <stdio.h>
#include <coralsrc/memory.h>
#include <coralsrc/string.h>
#include <coralsrc/list.h>
#include <coralsrc/io.h>
#include <coralsrc/file.h>

#include <coral/compat/string.h>
#include <coral/inline.h>
#include <coral/compat/unistd.h>
#include <coral/compat/stat.h>
#include <fcntl.h>

#ifdef __WIN32__
#include <windows.h>
#endif



/*------------------------------------------------------------------------*
   Coral Specific I/O Routines
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "fopenl.c" )
/* file_open */

#ifndef O_BINARY
#	define O_BINARY 0
#endif

int
file_open
	P_B(( io, filename, flag, size, p_fd ))
	P_1		struct _IO __F* io
	P_N		const char __F* filename
	P_N		unsigned flag
	P_N		unsigned size
	P_N		int __F* p_fd
	P_E
{
	int m, result;

	if( flag&IO_FLAG_READ ) {
		if( flag&IO_FLAG_WRITE ) {
			m = O_RDWR|O_BINARY|O_CREAT;
			goto l_write_mode;
		}
		else {
#			ifdef __UNIX__
			struct stat st;
			/* on unix, directory can be opened by 'open' for read-only mode,
			 so stat need to be called */
			if( stat(filename,&st)!=0 || S_ISDIR(st.st_mode) )
				return FAIL_IO;
#			endif
			m = O_RDONLY|O_BINARY;
		}
	}
	else if( flag&IO_FLAG_WRITE ) {
		m = O_WRONLY|O_BINARY|O_CREAT;
		l_write_mode:
		if( flag&FILE_FLAG_EXISTENCE ) {
			if( access(filename,F_OK)!=0 ) return FAIL_ENTRY;
		}
		else if( !(flag&FILE_FLAG_OVERWRITE) ) {
			/* if the given file name is present, opening is failed */
			if( access(filename,F_OK)==0 ) return FAIL_DUPLICATED;
		}
	}
	else {
		return FAIL_PARAM;
	}

	if( OP_ALLFLAG(flag,IO_FLAG_WRITE|FILE_FLAG_TRUNCATE) )
		m |= O_TRUNC;

	result = ioOpen( io, size );
	if( result==0 ) {
		int fd = open(filename,m,S_IREAD|S_IWRITE);
		if( fd==FAIL_IO ) {
			memFree( ioClose(io,NULL) );
			return FAIL_IO;
		}
		*p_fd = fd;
	}

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "fopen.c" )
/* fileOpen */

int __API
fileOpen
	P_B(( io, filename, flag ))
	P_1		struct _IO __F* io
	P_N		const char __F* filename
	P_N		unsigned flag
	P_E
{
	int fd;
	int result = file_open( io, filename, flag, sizeof(int), &fd );
	if( result==0 ) *(int __F*)io->handle = fd;
	return result;
}



/* fileClose */

int __API
fileClose
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	int result;
	void __F* handle = ioClose( io, &result );
	if( handle ) {
		if( close( *(int __F*)handle )!=0 ) result = FAIL_IO;
		memFree( handle );
	}
	return result;
}

__MODULEEND__



__MODULEBEGIN__( "fread.c" )
/* fileRead */



#if __ARCH__<=16

#ifdef __FARPTR__

/* if far and near pointer size is different, __FARPTR__ is defined */

_unsigned2 __API
file_short_read_proc
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		_unsigned2 size
	P_E
{
	void __F* cp;
	_unsigned1 interbuf[_IO_DEFAULT_BUFFER];
	_unsigned2 total;
	long longsize;

	total = 0;
	cp = p;
	longsize = size;
	while( longsize > 0 ) {
		/* calculate opeartion size */
		_unsigned2 opsize = memLongIter( &cp, &longsize, sizeof(interbuf) );
		_unsigned2 rbytes;

		/* read op bytes to intermediate buffer */
		rbytes = read( (int)io->handle, interbuf, opsize );

		/* transfer bytes in the intermediate buffer to destination */
		mem_short_copy( p, interbuf, opsize );

		total += rbytes;
		if( rbytes<opsize ) {
			if( rbytes==(_unsigned2)FAIL_IO ) return ERROR_IO;
			break;
		}
		p = cp;
	}
	return total;
}

#else /* !__FARPTR__ */

_unsigned2 __API
file_short_read_proc
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		_unsigned2 size
	P_E
{
	return read( (int)io->handle, p, size );
}

#endif /* __FARPTR__ */

#endif /* __ARCH__<=16 */



/* fileRead */

long __API
fileRead
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long size
	P_E
{
 /* WIN16 */
#if defined(__WIN16__)
	return _hread( (int)io->handle, p, size );

 /* 16 bit architecture */
# elif __ARCH__<=16
	return ioEnhance( io, p, size, file_short_read_proc );

 /* POSIX */
# else
	return read( *(int __F*)io->handle, p, size );

# endif
}
__MODULEEND__



__MODULEBEGIN__( "fwrite.c" )
/* fileWrite */



#if __ARCH__<=16

#ifdef __FARPTR__

_unsigned2 __API
file_short_write_proc
P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		_unsigned2 size
	P_E
{
	void __F* cp;
	_unsigned1 interbuf[_IO_DEFAULT_BUFFER];
	_unsigned2 total;
	long longsize;

	total = 0;
	cp = p;
	longsize = size;
	while( longsize > 0 ) {
		/* calculate the size to be operated */
		_unsigned2 opsize = mem_long_iter( &cp, &longsize, sizeof(interbuf) );
		_unsigned2 wbytes;

		/* transfer bytes into intermediate buffer */
		mem_short_copy( interbuf, p, opsize );

		/* swap intermediate buffer */
		wbytes = write( (int)io->handle, interbuf, opsize );

		total += wbytes;
		if( wbytes<opsize ) {
			if( wbytes==(_unsigned2)-1 ) return -1;
			break;
		}

		p = cp;
	}
	return total;
}

#else /* !__FARPTR__ */

_unsigned2 __API
file_short_write_proc
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		_unsigned2 size
	P_E
{
	return write( (int)io->handle, p, size );
}

#endif /* __FARPTR */

#endif /* __ARCH__<=16 */



/* fileWrite */

long __API
fileWrite
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long size
	P_E
{
 /* WIN16 */
#ifdef __WIN16__

	return _hwrite( (int)io->handle, p, size );

 /* 16 bit platform */
#else
#if __ARCH__<=16
	return ioEnhance( io, p, size, file_short_write_proc );

 /* POSIX */
#else

	return write( *(int __F*)io->handle, p, size );

#endif /* __ARCH__<=16 */
#endif /* __WIN16__ */
}

__MODULEEND__



__MODULEBEGIN__( "fseek.c" )
/* fileSeek */

long __API
fileSeek
P_B(( io, offset, offset_high, seekpos ))
	P_1		struct _IO __F* io
	P_N		long offset
	P_N		long __F* offset_high
	P_N		int seekpos
	P_E
{
#ifdef __WIN16__
	return _llseek( (int)io->handle, offset, seekpos );
#else
	offset = lseek( *(int __F*)io->handle, offset, seekpos );
	if( offset==-1 && offset_high ) {
    	*offset_high = -1;
	}
	return offset;
#endif /* __WIN16__ */
}
__MODULEEND__



__MODULEBEGIN__( "fseteof.c" )
/* XXX not implemented yet */
__MODULEEND__



#ifdef __WIN32__
__MODULEBEGIN__( "fopenn.c" )
/* fileOpenNative */

int __API
fileOpenNative
	P_B(( io, filename, flag ))
	P_1		struct _IO __F* io
	P_N		const char __F* filename
	P_N		unsigned flag
	P_E
{
	int result;
	unsigned long m;

	/* parse opening mode */
	if( flag&IO_FLAG_READ ) {
		if( flag&IO_FLAG_WRITE ) {
			m = GENERIC_READ|GENERIC_WRITE;
			goto l_write_mode;
		}
		else {
			m = GENERIC_READ;
		}
	}
	else if( flag&IO_FLAG_WRITE ) {
		m = GENERIC_WRITE;

		l_write_mode:
		if( flag&FILE_FLAG_EXISTENCE ) {
			/* error or directory : note that error code is 0xffffffff */
			if( GetFileAttributes(filename)&FILE_ATTRIBUTE_DIRECTORY )
				return FAIL_ENTRY;
		}
		else if( !(flag&FILE_FLAG_OVERWRITE) ) { /* overwriting was not requested */
			/* if the given file name is present, opening is failed */
			if( GetFileAttributes(filename)!=0xffffffffUL )
				return FAIL_DUPLICATED;
		}
	}
	else {
		return FAIL_PARAM;
	}

	result = ioOpen( io, 0 );
	if( result==0 ) {
		/* open a file */
		HANDLE handle =
			CreateFile( filename, m,
				FILE_SHARE_READ|FILE_SHARE_WRITE, NULL,
				flag&IO_FLAG_WRITE ? OPEN_ALWAYS : OPEN_EXISTING,
				FILE_ATTRIBUTE_NORMAL, NULL );
		if( handle==INVALID_HANDLE_VALUE ) {
			ioClose( io, NULL );
			return FAIL_IO;
		}

		io->handle = handle;

		/* truncate file if requested */
		if( OP_ALLFLAG(flag,
				IO_FLAG_WRITE|FILE_FLAG_OVERWRITE|FILE_FLAG_TRUNCATE) )
			SetEndOfFile( (HANDLE)io->handle );
	}

	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "fclosen.c" )
/* fileCloseNative */

int __API
fileCloseNative
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	int result;
	void __F* handle = ioClose( io, &result );
	if( handle ) {
		CloseHandle( handle );
	}
	return result;
}



__MODULEBEGIN__( "freadn.c" )
/* fileReadNative */

long __API
fileReadNative
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long size
	P_E
{
	DWORD readbytes;
	if( ReadFile( (HANDLE)io->handle, p, size, &readbytes, 0 ) )
		return readbytes;
	return FAIL_IO;
}
__MODULEEND__



__MODULEBEGIN__( "fwriten.c" )
/* fileWriteNative */

long __API
fileWriteNative
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long size
	P_E
{
	DWORD wbytes;

	if( WriteFile( (HANDLE)io->handle, p, size, &wbytes, 0 ) )
		return wbytes;

	return -1;
}
__MODULEEND__



__MODULEBEGIN__( "fseekn.c" )
/* fileSeekNative */

long __API
fileSeekNative
P_B(( io, offset, offset_high, seekpos ))
	P_1		struct _IO __F* io
	P_N		long offset
	P_N		long __F* offset_high
	P_N		int seekpos
	P_E
{
  DWORD result = SetFilePointer
		( (HANDLE)io->handle, offset, offset_high, seekpos );
  if( result==0xffffffff ) {
    if( GetLastError()!=NO_ERROR )
      if( offset_high ) *offset_high = -1;
  }
  return result;
}
__MODULEEND__



__MODULEBEGIN__( "fseteofn.c" )
/* fileSetEOFNative */

_bool __API
fileSetEOFNative( struct _IO __F* io )
{
	return SetEndOfFile( (HANDLE)io->handle );
}
__MODULEEND__



#endif /* __WIN32__ */



__MODULEBEGIN__( "fload.c" )
/* fileLoad */

char __F* __API
fileLoad
	P_B(( filename, buf, offset, offset_high, p_size ))
	P_1		const char __F* filename
	P_N		long offset
	P_N		long __F* offset_high
	P_N		_integer1 __F* buf
	P_N		long __F* p_size
	P_E
{
	struct _IO io;
	char __F* p = NULL;
	long size = 0;

	ioEmpty( &io );
	if( fileOpen(&io,filename,IO_FLAG_READ)==0 ) {

		if( p_size )
			size = *p_size;

	    if( size<=0 ) {
			if( offset_high==NULL )
    			size = (long) fileSeek(&io,0,NULL,IO_SEEK_END) - offset;
			if( size<=0 ) goto l_error;
		}

		if( buf ) {
			p = buf;
		}
		else {
			p = (char __F*) memAlloc( size );
			if( p==NULL ) goto l_error;
		}

		if( fileSeek( &io, offset, offset_high,  IO_SEEK_BEGIN )==-1 ||
		    fileRead( &io, p, size )!=size ) {
			if( buf==NULL ) {
				memFree( p );
				p = NULL;
			}
		}

		l_error:
		fileClose( &io );
	} /* end if( fileOpen ) */

	if( p_size )
		*p_size = size;

	return p;
}
__MODULEEND__



__MODULEBEGIN__( "fsave.c" )
/* fileSave */

_bool __API
fileSave
	P_B(( filename,  offset, offset_high, buf, size, flag ))
	P_1		const char __F* filename
	P_N		long offset
	P_N		long __F* offset_high
	P_N		_integer1 __F* buf
	P_N		long size
	P_N		unsigned flag
	P_E
{
	struct _IO io;
	_bool result = 0;

	ioEmpty( &io );
	if( fileOpen(&io,filename,IO_FLAG_WRITE|flag)==0 ) {
		if( fileSeek( &io, offset, offset_high,  IO_SEEK_BEGIN )!=-1 &&
		    fileWrite( &io, buf, size )==size )
			result = 1;
		fileClose( &io );
	} /* end if( fileOpen ) */

	return result;
}
__MODULEEND__




#if 0 /* XXX not completed */
__MODULEBEGIN__( "fcopy.c" )
/* fileCopy

srcoffset : 0, positive, negative(use destoffset)
destoffset : 0, positive, negative(end of file) */

//??? total size written can be calculated ?
//??? initial offset can be reported

int __API fileCopy
P_B(( destname, srcname, destoffset, srcoffset, totalsize, buf, bufsize, flag, progress, context ))
	P_1		const char __F* destname
	P_N		const char __F* srcname
	P_N		long destoffset
	P_N		long srcoffset
	P_N		long totalsize
	P_N		_integer1 __F* buf
	P_N		int bufsize
	P_N		unsigned flag
	P_N		int (__API* progress)(void __F* context,int size)
	P_N		void __F* context
	P_E
{
  struct _IO dest, src;
  int result=FILE_COPY_SRC_NOT_OPENED;;

  memSet( &dest, 0, sizeof(struct _IO) );
  memSet( &src, 0, sizeof(struct _IO) );
  dest.write = fileWrite;
  src.read = fileRead;

  if( fileOpen( &src, srcname, IO_FLAG_READ ) ) {

    if( fileOpen( &dest, destname, IO_FLAG_WRITE|flag ) ) {

      result = FAIL_IO;

      if( destoffset>0 )
        destoffset = fileSeek( &dest, destoffset, IO_SEEK_BEGIN );
      else if( destoffset<0 )

//??? todo: destoffset should be reported to callback function in this case.
        destoffset = fileSeek( &dest, 0, IO_SEEK_END );

      if( destoffset>=0 ) {

        if( srcoffset>0 )
          srcoffset = fileSeek( &src, srcoffset, IO_SEEK_BEGIN );
        else if( srcoffset<0 )
          srcoffset = fileSeek( &src, destoffset, IO_SEEK_BEGIN );

        if( srcoffset>=0 ) {
          if( flag&FILE_FLAG_OFFSET ) {
            ((long __F*)context)[0] = destoffset;
            ((long __F*)context)[1] = srcoffset;
          }
          result = ioTransfer( &dest, &src, totalsize, buf, bufsize, progress, context );
        } /* end if( srcoffset>=0 ) */
          
      } /* end if( destoffset>=0 ) */

      fileClose( &dest );
    }
    else {
      result = FILE_COPY_DEST_NOT_OPENED;
    } /* end if( fileOpen(&dest) ) */

    fileClose( &src );

  } /* end if( fileOpen(&src) ) */

  return result;
}

__MODULEEND__
#endif



/* EOF */

