/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* I/O */

/*
 Restrictions to use Coral Open I/O capabilities
	( summarized on 2000.07.10 )

 1. Procedure definitions issues for vtable entries
	(1) Read, write and seek procedure must return IO_ERROR(-1) if
		an error was occurred
	(2) Seek procedure must use IO_SEEK_BEGIN(0), IO_SEEK_CURRENT(1)
		and IO_SEEK_END(2) to define the stream position.

 2. Requested procedure issues
	 Coral Open I/O class contains 3 major vtable entries to read,
	write and seek an I/O stream and sometimes a procedure must be
	additionally supplied to use I/O buffer.

	(1) Seek procedure must be supplied for ioSyncBuffer if I/O
		stream was opened for read mode
	(2) Seek procedure must be supplied to use buffering for
		read/write operations if I/O stream was opened for read/write mode
	(3) Procedures which are explicitly denoted for a method, on a document.

 3. Buffer management issues
	Coral I/O buffer is independent to the property of an I/O stream
	so the following restrictions are present.

	(1) It is needed to call ioFlushBuffer to close an I/O stream.
	(2) It is needed to call ioTruncateBuffer after the stream
		size is successfully changed.
*/

/*

 To do :
	Further optimization for byte ordering is needed
	I/O buffering optimzation is needed

 1999.08.13
	Buffer API support, buffered file I/O written on 1997.10. was
 generalized.

 1999.03.22  Revision to read multibyte value

 1998.04.30  Module splitter spec.

 1997.10.    Some modifications

 1997.01.    Header creation
*/



#undef __DEBUG__ /* ignore globally defined debug request */
/* #define __DEBUG__ */



#include <coralsrc/memory.h>
#include <coralsrc/io.h>
#include <coralsrc/integer.h>
#include <coral/inline.h>
#include <coral/compat/string.h>



/*------------------------------------------------------------------------*
	Internal methods
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "ioreadyw.c" )
/* io_ready_wmode

 Seek procedure is needed for rw mode. */

int __API
io_ready_wmode
	P_B(( io, n ))
	P_1		struct _IO __F* io
	P_N		long n
	P_E
{
	int remained;
	struct _io_buffer __F* buffer = io->buffer;

	/* ioSyncBuffer refers the last operation so it needs to be evaluated
		before calling ioSyncBuffer */
	if( buffer->lastop==IO_FLAG_READ ) {
		/* by the restrictions mentioned, seek
			procedure is needed for read/write mode to use buffer */
		if( !ioSyncBuffer( io ) ) return FAIL_IO;
	}

	buffer->lastop = IO_FLAG_WRITE;

	remained = buffer->size - buffer->current;

	/* buffer is full ( buffer has no remained bytes ) and the buffer
		size is sufficient to dispatch the current request */
	if( remained==0 && n<=(long)buffer->size ) {
		remained = buffer->size;
		if( io->write( io, buffer->buf, remained )==FAIL_IO )
			return FAIL_IO;
		buffer->current = 0;
		buffer->wend = 0;

		DEBUG_PUTS(( "io_ready_wmode:no remained bytes and buffer size is sufficient" ));
	}

	/* return remained bytes */
	return remained;
}
__MODULEEND__



__MODULEBEGIN__( "ioreadyr.c" )
/* io_ready_rmode
 'n' can be zero to always try to fill buffer.
 Returns FAIL_IO if failed, remained size if successful. */

int __API
io_ready_rmode
	P_B(( io, n ))
	P_1		struct _IO __F* io
	P_N		long n
	P_E
{
	int remained;
	struct _io_buffer __F* buffer = io->buffer;

	/* flush buffer if last operation was writing */
	if( buffer->lastop==IO_FLAG_WRITE && !ioFlushBuffer(io) )
		return FAIL_IO;

	/* calculate remained bytes in the buffer */
	remained = buffer->rend - buffer->current;

	/* no remained bytes in the buffer and buffer size is sufficient
		to dispatch the current request : 'n' must be used for the following
		comparison in the current implementation because supplied value
		of 'n' is actually 'n-1' - see io_ready_rmode in ioRead*/
	if( remained==0 && n<=(long)buffer->size ) {
		/* fill buffer */
		remained = buffer->rend =
			(int)io->read( io, buffer->buf, buffer->size );
		if( remained==FAIL_IO ) return FAIL_IO;
		buffer->current = 0;
		DEBUG_PRINTF(( "io_ready_rmode:no remained cache and %d bytes are cached\n",
			remained ));
#if 0
		int mid = (buffer->rend+1) / 2;
		if( mid>0 ) {
				remained = buffer->rend+1-mid;
				mem_short_copy( buffer->buf,
					buffer->buf + remained, mid-1 );
		}
		buffer->rend =
			mid + io->read( io, buffer->buf + mid, buffer->size-mid );
		buffer->current = mid;
#endif
	}

	buffer->lastop = IO_FLAG_READ;

	return remained;
}
__MODULEEND__



__MODULEBEGIN__( "io_reset.c" )
/* io_reset_buffer

 defined on 2000.07.11 to reduce overhead caused by
 buffer = io->buffer
 if( buffer ) { ... } */

void __API
io_reset_buffer
	P_B(( buffer ))
	P_1		struct _io_buffer __F* buffer
	P_E
{
	buffer->rend = buffer->wend = buffer->current = 0;
	buffer->lastop = -1;
}
__MODULEEND__



__MODULEBEGIN__( "iooprec.c" )
/* io_operate_record

 [n], size, offset
 n is optional and turn on msb to supply it */

_bool __API
io_operate_record
	P_B(( io, p, ci, flag, ioproc ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		const void __F* ri
	P_N		unsigned flag
	P_N		_bool (__API* ioproc) P_A((struct _IO __F* io,void __F* p,long n,unsigned byteorder))
	P_E
{
	unsigned byteorder = flag&MASK_BYTE_ORDER;
	unsigned risize = flag&MASK_COMBINATED_SIZE;
	unsigned long msb = OP_MINBITINT(risize*8), mask=~msb;

	while( 1 ) {

		unsigned long val, v[3], *pv=v;
		int i=2;

		while( i-->0 ) {

			switch( risize ) {
				case 1: val = *(_unsigned1 __F*)ri; break;
				case 2: val = *(_unsigned2 __F*)ri; break;
				case 4: val = *(_unsigned4 __F*)ri; break;
				default: return 0;
			}

			if( pv==v ) { /* at the first time */
				if( val==0 ) return 1;

				if( val&msb ) {
					val &= mask; /* count */
					i++; /* count was specified so read one more value */
				}
				else {
					*pv++ = 1;
				}
			}

			*pv++ = val;
			ri = (_unsigned1 __F*)ri + risize;
		}

		/* n, size, offset */
		if( !ioproc( io, (_unsigned1 __H*)p+v[2], v[0], byteorder|v[1] ) )
			return 0;
	}
}
__MODULEEND__



__MODULEBEGIN__( "ioop2d.c" )
/* io_operate_2d */

_bool __API
io_operate_2d
	P_B(( io, p, row, item, colrow, colcol, variable, byteorder, ioproc ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long row
	P_N		unsigned short item
	P_N		const void __F* colrow
	P_N		const void __F* colcol
	P_N		_unsigned2 variable
	P_N		_unsigned2 byteorder
	P_N		_bool (__API* ioproc) P_A((struct _IO __F* io,void __F* p,long n,unsigned byteorder))
	P_E
{
	register void __F* __H* pp = (void __F* __H*)p;
	while( row-- ) {
		/* dispatch a column */
		if( !ioproc( io, *pp++,
					mem_iterate_2d(item,&colrow,&colcol,variable), byteorder ) )
			return 0;
	}
	return 1;
}

__MODULEEND__



__MODULEBEGIN__( "iotreadc.c" )
/* io_text_readchar

	Some editors use ASC II EOF character ( 0x1a ) at the end of a file and
 this one byte reader considers it.
*/

int __API
io_text_readchar
	P_B(( io, c ))
	P_1		struct _IO __F* io
	P_N		char __F* c
	P_E
{
	int readbytes = (int)ioRead( io, c, 1, 0 );
	if( readbytes!=FAIL_IO && *c==0x1a ) return IO_END;
	/* no need to consider readbytes==0, because IO_END==0 */
	return readbytes;
}
__MODULEEND__




/*------------------------------------------------------------------------*
	 I/O API
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "ioempty.c" )
/* ioEmpty */

void __API
ioEmpty
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	io->read = NULL;
	io->write = NULL;
	io->seek = NULL;
	io->handle = NULL;
	io->buffer = NULL;
	io->op = 0;
}
__MODULEEND__



__MODULEBEGIN__( "ioopen.c" )
/* ioOpen */

int __API
ioOpen
	P_B(( io, size ))
	P_1		struct _IO __F* io
	P_N		unsigned size
	P_E
{
	ioResetBuffer( io );
	return
		io->handle ? FAIL_DUPLICATED :
		size>0 && (io->handle=memAlloc( size ))==NULL ? FAIL_MEMORY :
		0;
}

__MODULEEND__



__MODULEBEGIN__( "ioclose.c" )
/* ioClose */

void __F* __API
ioClose
	P_B(( io, result ))
	P_1		struct _IO __F* io
	P_N		int __F* result
	P_E
{
	void __F* handle = NULL;
	int code = 0;

	if( ioFlushBuffer(io) ) {
		handle = io->handle;
		if( handle ) io->handle = NULL;
		else code = FAIL_UNAVAILABLE;
	}
	else {
		code = FAIL_IO;
	}

	if( result ) *result = code;

	return handle;
}
__MODULEEND__



__MODULEBEGIN__( "ioenhan.c" )
/* ioEnhance

 1997. 8.
 1997. 10.
 16 bit platform support. */

long __API
ioEnhance
	P_B(( io, p, size, ioproc ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long size
	P_N		_unsigned2 (__API* ioproc) P_A(( struct _IO __F* io, void __F* p, _unsigned2 size ))
	P_E
{
	long total = 0;
	void __F* cp = p;
	while( size>0 ) {
		_unsigned2 toread = mem_long_iter
		 ( (char __H* __F*)&cp, &size, _MEM_LONGITER );
		_unsigned2 readbytes = (*ioproc)( io, p, toread );
		total += readbytes;
		if( readbytes<toread ) {
			if( (short)readbytes==-1 ) return -1;
			break;
		}
		p = cp;
	}
	return total;
}
__MODULEEND__



__MODULEBEGIN__( "iobuf.c" )

/* I/O buffer API
 1999.08.12 */


/* ioResetBuffer */

void __API
ioResetBuffer
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	struct _io_buffer __F* buffer = io->buffer;

	if( buffer )
		io_reset_buffer( buffer );
}



/* ioAllocBuffer */

_bool __API
ioAllocBuffer
	P_B(( io, size ))
	P_1		struct _IO __F* io
	P_N		int size
	P_E
{
	_bool first;
	int reserved;
	struct _io_buffer __F* buffer;

	if( size==0 ) {
		/* zero size will allocate _IO_DEFAULT_BUFFER size */
		size=_IO_DEFAULT_BUFFER;
	}
	else if( size<0 ) {
		/* negative size will allocate at least _IO_DEFAULT_BUFFER size */
		size = -size;
		if( size<_IO_DEFAULT_BUFFER )
			size = _IO_DEFAULT_BUFFER;
	}

	/* check buffer was already attached */
	buffer = io->buffer;
	if( buffer ) {
		/* by the following test, buffer size will never be reduced -
		  it is needed because ioSyncBuffer is needed if size is reduced and
		  that function has a restriction which seek procedure must be
		  supplied */
		if( buffer->size >= size ) {
			/* buffer has sufficient size */
			return 1;
		}
		first = 0;
 		/* reserved size is still avaiable because buffer size is never
		   reduced */
		reserved = buffer->reserved;
	}
	else {
		first = 1;
		reserved = 0;
	}

	/* allocate buffer */
	buffer = (struct _io_buffer __F*) memResize( buffer,
		size + ( sizeof(struct _io_buffer)-sizeof(buffer->buf) ) );

	if( buffer ) {
		/* attach buffer to io object */
		io->buffer = buffer;

		/* initialize buffer */
		buffer->size = size;
		buffer->reserved = reserved;

		/* reset buffer for the first time allocation */
		if( first )
			io_reset_buffer( buffer );

		buffer->offset_high = 0;
		buffer->offset = io->handle && io->seek ?
			(unsigned long) io->seek
				( io, 0, &buffer->offset_high, IO_SEEK_CURRENT ) : 0;

		return 1;
	}
	return 0;
}



/* ioFreeBuffer */

_bool __API
ioFreeBuffer
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	_bool result = ioFlushBuffer(io);
	memFree( io->buffer );
	io->buffer = NULL;
	return result;
}

__MODULEEND__ /* I/O buffer API */



__MODULEBEGIN__( "iogetbs.c" )

int __API
ioGetBufferSize
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	struct _io_buffer __F* buffer = io->buffer;
	return buffer==NULL ? 0 : buffer->size;
}
__MODULEEND__



__MODULEBEGIN__( "iosetrbs.c" )

void __API
ioSetReservedBufferSize
	P_B(( io, reserved ))
	P_1		struct _IO __F* io
	P_N		int reserved
	P_E
{
	struct _io_buffer __F* buffer = io->buffer;
	if( buffer && buffer->size>reserved )
		buffer->reserved = reserved;
}
__MODULEEND__



__MODULEBEGIN__( "ioflush.c" )

/* ioFlushBuffer
 Revision was made from fileFlush, 1999.08.13 for buffer API */

_bool __API
ioFlushBuffer
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	struct _io_buffer __F* buffer = io->buffer;

	if( buffer ) {
		/* as a rule following test was previously processed by the caller */
		if( buffer->lastop==IO_FLAG_WRITE ) {
			register int wend = buffer->wend;
			if( wend ) {
				/* output must not be directed to intermediate buffer */
				if( io->write(io,buffer->buf,wend)!=wend )
					return 0;

				wend = buffer->current-wend;
				if( wend!=0 && io->seek(io,wend,NULL,IO_SEEK_CURRENT)<0 )
					return 0;

				/* all written buffer was flushed */
				buffer->wend = buffer->current = 0;
			}
		}
	}

	return 1;
}

__MODULEEND__



__MODULEBEGIN__( "iosync.c" )
/* ioSyncBuffer
 2000.07.10

	 To be escaped from the buffered sequence of I/O routines, after
 calling some buffered I/O methods in Coral, ioSyncbuffer need to be
 called. */

_bool __API
ioSyncBuffer
	P_B(( io ))
	P_1		struct _IO __F* io
	P_E
{
	int offset;
	struct _io_buffer __F* buffer = io->buffer;

	if( buffer ) {
		switch( buffer->lastop ) {

			case IO_FLAG_WRITE :
			/* it is equivalent to ioFlushBuffer for write mode */
			return ioFlushBuffer( io );

			case IO_FLAG_READ :

			/* following routines are moved from io_ready_wmode, 2000.07.11 */
			offset = buffer->current-buffer->rend;

			if( offset!=0 ) {
				if( !io->seek || io->seek(io,offset,NULL,IO_SEEK_CURRENT)==FAIL_IO )
					return 0;
			}

			/* actually resetting buffer is not needed if buffer->rend==0 */
			io_reset_buffer( buffer );
		} /* end switch( buffer->lastoop ) */
	} /* end if( buffer ) */

	return 1;
}

__MODULEEND__



__MODULEBEGIN__( "iotrunc.c" )
/* ioTruncateBuffer */

/*??? maybe lost in accident */

__MODULEEND__



__MODULEBEGIN__( "iolock.c" )
/* ioLockBuffer

	&locked, 0 : lock available size
	NULL, size : lock requested size
	&locked, size : lock at least the given size (minimum)

 See flags for this method defined in io.h to know the usage.
	This method allocates I/O buffer if it is not present, so
 deallocation request is implicitly exposed if I/O buffer
 was not allocated explicitly.
	Seek procedure must be needed to support read & write mode. See
 io_ready_wmode. */

_integer1 __F* __API
ioLockBuffer
	P_B(( io, locked, size, flag ))
	P_1		struct _IO __F* io
	P_N		int __F* locked
	P_N		int size
	P_N		unsigned flag
	P_E
{
	int remained, requested;
	struct _io_buffer __F* buffer = io->buffer;

	/* attach default buffer if there is no buffer */
	if( !buffer ) {
		if( !ioAllocBuffer( io, !locked?  -size : 0 ) )
			return NULL;
		buffer = io->buffer;
	}

	if( locked ) { /* available or minimum mode */
		/* lock available size */
		remained = ( flag&IO_FLAG_WRITE? io_ready_wmode:io_ready_rmode ) (io,0);

		/* check failure was occurred
			there was a trial to modify remained==FAIL_IO to remained<=0
			not to include zero byte remained case but when the return
			value of ioLockBuffer is NULL, that means an error was occurred.
			io loop can be terminated by testing available locked bytes */
		if( remained==FAIL_IO ) return NULL;

		/* if minimum locked size was specified and remained bytes are
			less than it : 'size>0' is not needed to detect minimum locked size mode
			because if size<=0, then remained<size cannot be true */
		if( remained<size ) {
			DEBUG_PRINTF((
				"ioLockBuffer:%d remained, re-lock buffer to get at least the given size %d\n",
				remained, size ));

			/* the following old code has a bug which cannot consider stream is
				ended before reading the requested size when the refilling is occurred
				in the requested mode.
				return ioLockBuffer
				 ( io, *(int __F*)size = minimum>buffer->size?minimum:buffer->size,
					IO_FLAG_REQUESTED|flag ); */

			requested = *locked = size>buffer->size?size:buffer->size;
			goto l_lock_requested;
		}

		*locked = remained;

	}
	else { /* requested mode */
		requested = size;

		l_lock_requested:

		/* requested size is needed, check buffer doesn't have potential
			problem - if above default buffer allocation was performed,
			buffer->size will be always greater than requested size */
		if( buffer->size<requested ) {
			if( !ioAllocBuffer( io, requested ) ) return NULL;
			buffer = io->buffer;
		}

		if( flag&IO_FLAG_WRITE ) { /* writing mode */
			remained = io_ready_wmode( io, 0 );
			/* if remained buffer is too short, flush it and need not to check
				buffer->lastop==IO_FLAG_WRITE because it is evident : above
				allocation routine prevents potential problem in buffer caused by
				its size so insufficient buffer evaluated after calling io_ready_wmode
				means there is some written bytes in the buffer. */
			if( remained<requested )
				ioFlushBuffer( io );
		}
		else { /* reading mode */
			remained = io_ready_rmode( io, 0 );
			if( remained<requested ) {
				int fillsize;

				/* move remained bytes to the beginning of the buffer,
					overlapped region need not to be considered because
					following operation will be occurred from high address to
					low address */
				mem_short_copy( buffer->buf, buffer->buf+buffer->current, remained );

				/* fill remained buffer, actually requested-remained need to
					be filled but current implementation maintains maximum cache
					for next operations ( more comments can be found in @cmt0004 */
				fillsize = io->read( io, &buffer->buf[remained], buffer->size-remained );

				if( fillsize==FAIL_IO ) return NULL;

				if( (fillsize+=remained)<requested ) {
					/* if( locked ) means the routine was jumped from minimum mode */
					if( locked && fillsize>=size )
						*locked = fillsize;
					else
						return NULL;
				}

				/* reset buffer */
				buffer->rend = fillsize;
				buffer->current = 0;
			}
		} /* end if( flag&IO_FLAG_WRITE ) */

		/* check failure was occurred */
		if( remained==FAIL_IO ) return NULL;

	} /* end if( flag&IO_FLAG_AVAILABLE ) */

	/* lock current buffer position */
	return buffer->buf + buffer->current;
}



/* ioUnlockBuffer */

void __API
ioUnlockBuffer
	P_B(( io, current ))
	P_1		struct _IO __F* io
	P_N		_integer1 __F* current
	P_E
{
	struct _io_buffer __F* buffer = io->buffer;
	int cp = (int)( current - buffer->buf );
	buffer->current = cp;
	if( buffer->lastop==IO_FLAG_WRITE && cp>buffer->wend )
		buffer->wend = cp;
}


/* ioUnlockRewindBuffer

	This function is only applicable for write mode lock */

void __API
ioUnlockRewindBuffer
	P_B(( io, current, int offset ))
	P_1		struct _IO __F* io
	P_N		_integer1 __F* current
	P_N		int offset
	P_E
{
	int cp;
	struct _io_buffer __F* buffer = io->buffer;
	buffer->current += offset;
	cp = (int)( current - buffer->buf );
	if( cp>buffer->wend )
		buffer->wend = cp;
}

__MODULEEND__


__MODULEBEGIN__( "ioinitb.c" )
/* ioInitializeBuffer */

void __API
ioInitializeBuffer
 ( struct _IOBUFFER __F* iob, struct _IO __F* io )
{
	iob->context = (ANYCONTEXT)io;
	iob->lock = (_IOLOCKBUFFER)ioLockBuffer;
	iob->unlock = (_IOUNLOCKBUFFER)ioUnlockBuffer;
}
__MODULEEND__



__MODULEBEGIN__( "iolockm.c" )
/* ioLockMemory
 Simple memory locking mechanism to use memory buffer directly */

_integer1 __F* __API
ioLockMemory
	P_B(( iom, int __F* locked, size, flag ))
	P_1		struct _IOLOCKMEMORY __F* iom
	P_N		int __F* locked
	P_N		int size
	P_N		unsigned flag
	P_E
{
	int current = iom->current;
	int remained = iom->size - current;

	/* the case 'remained==0' is not excluded and if *size==0 for IO_FLAG_AVAILABLE mode,
	 it means eof was reached */
	if( remained>=0 ) {
		void __F* p = &iom->p[current];

		if( size>0 ) { /* requested or minimum lock mode */
			if( locked ) *locked = remained; /* minimum mode */
			if( remained<size ) return NULL;
		}
		else { /* available mode */
			*locked = remained;
		}

		return p;
	}

	return NULL;
}

__MODULEEND__



__MODULEBEGIN__( "ioulockm.c" )
/* ioUnlockMemory */

void __API
ioUnlockMemory
	P_B(( iom, current ))
	P_1		struct _IOLOCKMEMORY __F* iom
	P_N		_integer1 __F* current
	P_E
{
	/* move cp */
	iom->current = (int)( current - iom->p );
}

__MODULEEND__



__MODULEBEGIN__( "ioinitm.c" )
/* ioInitializeMemoryBuffer */

void __API
ioInitializeMemoryBuffer
	P_B(( iob, iom, buf, bufsize ))
	P_1		struct _IOBUFFER __F* iob
	P_N		struct _IOLOCKMEMORY __F* iom
	P_N		_integer1 __F* buf
	P_N		int bufsize
	P_E
{
	iom->size = bufsize<=0? strlen(buf) : bufsize;
	iom->current = 0;
	iom->p = buf;

	iob->context = (ANYCONTEXT)iom;
	iob->lock = (_IOLOCKBUFFER)ioLockMemory;
	iob->unlock = (_IOUNLOCKBUFFER)ioUnlockMemory;
}

__MODULEEND__



__MODULEBEGIN__( "iolockc.c" )



/* ioLockConsole

	For 'write' function, the given buffer should be null terminated and
 the null terminated position is equal to 'bufsize' argument. */

_integer1 __F* __API
ioLockConsole
	P_B(( ioc, locked, size, lockflag ))
	P_1		struct _IOLOCKCONSOLE __F* ioc
	P_N		int __F* locked
	P_N		int size
	P_N		unsigned lockflag
	P_E
{
	int bufpos, buflen;
	_integer1 __F* buf;

	if( lockflag&IO_FLAG_WRITE ) return NULL; /* XXX write lock is unsupported now */
	if( size>0 ) return NULL; /* XXX explicit size is unsupported now */

	buf = ioc->buf;
	bufpos = ioc->inbufpos;
	buflen = ioc->inbuflen;

	if( bufpos<buflen ) {
		*locked = buflen - bufpos;
	}
	else {
		int state = ioc->state;
		if( state!=IO_END && bufpos==0 ) {
			if( (ioc->flag&IO_FLAG_PROMPT) && ioc->prompt ) ioc->prompt( ioc );
			if( ioc->read( ioc, buf, ioc->inbufsize )==0 ) {
				ioc->state = IO_NORMAL;
				ioc->inbuflen = *locked = strlen(buf);
			}
			else {
				ioc->state = IO_END;
				ioc->inbuflen = *locked = 0;
			}
		}
		else {
			ioc->state = IO_BREAK;
			ioc->inbuflen = *locked = 0;
			*locked = 0;
		}

		bufpos = ioc->inbufpos = 0;
	}
	return buf + bufpos;
}



/* ioUnlockConsole */

void __API
ioUnlockConsole
	P_B(( ioc, current ))
	P_1		struct _IOLOCKCONSOLE* ioc
	P_N		_integer1 __F* current
	P_E
{
	ioc->inbufpos = current - ioc->buf;
}



/* ioFlushConsole */

void __API
ioFlushConsole
	P_B(( ioc ))
	P_1		struct _IOLOCKCONSOLE __F* ioc
	P_E
{
	ioc->inbufpos = ioc->inbuflen = 0;
}
__MODULEEND__



__MODULEBEGIN__( "ioinicb.c" )
struct _IOLOCKCONSOLE __F* __API
ioInitializeConsoleBuffer
	P_B(( ioc, context, read, write, inbufsize, outbufsize, flag ))
	P_1		struct _IOLOCKCONSOLE __F* ioc
	P_N		void __F* context
	P_N		int (__API* prompt) P_A(( struct _IOLOCKCONSOLE __F* ioc ))
	P_N		int (__API* read) P_A(( struct _IOLOCKCONSOLE __F* ioc, _integer1 __F* buf, int bufsize ))
	P_N		int (__API* write) P_A(( struct _IOLOCKCONSOLE __F* ioc, _integer1 __F* buf, int bufsize ))
	P_N		int inbufsize
	P_N		int outbufsize
	P_N		unsigned flag
	P_E
{
	if( ioc==NULL ) {
		ioc = (struct _IOLOCKCONSOLE __F*)memAlloc
			( sizeof(struct _IOLOCKCONSOLE) + inbufsize + outbufsize + (2-1) );
		if( ioc==NULL ) return NULL;
	}

	ioc->context = context;
	ioc->prompt = prompt;
	ioc->read = read;
	ioc->write = write;
	ioc->flag = flag;
	ioc->lockflag = 0;
	ioc->state = IO_NORMAL;
	ioc->inbufsize = inbufsize;
	ioc->inbufpos = 0;
	ioc->inbuflen = 0;
	ioc->outbufsize = outbufsize;
	ioc->outbufpos = 0;

	return ioc;
}
__MODULEEND__



__MODULEBEGIN__( "iogetcst.c" )
/* ioGetConsoleState */

int __API
ioGetConsoleState( struct _IOLOCKCONSOLE __F* ioc )
{
	return ioc->state;
}
__MODULEEND__



__MODULEBEGIN__( "iosetcst.c" )
/* ioResetConsoleState */

void __API
ioResetConsoleState( struct _IOLOCKCONSOLE __F* ioc )
{
	ioc->state = IO_NORMAL;
}
__MODULEEND__



__MODULEBEGIN__( "ioread.c" )
/* ioRead */

long __API
ioRead
	P_B(( io, p, n, byteorder ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long n
	P_N		unsigned byteorder
	P_E
{
	if( n==0 ) return 0;

	if( io->read && io->handle ) {
		struct _io_buffer __F* buffer;
		int remained=0;
		void __F* pold = p;

		/* get total size */
		unsigned size = byteorder&MASK_COMBINATED_SIZE;
		if( size>1 )
			n *= (long)size;

		/* check buffer was allocated */
		buffer = io->buffer;
		if( buffer ) {
			/* it was considered to use ioLockBuffer but it cannot consider the
				case, remained byte is zero but requested size(n) exceeds buffer
				size : n+1 means cache is not needed if the requested size is
				equal to the buffer size */
			remained = io_ready_rmode( io, n+1 );
			if( remained==FAIL_IO ) return FAIL_IO;

			/* if 'remained' is zero, buffer size exceeds requested size 'n'
				: see io_ready_rmode */
			if( remained ) {
				/* transfer bytes from buffer */
				register _unsigned1 __F* buf = buffer->buf + buffer->current;
				int transfer = (long)remained>=n ? (int)n : remained;


				buffer->current += (int)transfer;

				switch( transfer ) {
					case 4 : ((_unsigned1 __F*)p)[3] = buf[3];
					case 3 : ((_unsigned1 __F*)p)[2] = buf[2];
					case 2 : ((_unsigned1 __F*)p)[1] = buf[1];
					case 1 : ((_unsigned1 __F*)p)[0] = buf[0]; break;
					default : mem_short_copy( p, buf, transfer );
				}

				/* remained bytes are sufficient to finish current opeartion ? */
				if( (long)remained>=n ) {
					DEBUG_PRINTF((
						"ioRead:remained %d bytes could dispatch %d bytes request\n",
						remained, n ));

					goto byte_order;
				}

				/* compensate cached bytes */
				n -= remained;
				p = (_integer1 __F*)p + remained;

				/* @CMT0009, bug fixed */
				buffer->rend = buffer->current = 0;

				DEBUG_PRINTF((
					"ioRead:%d bytes in buffer were used and %d bytes will be read\n",
					remained, n ));
			} /* end if(remained) */

		} /* end if( buffer ) */

		/* now file opeartion is occurred physically */
		n = io->read( io, p, n );
		if( n!=FAIL_IO ) {
			n += remained;

			/* manage byte order */
			byte_order:
			if( size>1 ) {
				n /= size;
				memReorder( pold, n, byteorder );
			}

			return n;
		} /* if( n!=FAIL_IO ) */

	} /* end if( io->read && n>0 ) */

	return FAIL_IO;
}
__MODULEEND__



__MODULEBEGIN__( "iowrite.c" )
/* ioWrite */

long __API
ioWrite
	P_B(( io, p, n, byteorder ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long n
	P_N		unsigned byteorder
	P_E
{
	if( n==0 ) return 0;
	if( io->write && io->handle ) {
		_bool reordered;
		struct _io_buffer __F* buffer;

		/* byte ordering */
		unsigned size = byteorder&MASK_COMBINATED_SIZE; /* elementary size */
		if( size>1 ) {
			/* contents in 'p' will be recovered so 'const' will be maintained */
			reordered = memReorder( (void __F*)p, n, byteorder );
			n *= size;
		}
		else reordered = 0;

		/* check a buffer was attached */
		buffer = io->buffer;
		if( buffer ) {
			/* read wmode, maximum size is not needed : see the comments in
				io_ready_rmode and ioRead to know the reason to use 'n+1' */
			int remained = io_ready_wmode( io, n+1 );
			if( remained==FAIL_IO )
				return FAIL_IO;

			/* remained bytes are sufficient to complete current operation ? */
			if( (long)remained>=n ) {
				int cp;
				mem_short_copy( buffer->buf+buffer->current, p, (int)n );
				cp = buffer->current+=(int)n;
				if( cp>buffer->wend )
					buffer->wend = cp;
				goto l_byte_order;
			}

			/* transfer cached bytes */
			if( buffer->current ) {
				if( io->write( io, buffer->buf, buffer->current )==FAIL_IO )
					return FAIL_IO;
				buffer->current = 0;
				buffer->wend = 0;
			}
		} /*end if( buffer ) */

		n = io->write( io, p, n );
		if( n!=FAIL_IO ) {
			/* manage byte order */
			l_byte_order:
			if( size>1 ) {
				n /= size;
				if( !(byteorder&FLAG_TEMPORARY) && reordered )
					memInvertB( (void __F*)p, n, size );
			}
			return n;
		} /* end if( n!=FAIL_IO ) */

	} /* end if( io->write && n>0 ) */

	return FAIL_IO;
}
__MODULEEND__


#define _NO_RETURN

__MODULEBEGIN__( "ioseek.c" )
/* ioSeek */

long
io_seek_backward
	P_B(( io, diff, offset_high ))
	P_1		struct _IO __F* io
	P_N		int diff
	P_N		long __F* offset_high
	P_E
{
	struct _io_buffer __F* buffer = io->buffer;
	_integer1 __F* buf = buffer->buf;
	int rend = buffer->rend;
	int unused = buffer->size - rend;
	int reserved = rend - diff;

	if( unused>diff )
		unused = diff;

	/*  offset----------- current      rend
	              |----------|-----------|
	      |_______|
	        diff                          */

	memmove( buf+diff, buf, reserved+unused );
	if( io->seek( io, -rend-diff, NULL, IO_SEEK_CURRENT )==FAIL_IO ||
	    io->read( io, buf, diff )!=diff )
		return FAIL_IO;
	buffer->current = 0;
	return io->seek( io, reserved, offset_high, IO_SEEK_CURRENT );
}

int
io_seek_forward
	P_B(( io, diff, size ))
	P_1		struct _IO __F* io
	P_N		int diff
	P_N		int size
	P_E
{
	/*         -------------------------offset
	     -----|--------------|
	        current         rend
	                         |____________|
	                             diff       */
	struct _io_buffer __F* buffer = io->buffer;
	_integer1 __F* buf = buffer->buf;
	register int reserved = buffer->reserved;

	memmove( buf, buf+size, reserved );
	size = io->read( io, buf+reserved, size );
	if( size==FAIL_IO ) return FAIL_IO;
	buffer->rend = size + reserved;
	buffer->current = diff + reserved;
	return 0;
}

long __API
ioSeek
	P_B(( io, offset, offset_high, seekpos ))
	P_1		struct _IO __F* io
	P_N		long offset
	P_N		long __F* offset_high
	P_N		int seekpos
	P_E
{
	if( io->seek && io->handle ) {
		struct _io_buffer __F* buffer = io->buffer;

		if( buffer ) {
			if( offset_high==NULL ||
			    ( *offset_high==0 &&
			    (unsigned long)offset<=OP_MAXINT(long) ) ) { /* long integer
					boundary */
				long diff;
				register int reserved, rsize, current, remained, rend;
				current = buffer->current;
				rend = buffer->rend;
				remained = rend - current;

				/* consideration for relative seek operation */
				if( seekpos==IO_SEEK_CURRENT ) {

					/* just detect current position */
					if( offset==0 ) {
						/* reviewed, 1998. 3. 31
							old code was
							seekpos + ( flag&FILE_FLAG_READ ? current-iof->read : current );
							but current-iof->read is sufficient because iof->read is zero if
							previous I/O operation was write - read buffer is cleared by write
							operation. */
						return io->seek(io,0,NULL,IO_SEEK_CURRENT) - remained;
					}

					if( buffer->lastop==IO_FLAG_READ ) {

						if( offset<0 ) {
							if( current>=-offset ) {
								buffer->current = current + (int)offset;
								return 0; /* XXX offset */
							}

#if 1
							diff = -offset - current;
							if( diff<rend )
								return io_seek_backward( io, diff, NULL );
#endif
						}
						else { /* offset > 0 */
							if( remained>offset ) {
								buffer->current = current + offset;
/* XXX offset
							return io->seek(io,0,NULL,IO_SEEK_CURRENT) +
								( buffer->current = current + offset )
								- buffer->rend;*/
								return 0;
							}


#if 1
							/* this optimization is meaningful when eof was
							   not reached */
							reserved = buffer->reserved;
							if( reserved>0 && rend==buffer->size ) {
								diff = offset - remained;
								rsize = rend - reserved;
								if( diff<rsize ) {
									/* XXX cur offset is needed ? */
									return io_seek_forward( io, diff, rsize );
								}
							}
#endif
						}

						offset -= remained;
					}
					else { /* write mode */
						diff = buffer->current + offset;
						if( diff>=0 && diff<=buffer->wend ) {
							buffer->current = diff;
#ifdef _NO_RETURN
							return 0;
#else
							return io->seek(io,0,NULL,IO_SEEK_CURRENT) + diff;
#endif
						}
					}
				}
				else { /* seek begin or end */
					if( buffer->lastop==IO_FLAG_READ ) {
						if( seekpos==IO_SEEK_BEGIN ) {
							diff = offset -
								io->seek(io,0,NULL,IO_SEEK_CURRENT) + remained;
/*							if( diff<0?
									(current>=-diff) :
									(remained>diff) ) {
								buffer->current = current + diff;
								return offset;
							}*/
							if( diff<0 ) {
								if( current>=-diff ) {
									buffer->current = current + diff;
									return offset;
								}
								diff = -diff - current;
#if 1
								if( diff<rend )
									return io_seek_backward( io, diff, NULL );
#endif
							}
							else {
								if( remained>diff ) {
									buffer->current = current + diff;
									return offset;
								}

#if 1
								reserved = buffer->reserved;
								if( reserved>0 && rend==buffer->size ) {
									diff = diff - remained;
									rsize = rend - buffer->reserved;
									if( diff<rsize ) {
										/* XXX cur offset is needed ? */
										if( io_seek_forward( io, diff, rsize )<0 )
											return FAIL_IO;
										return offset;
									}
								}
#endif
							}

						} /* seek begin */
					}
					else { /* write mode */
						if( seekpos==IO_SEEK_BEGIN ) {
							diff = offset - io->seek(io,0,NULL,IO_SEEK_CURRENT);
							if( diff>=0 && diff<=buffer->wend ) {
								buffer->current = (int)diff;
								return offset;
							}
						}

						else if( seekpos==IO_SEEK_END && offset==0 ) {

long end;
							diff = io->seek(io,0,NULL,IO_SEEK_CURRENT);
							end = io->seek(io,0,NULL,IO_SEEK_END);
							if( end==diff ) {
								return diff + buffer->current;
							}
							else {
								/* recover the current offset */
								io->seek(io,diff,NULL,IO_SEEK_BEGIN);
							}
						}
					} /* end if( buffer->lastop==IO_FLAG_READ ) */
				} /* end if( seekpos==IO_SEEK_CURRENT ) */


			}
			else { /* large io XXX test needed */

				if( seekpos==IO_SEEK_CURRENT ) {
					_integerx offsetx;
					register int current = buffer->current;

					x_store( offsetx, x_const(*offset_high,offset) );

					if( x_zero(offsetx) ) {
						long hoffset=0, loffset=io->seek(io,0,&hoffset,IO_SEEK_CURRENT);
						if( hoffset<0 ) goto l_error;
						x_store( offsetx, x_add( x_const(hoffset,loffset), x_short(current-buffer->rend) ) );
						*offset_high = x_high( offsetx );
						return x_low( offsetx );
					}

					if( buffer->lastop==IO_FLAG_READ )
						x_subs( offsetx, buffer->rend-current );
/*XXX offset ? */
					offset = x_low(offsetx);
					*offset_high = x_high(offsetx);
				}
			} /* end if( not large io ) */

			/* last operation was writing ( bug fixed, 1998.02 ),
				flush operation will make io position==buffer->current */
			if( buffer->lastop==IO_FLAG_WRITE )
				ioFlushBuffer( io );
			else
				buffer->rend = buffer->current = 0;

		} /* end if( buffer ) */

		return io->seek( io, offset, offset_high, seekpos );
	}

	l_error:
	if( offset_high ) *offset_high = FAIL_IO;
	return FAIL_IO;
}
__MODULEEND__



__MODULEBEGIN__( "ioffstrt.c" )
/* ioOffsetStart */

int __API
ioOffsetStart
	P_B(( iooffset ))
	P_1		struct _IOOFFSET __F* iooffset
	P_E
{
	/* start offset */
	long offset = x_low( iooffset->start_offset );
	long offset_high = x_high(iooffset->start_offset);
	if( ioSeek( &iooffset->io, offset, &offset_high,
			IO_SEEK_BEGIN )==FAIL_IO && offset_high==FAIL_IO )
		return FAIL_IO;

	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "ioffread.c" )
/* ioOffsetRead */

long __API
ioOffsetRead
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long size
	P_E
{
	long result = FAIL_IO;
	struct _IOOFFSET __F* iooffset = (struct _IOOFFSET __F*)io->handle;

	if( x_ge( iooffset->offset, x_const(0,0) ) ) {
		_integer8 remained;

		x_store( remained, x_sub( iooffset->size, iooffset->offset ) );
		if( x_gt( x_short(size), remained ) )
			size = x_conv(long,remained);

		if( size>0 ) {
			io->handle = iooffset->io.handle;
			result = iooffset->io.read( io, p, size );
			io->handle = iooffset;

			if( result>0 )
				x_adds( iooffset->offset, x_short(result) );
		}
		else {
			result = 0; /* eof */
		}
	}

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "ioffwrit.c" )
/* ioOffsetWrite */

long __API
ioOffsetWrite
	P_B(( io, p, size ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long size
	P_E
{
	long result = FAIL_IO;
	struct _IOOFFSET __F* iooffset = (struct _IOOFFSET __F*)io->handle;

	if( x_ge( iooffset->offset, x_const(0,0) ) ) {
		_integer8 remained;
		x_store( remained, x_sub( iooffset->limit, iooffset->offset ) );
		if( x_gt( x_short(size), remained ) )
			size = x_conv(long,remained);

		io->handle = iooffset->io.handle;
		result = iooffset->io.write( io, p, size );
		io->handle = iooffset;

		if( result>0 ) {
			x_adds( iooffset->offset, x_short(result) );
			if( x_gt( iooffset->offset, iooffset->size ) ) {
				x_store( iooffset->size, iooffset->offset );
			}
		}
	}

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "ioffseek.c" )
/* ioOffsetSeek */

long __API
ioOffsetSeek
	P_B(( io, offset, offset_high, pos ))
	P_1		struct _IO __F* io
	P_N		long offset
	P_N		long __F* p_offset_high
	P_N		int pos
	P_E
{
	long offset_high, offset_low;
	_integer8 newoffset, curoffset;
	struct _IOOFFSET __F* iooffset = (struct _IOOFFSET __F*)io->handle;

	if( pos==IO_SEEK_CURRENT ) {

		if( offset==0 ) {
			if( p_offset_high==NULL )
				return x_low(iooffset->offset);
			if( *p_offset_high==0 ) {
				*p_offset_high = x_high(iooffset->offset);
				return x_low(iooffset->offset);
			}
		}

		io->handle = iooffset->io.handle;
		offset = iooffset->io.seek
			( io, offset, p_offset_high, IO_SEEK_CURRENT );
		io->handle = iooffset;

		if( offset==FAIL_IO &&
			( p_offset_high==NULL || *p_offset_high==FAIL_IO ) ) return FAIL_IO;

		x_store( newoffset, x_sub( p_offset_high?
			x_const(*p_offset_high,offset):x_short(offset),
			iooffset->start_offset ) );

		x_store( iooffset->offset, newoffset );


		if( p_offset_high ) *p_offset_high = x_high(newoffset);
		return x_low(newoffset);
	}

	if( pos==IO_SEEK_END )
		x_store( curoffset, x_add( iooffset->size, p_offset_high?
			x_const(*p_offset_high,offset) : x_short(offset) ) );
	else
		x_store( curoffset, p_offset_high?
			x_const(*p_offset_high,offset) : x_short(offset) );

	x_store( newoffset, x_add( iooffset->start_offset, curoffset ) );

	offset_high = x_high(newoffset);

	io->handle = iooffset->io.handle;
	offset_low = iooffset->io.seek
		( io, x_low(newoffset), &offset_high, IO_SEEK_BEGIN );
	io->handle = iooffset;

	if( offset_low==FAIL_IO && offset_high==FAIL_IO ) {
		if( p_offset_high ) *p_offset_high = FAIL_IO;
		return FAIL_IO;
	}

	x_store( iooffset->offset, curoffset );

	if( pos==IO_SEEK_END ) {
		if( p_offset_high ) *p_offset_high = offset_high;
		return x_low(curoffset);
	}

	return offset;
}
__MODULEEND__



__MODULEBEGIN__( "ioskip.c" )
/* ioSkip

 added on 1999.07.12 */

_bool __API
ioSkip
	P_B(( io, size ))
	P_1		struct _IO __F* io
	P_N		long size
	P_E
{
	if( size>0 ) {
		/* if seek procedure was supplied, just use it */
		if( io->seek ) {
			if( ioSeek(io,size,NULL,IO_SEEK_CURRENT)==FAIL_IO )
				return 0;
		}
		else {
			_unsigned1 buf[256];

			/* reduce requested size by iteration */
			while( size>sizeof(buf) ) {
				if( !ioFullRead(io,buf,sizeof(buf),0) ) return 0;
				size-=sizeof(buf);
			}

			/* dispatch spares : notice that 'size' cannot be negative
				value because subtraction in the above loop can be performed
				only when size>sizeof(buf) */
			if( !ioFullRead(io,buf,size,0) ) return 0;
		} /* end if( io->seek ) */
	} /* end if( size>0 ) */
	else if( size<0 ) {
		return 0;
	}

 return 1;
}
__MODULEEND__



/* ioSkipBuffer */
__MODULEBEGIN__( "ioskipb.c" )

_bool __API
ioSkipBuffer
	P_B(( iob, size ))
	P_1		struct _IOBUFFER __F* iob
	P_N		int size
	P_E
{
	while( 1 ) {
		int avail;
		_unsigned1 __F* p = (_unsigned1 __F*) iob->lock( iob->context, &avail, 0, IO_FLAG_READ );
		if( !p ) return 0;

		iob->unlock( iob->context, p + (avail>=size?size:avail) );
		size-=avail;
		if( size<=0 ) break;
	}

	return 1;
}

__MODULEEND__



__MODULEBEGIN__( "iofread.c" )
/* ioFullRead */

_bool __API
ioFullRead
	P_B(( io, p, n, byteorder ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long n
	P_N		unsigned byteorder
	P_E
{
	return ioRead(io,p,n,byteorder)==n;
}
__MODULEEND__



__MODULEBEGIN__( "iofwrite.c" )
/* ioFullWrite */

_bool __API
ioFullWrite
	P_B(( io, p, n, byteorder ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long n
	P_N		unsigned byteorder
	P_E
{
	return ioWrite(io,p,n,byteorder)==n;
}
__MODULEEND__



__MODULEBEGIN__( "iowint.c" )
/* ioWriteInt
	I think this method can be useful if a default integer is used in
 the program but it need to be explicitly sized when it is stored.
	If source size was not specified, 4 byte is assumed.
 Added on 2000.06.05 */

_bool __API
ioWriteInt
	P_B(( io, value, flag ))
	P_1		struct _IO __F* io
	P_N		long value
	P_N		unsigned flag
	P_E
{
	memReduceInt( &value, flag&MASK_COMBINATED_SIZE );

	return ioFullWrite( io, &value, 1, flag );
}
__MODULEEND__



__MODULEBEGIN__( "iordrec.c" )
/* ioReadRecord */

_bool __API
ioReadRecord
	P_B(( io, p, ri, flag ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		const void __F* ri
	P_N		unsigned flag
	P_E
{
	return io_operate_record( io, p, ri, flag, ioFullRead );
}
__MODULEEND__



__MODULEBEGIN__( "iowrrec.c" )
/* ioWriteRecord */

_bool __API
ioWriteRecord
	P_B(( io, p, ci, flag ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		const void __F* ri
	P_N		unsigned flag
	P_E
{
	return io_operate_record( io, p, ri, flag, ioFullWrite );
}
__MODULEEND__



__MODULEBEGIN__( "iord2d.c" )
/* ioRead2D

	Parameters for this method is compatible to 2 dimensional memory
 methods but byte ordering must be considered for I/O operations
 and 'width' parameters of 2 dimensional memory methods is separated
 into two parts, 'item' and 'unit size' supplied by 'byteorder'. */

_bool __API
ioRead2D
	P_B(( io, p, row, item, colrow, colcol, variable, byteorder ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long row
	P_N		unsigned short item
	P_N		const void __F* colrow
	P_N		const void __F* colcol
	P_N		_unsigned2 variable
	P_N		_unsigned2 byteorder
	P_E
{
	return io_operate_2d
	 ( io, p, row, item, colrow, colcol, variable, byteorder, ioFullRead );
}
__MODULEEND__



__MODULEBEGIN__( "iowr2d.c" )
/* ioWrite2D

 Read comments in ioRead2D */

_bool __API
ioWrite2D
	P_B(( io, p, row, item, colrow, colcol, variable, byteorder ))
	P_1		struct _IO __F* io
	P_N		void __F* p
	P_N		long row
	P_N		unsigned short item
	P_N		const void __F* colrow
	P_N		const void __F* colcol
	P_N		_unsigned2 variable
	P_N		_unsigned2 byteorder
	P_E
{
	return io_operate_2d
	 ( io, p, row, item, colrow, colcol, variable, byteorder,
		 (_bool (__APICAST*) P_A((struct _IO __F*,void __F*,long,unsigned))) ioFullWrite );
}
__MODULEEND__



/* EOF */

