/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Text I/O operations */
/*
 1997.10
 1999.06.05
 1999.11.03  Revision using I/O buffer API
 2000.07.26  Reviewed for PARAM spec to support classic param style
 2001.03.14  The definition of I/O buffer
*/

#undef __DEBUG__ /* ignore globally defined debug request */
/* #define __DEBUG__ */



#include <coralsrc/io.h>
#include <coralsrc/memory.h>
#include <coralsrc/string.h>
#include <coral/inline.h>
#include <coral/compat/ctype.h>
#include <coral/compat/string.h>




/*------------------------------------------------------------------------*
   Internal functions
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "iotdat.c" )
struct _IOTEXT  io_text_default;
__MODULEEND__



/* iotext_skip_lineend

	This function is ( should be ) called after p is increased at least by 1
	*p must be '\r' or '\n' and at least, this function will return p+1.

	This function can modify the given size, so the size after calling
	this function should be carefully evaluated. */

_integer1 __F*
iotext_skip_lineend
	P_B(( iob, p, size ))
	P_1		struct _IOBUFFER __F* iob
	P_N		_integer1 __F* p
	P_N		int __F* size
	P_E
{
	register char c1, c2;
	int i;

	/* evidently c1 should be line end character ( it's a rule for this
	  function )
	   this function is ( should be ) called after p is increased at least
	  by 1 */
	c1 = *(p-1);
	i = *size;

	if( i<=0 ) {
		iob->unlock( iob->context, p );
		p = (_integer1 __F*)iob->lock( iob->context, &i, 0, IO_FLAG_READ );
		if( !p ) return NULL;
		if( i==0 ) return p; /* go on further buffered io operations */
	}

	c2=*p;

	if( ( c2!='\r' && c2!='\n' ) || c1==c2 ) {
		/* the current character is not a line end character or
			 \n\n or \r\r means two lines */
		*size = i;
		return p;
	}

	/* the current character is a line end character and the same
	  line end character was not appeared ( \r\n or \n\r is allowed ) */

	*size = i-1;
	return p+1;
}



/*------------------------------------------------------------------------*
   API
 *------------------------------------------------------------------------*/



/* iotext_probe_string

  The first character must be tested before calling this function */

__MODULEBEGIN__( "iotprbs.c" )

int
iotext_probe_string
	P_B(( iob, iot, locked, remained, s, len ))
	P_1		struct _IOBUFFER __F* iob
	P_N		struct _IOTEXT __F* iot
	P_N		_integer1 __F*__F* locked
	P_N		int __F* remained
	P_N		const char __F* s
	P_N		int len
	P_E
{
	_integer1 __F* p=*locked;
	if( len>1 ) {
		int size=*remained;
		if( size<len ) { /* remained locked chars are shorter than
							the given string */
			iob->unlock( iob->context, p );
			p = iob->lock( iob->context, &size, len, IO_FLAG_READ );
			/*  at least one byte is remained because the first character
			   should be tested before calling this function, so though p is
			   NULL now, the following lock operation will be evidently
			   successful : returning error code will cause more difficult
			   exception handling in caller side */
			if( !p ) p = iob->lock( iob->context, &size, 0, IO_FLAG_READ );
			*locked = p;
			*remained = size;

			/* insufficient locked chars so it is not a comment */
			if( size<len ) return FAIL_UNAVAILABLE;
		}

		if( strncmp((const char __F*)p,s,len)!=0 ) return 1;
	}
	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "iotread.c" )
/* iotextRead */

#define _IOTEXT_RESET_COLUMN(col,lastcol) \
	if( lastcol==0 ) lastcol = col; \
	col = 0;

static int
iotext_read
	P_B(( iob, delimiter, delimiter_len, buf, bufsize, flag, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* delimiter
	P_N		int delimiter_len
	P_N		char __F* buf
	P_N		int bufsize
	P_N		unsigned flag
	P_N		struct _IOTEXT __F* iot
	P_E
{
	struct _IOBUFFER iobcache;
	_integer1 __F* p;
	unsigned long line;
	unsigned col, lastcol, eventmask;

	char c, __F* curbuf;
	int i, result=IO_END, len=0;

	OP_COPY( iobcache, *iob );
	curbuf = buf;

	/*  following routine must be done after the above routine
	   because the current line can be modified by it */
	line = iot->line;
	col = iot->col;
	lastcol = 0;
	eventmask = iot->eventmask;

	while( 1 ) {

		/* lock available buffer */
		int size;
		p = iobcache.lock( iobcache.context, &size, 0, IO_FLAG_READ );
		if( !p ) return FAIL_IO;

		/* no locked chars, terminate loop */
		if( size==0 ) break;

		/* parse locked contents */
		while( size>0 ) {
			_bool lineend_found;

			c = *p;

			if( !(flag&IOTEXT_FLAG_DISABLE_COMMENT) ) {
				int lcmt1=iot->llinecmt, lcmt2=1;
				const char __F* cmt1=iot->linecmt, __F* cmt2 = "";
				unsigned eventflag=IOTEXT_FLAG_LINE_COMMENT;
				unsigned charflag = IOTEXT_FLAG_LINE_COMMENT_CHAR;
				unsigned startevent = IOTEXT_BEGIN_LINE_COMMENT;
				i = 0;
				while( 1 ) {
					if( lcmt1 && *cmt1==c &&
						iotext_probe_string
							(iob,iot,&p,&size,cmt1,lcmt1)==0 ) {
						unsigned long cmtline;
						unsigned charevent, eventmask;

						iobcache.unlock( iobcache.context, p+lcmt1 );
						col+=lcmt1;
						eventmask = iot->eventmask;
						if( eventmask&eventflag ) {
							if( iot->event( iob, iot, startevent,
									(void __F*)cmt1 )==IOTEXT_DISPATCH_FAIL )
								return IOTEXT_DISPATCH_FAIL;
							eventmask = iot->eventmask;
						}

						result = IO_END;
						cmtline=0;
						charevent = startevent+2;

						do {

							p = iob->lock
								( iob->context, &size, 0, IO_FLAG_READ );

							if( p==NULL ) {
								result = FAIL_IO;
								break;
							}

							if( size<=0 ) {
								if( !cmt2[0] && lcmt2 ) result = IOTEXT_COMMENT;
								else result = IOTEXT_INVALID_COMMENT;
								iob->unlock( iob->context, p );
								break;
							}

							while( size>0 ) {

								c = *p;
								if( c=='\n' || c=='\r' ) {
									cmtline++;
									_IOTEXT_RESET_COLUMN(col,lastcol);
									size--;
									p = iotext_skip_lineend( iob, p+1, &size );
									if( !p ) {
										result = FAIL_IO;
										break;
									}
									if( !cmt2[0] ) {
										result = IOTEXT_COMMENT;
										break;
									}
									/* XXX parsing for ' # endif' support */
									c = IOTEXT_LINE_END;
								}
								else {

									if( *cmt2==c ) {

										int cmp = iotext_probe_string
											(iob,iot,&p,&size,cmt2,lcmt2);

										if( cmp==FAIL_UNAVAILABLE ) {
											/* insufficient buffer means
												trailing specifier was
												not found */
											result = IOTEXT_INVALID_COMMENT;
											break;
										}

										if( cmp==0 ) {
											DEBUG_PUTS(( "iotext_skip_comment:a comment found" ));
											col+= lcmt2;
											p+=lcmt2;
											result = IOTEXT_COMMENT;
											break;
										}

									} /* end if( *cmt2==c ) */

									size--;
									p++;
									col++;
								} /* end if( line end is found ) */

								if( eventmask&charflag ) {
									if( iot->event(iob,iot,charevent,&c)
											==IOTEXT_DISPATCH_FAIL ) {
										result = IOTEXT_DISPATCH_FAIL;
										break;
									}
									/* update event mask */
									eventmask = iot->eventmask;
								}
							} /* end while( locked chars ) */

							iob->unlock( iob->context, p );

						} while( result==IO_END );

						if( result==IOTEXT_COMMENT && (eventmask&eventflag) ) {
							if( iot->event( iob, iot, startevent+1,
									size<=0? NULL:(void __F*)cmt2 )
								==IOTEXT_DISPATCH_FAIL )
								result = IOTEXT_DISPATCH_FAIL;
						}

						iot->invalidcmt = line;
						line += cmtline;
						goto l_result;
					}
					if( i>0 ) break;
					i++;
					lcmt1 = iot->lblockcmt1;
					cmt1 = iot->blockcmt1;
					lcmt2 = iot->lblockcmt2;
					cmt2 = iot->blockcmt2;
					eventflag = IOTEXT_FLAG_BLOCK_COMMENT;
					charflag = IOTEXT_FLAG_BLOCK_COMMENT_CHAR;
					startevent = IOTEXT_BEGIN_BLOCK_COMMENT;
				}

				lcmt2 = iot->lblockcmt2;
				cmt2 = iot->blockcmt2;
				if( lcmt2 && *cmt2==c &&
				    iotext_probe_string(iob,iot,&p,&size,cmt2,lcmt2)==0 ) {
					iobcache.unlock(iobcache.context,p+lcmt2);
					/* invalid block comment : trailing specifier is found
					   without leading specifier */
					col+=lcmt2;
					iot->invalidcmt = line;
					result = IOTEXT_INVALID_COMMENT;
					goto l_result;
				}
			} /* end if( !(flag&IOTEXT_FLAG_DISABLECOMMENT) ) */

			size--;
			p++;

#if 0
			if( !(flag&IOTEXT_FLAG_PRESERVE_LINE_END) &&
				(c=='\n' || c=='\r') ) {
				/*  io_skip_lineend should be called after finding
				   exclusive mode delimiter because buffer can be regressed
				   in that routine */
				line++;
				c = IOTEXT_LINE_END;
			}
#endif

			if( c=='\n' || c=='\r' ) {
				line++;
				if( !(flag&IOTEXT_FLAG_PRESERVE_LINE_END ) )
					c = IOTEXT_LINE_END;
				lineend_found = 1;
			}
			else {
				lineend_found = 0;
			}

			/* progress locked char, must be here not to be ignored by
				executing unlock */

			if( !( flag&IOTEXT_FLAG_DISABLE_DELIMITER ) &&
			     ( flag&IOTEXT_FLAG_EXCLUSIVE?
			             memchr(delimiter,c,delimiter_len)==NULL :
			             memchr(delimiter,c,delimiter_len)!=NULL ) ) {
				iot->lastchar = c;

				if( flag&IOTEXT_FLAG_EXCLUSIVE ) {
					/* ignore non-delimiter character */
					size++;
					p--;
					if( lineend_found ) line--;
				}
				else {
					/* post event
					    IOTEXT_DELIMITER is not posted for the exclusive
						delimiter because it is ignored and IOTEXT_CHAR will be
					    posted for that character in the future - one event
					    for one character */
					if( (eventmask&IOTEXT_FLAG_DELIMITER) &&
					    iot->event(iob,iot,IOTEXT_DELIMITER,&c)==
					    	IOTEXT_DISPATCH_FAIL )
						result = IOTEXT_DISPATCH_FAIL;

					if( lineend_found ) {
						_IOTEXT_RESET_COLUMN(col,lastcol);
						p = iotext_skip_lineend( iob, p, &size );
						if( !p ) {
							result = FAIL_IO;
							goto l_result;
						}
					}
					else {
						col++;
					}
				}

				/* escape duplicated loops */
				result = IOTEXT_DELIMITER;
				goto l_terminate;
			}

			if( lineend_found ) {
				p = iotext_skip_lineend( iob, p, &size );
				if( !p ) {
					result = FAIL_IO;
					goto l_result;
				}
				_IOTEXT_RESET_COLUMN(col,lastcol);
			}
			else {
				col++;
			}


			/* post event : a meaningful character was read */
			if( eventmask&IOTEXT_FLAG_CHAR ) {
				switch( iot->event( iob, iot, IOTEXT_CHAR, &c ) ) {
					/* currently the buffer is locked so unlocking is needed */
					case IOTEXT_DISPATCH_FAIL :
					result = IOTEXT_DISPATCH_FAIL;
					goto l_terminate;

					case IOTEXT_SKIP_CHAR : continue;
				}
				eventmask = iot->eventmask; /* update event mask */
			}

			/* store the character if needed */
			if( bufsize>len ) {
				if( curbuf ) *curbuf++ = c;
			}
			else if( !(flag&IOTEXT_FLAG_FORCE) ) {
				if( curbuf ) {
					/* buffer is too short and flag to force not to stop
					   until a delimiter is found was not specified */
					iot->lastchar = c;
					result = FAIL_BUFFER;
					goto l_terminate;
				}
			}

			/* increase read length */
			len++;

		} /* end while( iteration for locked chars ) */

		iobcache.unlock( iobcache.context, p );
	} /* end while(1) */


	/* post process to report results */

	l_terminate:
	iobcache.unlock( iobcache.context, p );

	l_result:

	/* terminate string if its buffer was supplied */
	if( curbuf ) {
		*curbuf = 0;
		/* trailing blanks were not wanted */
		if( !(flag&IOTEXT_FLAG_KEEP_TRAIL_BLANK) ) {
			i = strBackwardTokenIndex( curbuf, len, str_white_char,
				STR_FLAG_END_OF_STRING|STR_FLAG_EXCLUSIVE );
			if( i>=0 ) {
				buf[i] = 0;
				len = i;
			}
		}
	} /* end if( curbuf ) */

	iot->line = line;
	iot->col = col;
	iot->lastcol = lastcol;
	iot->length = len;

	return result;
}

#undef _IOTEXT_RESET_COLUMN


int __API
iotextRead
	P_B(( iob, delimiter, buf, bufsize, flag, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* delimiter
	P_N		char __F* buf
	P_N		int bufsize
	P_N		unsigned flag
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result, wclen, tokenlen;
	char wcbuf[OP_ELEM_COUNT(str_white_char)];
	const char __F * wc;

	bufsize--; /* reserve a byte for string termination */

	/* set default delimiter if there is no given delimiter */
	if( !delimiter ) delimiter = str_white_char;

	if( !iot ) { /* set default io text data if there is no given data */
		iot = &io_text_default;
		OP_ENABLE( flag, IOTEXT_FLAG_DISABLE_COMMENT );
	}

	/* initialize white character delimiter for initial skipping */

	if( flag&IOTEXT_FLAG_KEEP_LEAD_BLANK ) {
		/* formal initialization not to cause compiler warning */
		wc = NULL;
		wclen = 0;
	}
	else {
		if( flag&IOTEXT_FLAG_KEEP_LEAD_DELIMITER ) {
			/* exclude white characters in the supplied tokens */
			int i;
			char c, __F* curbuf=wcbuf;
			for( i=0, wclen=0; i<OP_ELEM_COUNT(wcbuf); i++ ) {
				/*  the current white character is not found in the delimiter,
				   so add it */
				if( !strchr(delimiter,c=str_white_char[i]) ) {
					*curbuf++ = c;
					wclen++;
				}
			}
			*curbuf=0; /* XXX maybe not needed because memchr is used currently - inspect it */
			wc = wcbuf;
		}
		else {
			wc = str_white_char;
			wclen = OP_ELEM_COUNT(str_white_char)-1;
		}
	}

	tokenlen = strlen(delimiter);
	while( 1 ) {
		if( !(flag&IOTEXT_FLAG_KEEP_LEAD_BLANK) ) {
			result = iotext_read( iob, wc, wclen, NULL, 0,
				IOTEXT_FLAG_EXCLUSIVE|IOTEXT_FLAG_DISABLE_COMMENT, iot );
			/* if return value is IO_END, the given string buffer should be
			  terminated, so result==0 is excluded */
			if( result<0 ) break;
		}

		result = iotext_read(iob,delimiter,tokenlen,buf,bufsize,flag,iot);
		if( result!=IOTEXT_COMMENT || iot->length>0 ||
			(flag&IOTEXT_FLAG_KEEP_LEAD_COMMENT) )
			break;

	}

	return result;
}

__MODULEEND__



__MODULEBEGIN__( "iotrdstr.c" )
/* iotextReadString

  leadchar can be zero if it is not present - evidently possible case - but
 trailchar must not be zero */

int __API
iotextReadString
	P_B(( iob, bufptr, bufsize, leadchar, trailchar, lastchar, flag, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		char* bufptr
	P_N		int bufsize
	P_N		char leadchar
	P_N		char trailchar
	P_N		char lastchar
	P_N		unsigned flag
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;
	char delimiter[2];
	delimiter[0] = leadchar;
	delimiter[1] = 0;
	result = iotextProbeString( iob, delimiter, 1, lastchar, iot );
	if( result==0 ) {
		delimiter[0] = trailchar;
		return iotextReadStringExt
			( iob, bufptr, bufsize, delimiter, 1, flag, iot );
	}
	return result;
}

__MODULEEND__



__MODULEBEGIN__( "iotprbsa.c" )
/* iotextProbeString */

int __API
iotextProbeString
	P_B(( iob, s, len, lastchar, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* s
	P_N		int len
	P_N		char lastchar
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;
	char delimiter[2];
	if( len<=0 ) len = strlen(s);
	if( (delimiter[0]=*s)!=lastchar ) {	
//		/* wait the first character without garbage */
//		delimiter[1] = 0;
//		result = iotextWait( iob, delimiter, iot );
		result = iotextObserve( iob, NULL, iot );
		if( result<=0 )	return result;
		if( *s!=iot->lastchar ) return IOTEXT_GARBAGE_FOUND;
	}
	else {
		s++;
		len--;
	}
	if( len>0 ) {
		int size;
		_integer1 __F* p = iob->lock( iob->context, &size,len, IO_FLAG_READ );
		if( p ) {
			result = iotext_probe_string( iob, iot, &p, &size, s, len );
			iob->unlock( iob->context, result==0? p+len : p );	
			return result;
		}
		return FAIL_IO;
	}
	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "iotrdsx.c" )
/* iotextReadStringExt */

int __API
iotextReadStringExt
	P_B(( iob, bufptr, bufsize, trail, trail_len, flag, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		char* bufptr
	P_N		int bufsize
	P_N		const char __F* trail
	P_N		int trail_len
	P_N		unsigned flag
	P_N		struct _IOTEXT __F* iot
	P_E
{
	struct _IOBUFFER iobcache;
	int result, len;
	char delimiter[5];
	unsigned iotrdflag;

	/* set default data if data was not supplied */
	if( !iot ) iot = &io_text_default;

	/* skip leading blanks if requested */
	if( flag&IOTEXT_FLAG_SKIP_LEAD_BLANK )
		iotextObserve( iob, NULL, iot );

	/* find trailing string specifier */

	delimiter[0] = '\n';
	delimiter[1] = '\r';
	delimiter[2] = '\\';
	delimiter[3] = trail ? *trail : 0;
	delimiter[4] = 0;

	len = 0;

	/* flags for keeping leading and trailing blanks are basically enabled
		and if they are disabled, internal blanks will be skipped if
		comment is enabled to read a string */
	iotrdflag = (flag&(IOTEXT_FLAG_PRESERVE_LINE_END|IOTEXT_FLAG_FORCE))
		|IOTEXT_FLAG_KEEP_LEAD_BLANK|IOTEXT_FLAG_KEEP_TRAIL_BLANK;
	OP_COPY( iobcache, *iob );

	if( !(flag&IOTEXT_FLAG_ENABLE_COMMENT) )
		OP_ENABLE( iotrdflag, IOTEXT_FLAG_DISABLE_COMMENT );

	while( (result=iotextRead( iob, delimiter, bufptr, bufsize,
					iotrdflag, iot ))>=0 ) {

		register int l = iot->length;
		len += l;
		if( bufsize>0 ) {
			bufptr += l;
			bufsize -= l;
		}
		if( result==0 ) break;
		if( result!=IOTEXT_COMMENT ) {
			char lastchar;
			/* evidently returned value is IOTEXT_DELIMITER */

			lastchar = (char)iot->lastchar;

			if( lastchar==delimiter[sizeof(delimiter)-2] ) {
				if( iotextProbeString(iob,trail,trail_len,lastchar,iot)==0 ) {
					result = 1; /* success */
					break;
				}
				if( bufsize>0 ) {
					bufsize--;
					*bufptr++ = lastchar;
				}
				else if( !(flag&IOTEXT_FLAG_FORCE) ) {
					return FAIL_BUFFER;
				}
			}
			else {
				char __F* p;
				switch( lastchar ) {

					case '\\' :
					p = (char __F*) iobcache.lock
							( iobcache.context, NULL, 2, IO_FLAG_READ );
					if( p ) {
						char c=*p++;
						/* \ is the line continuation character */
						if( c=='\n' || c=='\r' ) {
							if( flag&IOTEXT_FLAG_NO_LINE_CONT ) {
								goto l_plain_bkslash;
							}
							else {
								if( c=='\n' ) {
									/* if cr followed, eat it */
									if( *p=='\r' ) {
										p++;
										lastchar = '\r';
									}
									else {
										lastchar = '\n';
									}
								}
								else {
									/* if lf followed, eat it */
									if( *p=='\n' ) {
										p++;
										lastchar = '\n';
									}
									else {
										lastchar = '\r';
									}
								} /* end if( c=='\n' ) */

							} /* end if( flag&IOTEXT_FLAG_NO_LINE_CONT ) */
						}
						else {
							/* \ is the character escape sequence */

							if( flag&IOTEXT_FLAG_RESERVE_CHAR_ESC ) {
								l_plain_bkslash:
								len+=2;
								if( bufsize>2 ) {
									*bufptr++ = '\\';
									*bufptr++ = c;
									bufsize-=2;
								}
								else if( !(flag&IOTEXT_FLAG_FORCE) )
									return FAIL_BUFFER;
							}
							else {
								switch( c ) {
									case 'n'	: c = '\n'; break;
									case 'r'	: c = '\r'; break;
									case 't'	: c = '\t'; break;
									case '\\' : c = '\\'; break;
									case '0'	: c = 0;
								}

								iot->length++;
								if( bufsize>1 ) {
									*bufptr++=c;
									bufsize--;
								}
								else if( !(flag&IOTEXT_FLAG_FORCE) ) {
									return FAIL_BUFFER;
								}
							}
						} /* end if( c=='\n' || c=='\r' ) */
						iobcache.unlock( iobcache.context, p );
					}
					else { /* iob_lock is failed */
						return FAIL_IO;
					}
					break;

					case '\r' :
					case '\n' :
					if( flag&IOTEXT_FLAG_ALLOW_LINE_END ) {
						len++;
						if( bufsize>1 ) {
							*bufptr++ = lastchar;
							bufsize--;
						}
						else if( !(flag&IOTEXT_FLAG_FORCE) ) {
							return FAIL_BUFFER;
						}
					}
					else {
						return IOTEXT_INVALID_STRING;
					}

				} /* end switch( lastchar ) */

			} /* end if */

		} /* end if( result==IOTEXT_COMMENT ) */

	} /* end while( iotextRead ) */

	if( trail && result==IO_END ) return IOTEXT_INVALID_STRING;

	if( result>=0 ) {
		/* skipping trailing blanks before the trailing character
			was requested */
		if( flag&IOTEXT_FLAG_SKIP_TRAIL_BLANK )
			len -= bufptr-strBackwardToken( bufptr, len, str_white_char,
				STR_FLAG_END_OF_STRING|STR_FLAG_EXCLUSIVE )+1;

		iot->length = len;
	}

	return result;
}

__MODULEEND__



__MODULEBEGIN__( "iotsleep.c" )
/* iotextSleep

 Sleep until a delimiter is found : garbage is allowed

  IOTEXT_FLAG_KEEP_LEAD_BLANK is applied for iotextSleep on
 2000.05.31 - actually bug fix for the case, for example, "\n" is specified
 as a delimiter but the initial skipping routine in iotextRead will kill a '\n'
 which must not be wasted at there and another line which must not be
 skipped will be skipped. */

int __API iotextSleep
	P_B(( iob, delimiter, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* delimiter
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;

	while( (result=iotextRead( iob, delimiter, NULL, 0,
					 IOTEXT_FLAG_KEEP_LEAD_BLANK, iot ))==IOTEXT_COMMENT ) {
	}

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "iotwait.c" )
/* iotextWait

 Wait until a delimiter is found : garbage is not allowed */

int __API iotextWait
	P_B(( iob, delimiter, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* delimiter
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;
	struct _IOTEXT iotdef;

	/* set default data if there is no given data */
	if( !iot ) {
		mem_short_set( &iotdef, 0, sizeof(iotdef) );
		iot = &iotdef;
	}

#if 0
	/* ignore the case which returned by comment */
	while( (result=iotextRead( iob, delimiter, NULL, 0,
					 IOTEXT_FLAG_KEEP_LEAD_DELIMITER, iot ))==IOTEXT_COMMENT ) {
		if( iot->length>0 )
			return IOTEXT_GARBAGE_FOUND;
	}

	if( result>=0 && iot->length>0 )
		return IOTEXT_GARBAGE_FOUND;

	return result;
#else
	result = iotextObserve(iob,NULL,iot);
	if( result>0 ) {
		char buf[2];
		buf[0] = iot->lastchar;
		buf[1] = 0;
		result = iotextRead(iob,buf,NULL,0,0,iot);
		if( result>0 ) {
			if( memchr(delimiter,iot->lastchar,strlen(delimiter))==NULL )
				return IOTEXT_GARBAGE_FOUND;
		}
	}
#endif
	return result;
}
__MODULEEND__



__MODULEBEGIN__( "iotskipl.c" )
/* iotextSkipLine */

int __API
iotextSkipLine
	P_B(( iob, n, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		unsigned long n
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result=FAIL_IO;
	while( n-->0 ) {
		result = iotextSleep(iob,str_lineend_char,iot);
		if( result<=0 ) break;
	}
	return result;
}

__MODULEEND__



__MODULEBEGIN__( "iotobsrv.c" )
/* iotextObserve

 Observe specified tokens. */

int __API iotextObserve
	P_B(( iob, delimiter, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* delimiter
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;
	while( (result=iotextRead( iob,delimiter, NULL, 0,
	                 IOTEXT_FLAG_KEEP_LEAD_DELIMITER|
	                 IOTEXT_FLAG_EXCLUSIVE, iot ))==IOTEXT_COMMENT );
	return result;
}
__MODULEEND__



__MODULEBEGIN__( "iotconv.c" )
/* iotextConvert

 Returns source format.

 originated from tounix utility
 1998. 10. 9  DOS text to UNIX text conversion
 1998. 10. 25 UNIX text to DOS text conversion
 Hee S. Chung */

unsigned __API
iotextConvert
	P_B(( dest, src, endline ))
	P_1		struct _IO __F* dest
	P_N		struct _IO __F* src
	P_N		unsigned endline
	P_E
{
	_integer1 c;
	long n=0;
	unsigned result=0, leader=0;
	char lineend=0;

	char __F* str= (char __F*)src;

	/* string support on 2000.03.13 using internal flag for iotextWrite */

	while( endline&_IOTEXT_FLAG_MEMORY?
				 (c=*str++)!=0 : (n=ioRead(src,&c,1,0))==1 ) {

		switch( c ) {
			_integer1 cr, lf;

			/* if a line end was found */
			case '\r' :
			result |= IOTEXT_CR;
			if( !leader ) leader=IOTEXT_CR;
			goto lineend_found;

			case '\n' :
			result |= IOTEXT_LF;
			if( !leader ) leader=IOTEXT_LF;

			lineend_found:
			/* line end char was not found or found line end char is
				appeared again, that means successive new line is found
				without non-blank chars, bug fixed 1999.05.28 */
			if( !lineend || lineend==c ) {
				/* consider inverted case */
				if( endline&0x4 )
					cr='\n', lf='\r';
				else
					cr='\r', lf='\n';

				/* write line end chars */
				if( endline&IOTEXT_CR )
					if( !ioFullWrite(dest,&cr,1,0) ) return 0;
				if( endline&IOTEXT_LF )
					if( !ioFullWrite(dest,&lf,1,0) ) return 0;

				/* currently accepted line end */
				lineend = c;
			}
			break;

			default :
			/* line end char was not found */
			lineend = 0;
			/* just write the read byte to destination */
			if( !ioFullWrite(dest,&c,1,0) ) return 0;
		}
	}

	if( n==FAIL_IO ) return 0;

	return
		/* if result was not modified */
		result==0? IOTEXT_UNKNOWN :
		/* if lf cr found */
		result==IOTEXT_CRLF && leader==IOTEXT_LF? IOTEXT_LFCR :
		/* plain case */
		result;
}

__MODULEEND__



__MODULEBEGIN__( "iotfrmat.c" )
/* iotextFormat */

unsigned __API iotextFormat
	P_B(( src, readbytes ))
	P_1		struct _IO __F* src
	P_N		long __F* readbytes
	P_E
{
	char c, next;
	long n, total=0;

	while( (n=ioRead(src,&c,1,0))==1 ) {

		total++;

		/* check line end */
		switch( c ) {
			case '\r' :
			case '\n' :
			n = ioRead(src,&next,1,0);
			if( n==FAIL_IO ) return 0;
			if( readbytes ) *readbytes = total + n;
			if( c=='\r' ) {
				if( n==0 || next!='\n' ) return IOTEXT_CR;
				return IOTEXT_CRLF;
			}
			else {
				if( n==0 || next!='\r' ) return IOTEXT_LF;
				return IOTEXT_LFCR;
			}
		} /* end switch( currently read char ) */
	} /* end while( a byte can be read from src ) */

	if( readbytes )
		*readbytes = total;

	return n==FAIL_IO ? 0 : IOTEXT_UNKNOWN;
}

__MODULEEND__



__MODULEBEGIN__( "iotwrite.c" )
/* iotextWrite

 Returns string format or FAIL_IO if io operation was failed. */

_bool __API
iotextWrite
	P_B(( io, buf, endline ))
	P_1		struct _IO __F* io
	P_N		const char __F* buf
	P_N		unsigned endline
	P_E
{
	/* call iotextConvert with _IOTEXT_FLAG_MEMORY */
	return iotextConvert( io, (struct _IO __F*)buf, endline|_IOTEXT_FLAG_MEMORY )>0;
}
__MODULEEND__



__MODULEBEGIN__( "iotwstr.c" )
/* iotextWriteString */

_bool __API
iotextWriteString
	P_B(( io, s ))
	P_1		struct _IO __F* io
	P_N		const char __F* s
	P_E
{
	/* no byte ordering will be occurred so casting to char* is valid */
	return ioFullWrite( io, (void __F*)s, strlen(s), 0 );
}

__MODULEEND__


/* end of iotext.c */

