/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */



/* I/O Extra Features */



#include <coralsrc/memory.h>
#include <coralsrc/io.h>
#include <coral/compat/string.h>
#include <coral/inline.h>



__MODULEBEGIN__( "iofill.c" )
/* ioFill */

_bool __API
ioFill
	P_B(( io, buf, nbuf, n, byteorder ))
	P_1		struct _IO __F* io
	P_N		void __F* buf
	P_N		unsigned nbuf
	P_N		unsigned long n
	P_N		unsigned byteorder
	P_E
{
	if( nbuf>1 )
		memSetExt( buf, nbuf, byteorder&MASK_COMBINATED_SIZE, NULL );
	else {
		if( io->buffer ) {
			int value = *(_integer1 __F*)buf;
			while( n>0 ) {
				int avail;
				_integer1 __F* p = ioLockBuffer( io, &avail, 0, IO_FLAG_WRITE );
				if( !p ) return 0;
				if( (unsigned long)avail<=n ) {
					n -= avail;
				}
				else {
					avail = (int)n;
					n = 0;
				}
				mem_short_set( p, value, avail );
				ioUnlockBuffer( io, p+avail );
			}
			return 1;
		} /* end if( io->buffer ) */
		nbuf = 1; /* usually this case is used when the io buffer was assigned */
	}

	/* reduce requested size by iteration */
	while( n>nbuf ) {
		if( !ioFullWrite(io,buf,nbuf,byteorder) ) return 0;
		n-=nbuf;
	}
	if( !ioFullWrite(io,buf,n,byteorder) ) return 0;

	return 1;
}
__MODULEEND__



__MODULEBEGIN__( "iordstr.c" )
/* ioReadString

 2000.02.21 */

char __F* __API
ioReadString
	P_B(( io, buf, bufsize, byteorder ))
	P_1		struct _IO __F* io
	P_N		char __F* buf
	P_N		int bufsize
	P_N		unsigned byteorder
	P_E
{
	long l;

	/* read string length */
	if( ioFullRead(io,&l,1,byteorder) ) {
		int skip = 0;
		register int n = memRestoreInt( &l, byteorder&MASK_COMBINATED_SIZE );

		/* read string */
		if( buf ) {
			bufsize--;
			if( bufsize>=0 ) {
				if( bufsize<n ) {
					/* the given buffer is too short and some bytes need to be skipped */
					skip = n - bufsize;
					n = bufsize;
				}
			}
			else {
				buf = NULL;
			}
		}
		else {
			buf = (char __F*) memAlloc( n+bufsize+1 );
		} /* end if( buf ) */

		if( buf ) {
			/* zero length string or io operation is successful */
			if( n==0 || ( ioFullRead(io,buf,n,0) && ioSkip(io,skip) ) ) {
				buf[n] = 0;
				return buf;
			}
		}

	} /* end if( read string length ) */

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "iowrstr.c" )
/* ioWriteString

 2000.02.21 */

_bool __API
ioWriteString
	P_B(( io, s, len, byteorder ))
	P_1		struct _IO __F* io
	P_N		const char __F* s
	P_N		int len
	P_N		unsigned byteorder
	P_E
{
	long n;

	if( len==0 && s ) {
		/* length was not calculated */
		len=strlen(s);
	}
	else if( s==NULL ) {
		/* buffer was not supplied so write null string specifier */
		n = 0;
		return ioFullWrite( io, &n, 1, byteorder );
	}

	n = len;
	memReduceInt( &n, byteorder&MASK_COMBINATED_SIZE );

	return ioFullWrite( io, &n, 1, byteorder ) &&
				 ioFullWrite( io, (void __F*)s, len, 0 );
}
__MODULEEND__



__MODULEBEGIN__( "iotransf.c" )
/* ioTransfer

 1999.04.04 */

int __API
ioTransfer
	P_B(( dest, src, totalsize, buf, bufsize, progress, context ))
	P_1		struct _IO __F* dest
	P_N		struct _IO __F* src
	P_N		long totalsize
	P_N		_integer1 __F* buf
	P_N		int bufsize
	P_N		int (__API* progress) P_A(( void __F* context, int bufsize ))
	P_N		void __F* context
	P_E
{
	int result=0;

	/* callback to indicate the start of progression */
	if( !progress || (result=progress(context,0))==0 ) {

		_integer1 __F* p;
		_bool loop;
		int r, w, normal=1;

		struct _io_buffer __F* buffer = src->buffer;

		/* transfer buffer was supplied */
		if( buf ) {
			if( bufsize<=0 ) return FAIL_BUFFER; /* trivial case */
			p = buf;
		}
		else {

			if( buffer && buffer->size>=bufsize ) {
				int remained;

				bufsize = buffer->size;

				/* buffer pointer will be used to update 'p' and it should
					not be freed : 'if(!buf)' is checked before freeing 'p' */
				buf = buffer->buf;

				/* check there is remained bytes in the buffer :
					refill is not allowed */
				remained = io_ready_rmode( src, OP_MAXINT(long) );
				if( remained>0 ) {
					p = ioLockBuffer( src, &bufsize, 0, IO_FLAG_READ );
					if( !p ) return FAIL_IO;
					normal = 0;
				}
				else {
					p = buf;
				} /* end if remained>0 ) */
			}
			else {
				/* allocate transfer buffer */
				if( bufsize<=0 ) return FAIL_BUFFER; /* buffer size was not explicitly specified */
				p = (_integer1 __F*)memAlloc(bufsize);
				if( !p ) return FAIL_MEMORY;
			}

		} /* end if ( buf ) */


		loop = 1;

		if( normal )
			src->buffer = NULL; /* disable io buffering */

		while( loop ) {

			if( totalsize>0 ) {
				/*  if initially totalsize>0, it will be maintained by the
				   following routine and totalsize>0 is alway valid to know
				   its initial state */
				if( totalsize>bufsize ) {
					totalsize -= bufsize;
				}
				else {
					bufsize = (int)totalsize;
					loop = 0;
				}
			} /* end if( totalsize>0 ) */

			if( normal ) {

				r = ioRead( src, p, bufsize, 0 );

				if( r<bufsize ) {

					if( r<0 ) {
						result = r;
						break;
					}

					if( totalsize>0 )
						result = IO_INSUFFICIENT_SOURCE;

					if( r==0 )
						break;

					loop = 0;
			
				} /* end if( r<bufsize ) */

				w = ioWrite( dest, p, r, 0 );
			}
			else {
				normal = 1;

				/* if bufsize is truncated above, remained bytes in the buffer should be used later */
				ioUnlockBuffer( src, p+bufsize );

				w = ioWrite( dest, p, bufsize, 0 );

				r = bufsize;
				bufsize = buffer->size;
				p = buf;
				src->buffer = NULL;
			} /* end if( normal ) */

			if( w>0 && progress ) {
				result = progress( context, w );
				if( result!=0 )
					loop = 0;
			}

			if( w<r ) {
				result = FAIL_IO;
				break;
			}

		} /* end while( loop ) */

		src->buffer = buffer; /* recover original io buffer */

		/* free transfer buffer */
		if( !buf )
			memFree( p );
	} /* if( !progress || (result=progress(context,0))==0 ) */

	return result;
}

__MODULEEND__



__MODULEBEGIN__( "iobsrch.c" )
/* ioBinarySearch

 Parameter 'cl' in supplied function pointer 'compare'. :
 *cl==0 if p is io object, *cl==1 if p is a cached element */

int __API
ioBinarySearch( io, n, size, key, compare, param, index )
	struct _IO __F* io;
	unsigned long n;
	unsigned size;
	const void __F* key;
	int (__API* compare) P_A(( struct _IO __F* io, const void __F* key, void __F* param, int __F* cl ));
	void __F* param;
	unsigned long __F* index;
{
	/* following routines are equivalent to memBinarySearch */

	int result = FAIL_UNAVAILABLE;

	if( n>0 ) {
		unsigned long offset = 0; /* offset relative to the current position */
		unsigned long l=0, m=0;
		result = FAIL_ENTRY;
		n--;
		while( l<=n ) {

			int cl, keyend_offset;
			unsigned long newoffset;
			long diff;

			/* middle position of l and n */
			m = (l+n)>>1;

			/* get the new offset */
			newoffset = (unsigned long)m*(unsigned long)size;

			/* move stream position relatively */
			if( newoffset>=offset ) diff = (long)(newoffset-offset);
			else diff = -(long)(offset-newoffset);
			if( ioSeek( io, diff, NULL, IO_SEEK_CURRENT )==FAIL_IO ) {
				result = FAIL_IO;
				break;
			}

			offset = newoffset;

			/* the major problem which this method cannot read a key is byte ordering */
			cl = 0; /* not cached */
			if( (keyend_offset=compare( io, key, param, &cl ))<=0 ) {
				result = keyend_offset;
				break;
			}

			offset += (unsigned long)keyend_offset;

			/* check result and determine to continue iteration or not */
			if( cl==0 ) {
				result = 0;
				break;
			}
			if( cl<0 ) {
				l = m+1;
			}
			else { /* cl>0 */
				if( m<=0 ) break;
				n = m-1;
			}
		} /* end while( l<=n && n!=(unsigned)-1 ) */

		if( index ) *index = m;
	} /* end if( n>0 ) */

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "iordhier.c" )
/* ioReadHierarchy */

int __API
ioReadHierarchy
	P_B(( io, excl, nexcl, current_key ))
	P_1		struct _IO __F* io
	P_N		char __F*__F*   excl
	P_N		int             nexcl
	P_N		char __F*__F*   current_key
	P_E
{
	_unsigned1 property;
	int result;
	unsigned byte_order = FLAG_LITTLE_ENDIAN;

	if( current_key )
		*current_key = NULL;

	while( 1 ) {
		result = ioRead(io,&property,1,0);
		if( result==1 ) {
			unsigned long size;
			unsigned size_field, key_field;

			switch( property&IO_MASK_NODE_PROPERTY ) {
				case IO_TERMINAL_NODE:
				case IO_ENTER_INTERNAL_NODE:
				break;
				case IO_LEAVE_INTERNAL_NODE:
				return IO_LEAVE_INTERNAL_NODE;
				default:
				return FAIL_UNAVAILABLE;
			}

			/* read subtree size */

			switch( property&IO_MASK_SUBTREE_SIZE_FIELD ) {
				case IO_SUBTREE_SIZE_FIELD_1: size_field = 1; break;
				case IO_SUBTREE_SIZE_FIELD_2: size_field = 2; break;
				case IO_SUBTREE_SIZE_FIELD_4: size_field = 4; break;
				default: size_field = 0;
			}

			if( size_field>0 ) { /* subtree skip info is present */

				result = (int)ioRead( io, &size, 1, byte_order|size_field );
				if( result<=0 ) return result;

				size = (unsigned long)memRestoreInt( &size, size_field );
			}
			else {
				size = 0;
			}

			/* get subtree key */

			switch( property&IO_MASK_KEY_LENGTH_FIELD ) {
				case IO_KEY_LENGTH_FIELD_1:
				key_field = 1;
				break;
				case IO_KEY_LENGTH_FIELD_2:
				key_field = 2;
				break;
				default:
				key_field = 0;
			}

			if( key_field>0 ) {
				_bool found;
				char __F* k, __F* key = ioReadString( io, NULL, 0,
									byte_order|key_field );
				if( key==NULL ) return FAIL_IO;

				found = 0;
				if( nexcl>0 ) {
					unsigned nk = nexcl;
					MEM_BEGIN_BSEARCH( nk )
						MEM_ENTER_BSEARCH_LOOP( nk )
							int cl = strcmp(excl[m],key);
							if( cl==0 ) {
								found = 1;
								break;
							}
							if( cl<0 ) {
								MEM_BSEARCH_LESS
							}
							else {
								MEM_BSEARCH_GREATER(nk)
							}
						MEM_LEAVE_BSEARCH_LOOP
					MEM_END_BSEARCH
				}
				else if( excl ) {
					char __F*__F* kt = excl;
					while( (k=*kt)!=NULL ) {
						if( strcmp(k,key)==0 ) {
							found = 1;
							break;
						}
						kt++;
					}
				}

				if( found && size>0 ) {
					size -= 1 + size_field + key_field + strlen(key);
/* XXX check seek error */
					memFree( key );
/*					ioSeek( io, size, NULL, IO_SEEK_CURRENT );*/
					ioSkip( io, size );
					continue;
				}

				if( current_key ) *current_key = key;
				else memFree( key );
			} /* if( key_field>0 ) */

			/* not to return zero for the normal case, 0x100 is combinated */
			return (int)(0x100|(unsigned)property);
		}
		break;
	}
	return result;
}
__MODULEEND__



__MODULEBEGIN__( "ioenhier.c" )
/* ioEnterHierarchy
 header format : property(1)  size  key_len  key */

int __API
ioEnterHierarchy
	P_B(( io, property, size, key, stk ))
	P_1		struct _IO __F* io
	P_N		unsigned property
	P_N		unsigned long __F* size
	P_N		const char __F* key
	P_N		struct _IO_HIERARCHY_STACK __F* stk
	P_E
{
	unsigned long total_size, data_size;
	unsigned key_field, key_len;
	unsigned size_field, byte_order = property&0xff00;
	property &= 0x00ff;


	/* key field */

	if( key ) {
		key_len = strlen(key);
		if( key_len<=OP_MAXUINT(_unsigned1) ) {
			property |= IO_KEY_LENGTH_FIELD_1;
			key_field = 1;
		}
		else {
			key_field = 2;
			property |= IO_KEY_LENGTH_FIELD_2;
		}
	}
	else {
		key_len = key_field = 0;
	}

	/* size field */

	data_size = *size;
	total_size = 1 + data_size + key_len + key_field;

	size_field = 0;

	/* explicit size for terminal node was supplied */
	if( data_size>0 && (property&IO_MASK_NODE_PROPERTY)==IO_TERMINAL_NODE ) {
		/* variable size field for the terminal node */
		OP_DISABLE( property, IO_MASK_SUBTREE_SIZE_FIELD );
		if( total_size+1<=OP_MAXUINT(_unsigned1) ) {
			property |= IO_SUBTREE_SIZE_FIELD_1;
			goto l_size_field_1;
		}
		else if( total_size+2<=OP_MAXUINT(_unsigned2 ) ) {
			property |= IO_SUBTREE_SIZE_FIELD_2;
			goto l_size_field_2;
		}
		else {
			property |= IO_SUBTREE_SIZE_FIELD_4;
			goto l_size_field_4;
		}
	}
	else {
		switch( (property&IO_MASK_SUBTREE_SIZE_FIELD) ) {
			case IO_SUBTREE_SIZE_FIELD_1: l_size_field_1:
			total_size++;
			size_field = 1;
			break;
			case IO_SUBTREE_SIZE_FIELD_2: l_size_field_2:
			total_size += 2;
			size_field = 2;
			break;
			case IO_SUBTREE_SIZE_FIELD_4: l_size_field_4:
			total_size += 4;
			size_field = 4;
		}
	} /* if( terminal node ) */

	if( !ioWriteInt( io, property, 1 ) ) return FAIL_IO;

	if( (property&IO_MASK_NODE_PROPERTY)!=IO_TERMINAL_NODE ) {
			total_size++; /* field to store indicator to leave subtree */
		/* get size field offset */
		stk->offset_high = 0;
		stk->offset =
			ioSeek( io, 0, &stk->offset_high, IO_SEEK_CURRENT );
		stk->size_field = size_field;
	}

	if( size_field>0 && !ioWriteInt( io, total_size, byte_order|size_field ) )
		return FAIL_IO;

	if( key && !ioWriteString( io, key, key_len, byte_order|key_field ) )
		return FAIL_IO;

	*size = total_size;

	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "iolehier.c" )
/* ioLeaveHierarchy */

int __API
ioLeaveHierarchy
	P_B(( io, stk, byte_order ))
	P_1		struct _IO __F* io
	P_N		struct _IO_HIERARCHY_STACK __F* stk
	P_N		unsigned byte_order
	P_E
{
	unsigned long offset;
	long offset_high;
	unsigned size_field;

	if( !ioWriteInt( io, IO_LEAVE_INTERNAL_NODE, 1 ) ) return FAIL_IO;

	size_field = stk->size_field;
	if( size_field>0 ) {
		/* XXX check return values for ioSeek */
		offset_high = 0;
		offset = ioSeek( io, 0, &offset_high, IO_SEEK_CURRENT );

		/* write size field */
		ioSeek( io, stk->offset, &stk->offset_high, IO_SEEK_BEGIN );
		if( !ioWriteInt( io, stk->size, byte_order|size_field ) )
			return FAIL_IO;

		ioSeek( io, offset, &offset_high, IO_SEEK_BEGIN );
	}
	return 0;
}
__MODULEEND__



/* EOF */

