/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */



/* Coral Open List

 1996.01.
 1998.02.   Reviewed
 2000.08.08 List transformation macros are introduced
            Reviewed using parameter compatibility macro spec.
 2000.12.10 Variable item list
 2001.03.27 Reviewed
 2001.05.11 Reviewed for LGPL release

	Circular list can be detected by checking head->prev, so it is possible
 to supply functions which can be used for both circular and linear list.
 But those functions cannot be used for singly linked list. It maybe rare case
 but with the current implementation of coral linked list, there is a chance
 to supply a node which is not the head node, as a head node to be iterated */



#undef __DEBUG__
#include <coralsrc/memory.h>
#include <coralsrc/list.h>
#include <coralsrc/string.h>
#include <coral/inline.h>
#include <coral/compat/string.h>
__MODULEEND__



#ifndef __TEMPLATE__



/*------------------------------------------------------------------------*
	memAlloc based list node allocation and deallocation
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "lstack.c" )

/* listPush

 Prototype changed, 1998.03.23 for MLLOG creation
 ( ML structure reordering ) */

struct _LIST __F* __API
listPush
	P_B(( head, size ))
	P_1		struct _LIST __F*__F* head
	P_N		long size
	P_E
{
	struct _LIST __F* newnode = (struct _LIST __F*) memAlloc( size );

	if( newnode ) {
		/* the following routine is used in listAdd but it is very short and
			simple so re-written on there */
		newnode->next = *head,
		*head=newnode;
	}

	return newnode;
}



/* listPop */

void __API
listPop
	P_B(( head ))
	P_1		struct _LIST __F*__F* head
	P_E
{
	struct _LIST __F* delnode = *head;
	if( delnode ) { /* exclude underflow */
		*head = delnode->next;
		memFree( delnode );
	}
}

__MODULEEND__



__MODULEBEGIN__( "lsingle.c" )

/* listCreate */

struct _LIST __F* __API
listCreate
	P_B(( head, where, size ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* where
	P_N		long size
	P_E
{
	struct _LIST __F* what = (struct _LIST __F*) memAlloc( size );
	listAdd( head, where, what );
	return what;
}



/* listRemove */

void __API
listRemove
	P_B(( head, what, prev ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* what
	P_N		struct _LIST __F* prev
	P_E
{
	listCut( head, what, prev );
	memFree(what);
}

__MODULEEND__



__MODULEBEGIN__( "lclear.c" )
/* listClear

	Here, 'for each' procedure is not considered, and pre-iteration
 to call destructor for each nodes is needed. */

void __API
listClear
	P_B(( head ))
	P_1		struct _LIST __F*__F* head
	P_E
{
	while( *head )
		listPop( head );
}
__MODULEEND__



__MODULEBEGIN__( "ldouble.c" )
/* listDNCreate */

struct _LIST __F* __API
listDNCreate
	P_B(( head, where, size, flag ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* where
	P_N		long size
	P_N		unsigned flag
	P_E
{
	struct _LIST __F* newnode = (struct _LIST __F*)
		( flag&FLAG_CLEAR? memCAlloc : memAlloc )( size );
	if( newnode ) listDNAdd( head, where, newnode, flag );
	return newnode;
}



/* listDNRemove */

void __API
listDNRemove
	P_B(( head, what ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* what
	P_E
{
	listDNCut( head, what );
	memFree( what );
}

__MODULEEND__



__MODULEBEGIN__( "lcclear.c" )
/* listCNClear */

void __API
listCNClear
	P_B(( head ))
	P_1		struct _LIST __F*__F* head
	P_E
{
	if( *head ) {
		LIST_BEGIN_LINEAR( *head );
		listClear( head );
	}
}
__MODULEEND__



/*------------------------------------------------------------------------*
	Add & cut operations
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "ladd.c" )
/* listAdd */

void __API
listAdd
	P_B(( head, where, what ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* where
	P_N		struct _LIST __F* what
	P_E
{
	if( what ) {
		register struct _LIST __F* headnode = *head;
		if( where && headnode ) {
			what->next  = where->next;
			where->next = what;
		}
		else {
			what->next = headnode;
			*head=what;
		}
	}
}
__MODULEEND__



__MODULEBEGIN__( "lcut.c" )
/* listCut */

void __API
listCut
	P_B(( head, what, prev ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* what
	P_N		struct _LIST __F* prev
	P_E
{
	if( what ) {
		if( *head==what ) {
			*head=what->next;
		}
		else {
			if( !prev )
				prev = listFindPrev( *head, what );

			if( prev )
				prev->next = what->next;
		}
	}
}

__MODULEEND__



__MODULEBEGIN__( "ldadd.c" )
/* listDNAdd */

void __API
listDNAdd
	P_B(( head, where, what, flag ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* where
	P_N		struct _LIST __F* what
	P_N		unsigned flag
	P_E
{
	if( what ) {
		register struct _LIST __F* headnode = *head;

		if( headnode ) {
			if( where==NULL ) {
				where = headnode;
				goto l_add_prev;
			}

			if( flag&LIST_FLAG_ADD_PREV ) {
				register struct _LIST __F* whereprev;
				l_add_prev:
				whereprev = where->prev;
				what->prev = whereprev;
				if( whereprev )
					whereprev->next = what;

				what->next = where;
				where->prev = what;

				/* modification was done for the case when the insertion
				  position is 'head', 'head' should be the inserted node for
				  LIST_FLAG_ADD_PREV mode ( regardless of linear or circular
				  node list ) and to support ordered list (queue) for circular
				  list which was accomplished by adding a node to head in
				  LIST_FLAG_ADD_PREV mode, which became unavailable after above
				  modification, flag&LIST_FLAG_QUEUE is added */
				if( (flag&LIST_FLAG_QUEUE) || where!=headnode )
					return;
			}
			else {
				register struct _LIST __F* wherenext = where->next;
				what->next = wherenext;
				if( wherenext )
					wherenext->prev = what;

				what->prev = where;
				where->next = what;

				/* '*head = what' routine is shared to update list header for
				  LIST_FLAG_ADD_PREV mode so 'return' must be used here for
				  LIST_FLAG_ADD_NEXT mode */
				return;
			}
		}
		else {
			/* linear and circular list is settled only when the first node
			  is added into the list */
			what->next = what->prev = flag&LIST_FLAG_CIRCULAR? what:NULL;
		}

		/* following code is shared for the first node initialization and
		  LIST_FLAG_ADD_PREV mode */
		*head = what;
	} /* end if( what ) */
}
__MODULEEND__



__MODULEBEGIN__( "ldcut.c" )
/* listDNCut */

void __API
listDNCut
	P_B(( head, what ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* what
	P_E
{
	if( what ) {
		register struct _LIST __F* whatprev, __F* whatnext = what->next;
		if( *head==what )
			*head = *head==whatnext ? NULL : whatnext;

		whatprev = what->prev;
		if( whatprev )  whatprev->next = whatnext;
		if( whatnext )  whatnext->prev = whatprev;
	}
}
__MODULEEND__



#ifdef RESERVED
__MODULEBEGIN__( "ldattach.c" )
/* listDNAttach

	'to' can be NULL and for this case,
 1. given lists must be linear list
 2. the insertion position must be the tail node of the destination list.
 3. insertion mode must be LIST_FLAG_ADD_NEXT */

void __API
listDNAttach
	P_B(( head, where, from, to, flag ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* where
	P_N		struct _LIST __F* from
	P_N		struct _LIST __F* to
	P_N		unsigned flag
	P_E
{
	register struct _LIST __F* headnode;

	headnode = *head;

	if( headnode && from ) {
		/* default insertion position is the head node */
		if( !where ) where = headnode;

		if( flag&LIST_FLAG_ADD_PREV ) {
			register struct _LIST __F* whereprev = where->prev;
			from->prev = whereprev;
			if( whereprev )
				whereprev->next = from;

			to->next = where;
			where->prev = to;

			/* if insertion position was 'head', 'head' need to be the inserted
			  node for LIST_FLAG_ADD_PREV mode ( regardless of linear or
			  circular node list ) and to support ordered list (queue) for
			  circular list which was accomplished by adding a node to head in
			  LIST_FLAG_ADD_PREV mode, which became unavailable after above
			  modification, flag&LIST_FLAG_QUEUE is added */

			if( (flag&LIST_FLAG_QUEUE) || where!=headnode )
				return;
		}
		else {
			if( to ) {
				register struct _LIST __F* wherenext = where->next;
				to->next = wherenext;
				if( wherenext )
					wherenext->prev = to;
			}

			from->prev = where;
			where->next = from;

			/* '*head = from' routine is shared to update list header for
			  LIST_FLAG_ADD_PREV mode so 'return' must be used here for
			  LIST_FLAG_ADD_NEXT mode */
			return;
		}
	}
	else {
		/* linear and circular list is settled only when the first node
		  is added into the list */
		if( to ) {
			if( flag&LIST_FLAG_CIRCULAR ) {
				from->prev = to;
				to->next = from;
			}
			else {
				from->prev = to->next = NULL;
			}
		}
	}

	/* following code is shared for the first node initialization and
		LIST_FLAG_ADD_PREV mode */
	*head = from;
}

__MODULEEND__



__MODULEBEGIN__( "ldcut.c" )
/* listDNDetach

 'to' can be NULL and for this case, the given list must be linear list. It is
 equivalent if 'to' is the tail node of the given list. */

void __API
listDNDetach
	P_B(( head, from, to ))
	P_1		struct _LIST __F*__F* head
	P_N		struct _LIST __F* from
	P_N		struct _LIST __F* to
	P_E
{
	if( from ) {
		struct _LIST __F* fromprev, __F* tonext = to? to->next : NULL;

		if( *head==from )
			*head = *head==from->next ? NULL : tonext;

		fromprev = from->prev;
		if( fromprev )  fromprev->next = tonext;
		if( tonext )  tonext->prev = fromprev;
	}
}
__MODULEEND__
#endif



/*-------------------------------------------------------------------------*
	 Linear list iterations
 *-------------------------------------------------------------------------*/



__MODULEBEGIN__( "lcount.c" )
/* listCount */

unsigned long __API
listCount
	P_B(( head ))
	P_1		struct _LIST __F* head
	P_E
{
	register unsigned long count = 0;

	while( head ) {
		count++;
		head = head->next;
	}

	return count;
}
__MODULEEND__



__MODULEBEGIN__( "lnode.c" )
/* listNode */

struct _LIST __F* __API
listNode
	P_B(( head, i ))
	P_1		struct _LIST __F* head
	P_N		long i
	P_E
{
	if( head && i>=0 ) {

		while( i-- ) {
			head = head->next;
			if( !head ) break;
		}

		return head;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lindex.c" )
/* listIndex */

long __API
listIndex
	P_B(( head, node ))
	P_1		struct _LIST __F* head
	P_N		struct _LIST __F* node
	P_E
{
	if( head && node ) {
		long index=0;
		while( head ) {
			if( head==node ) return index;
			index++;
			head = head->next;
		}
	}
	return -1;
}
__MODULEEND__



__MODULEBEGIN__( "linsnod.c" )
/* listInsertionNode */

struct _LIST __F* __API
listInsertionNode
	P_B(( head, index, flag ))
	P_1		struct _LIST __F* head
	P_N		long index
	P_N		unsigned __F* flag
	P_E
{
	if( index!=0 ) {
		struct _LIST __F* headprev = NULL;

		while( --index!=0 ) {
			if( !head ) {
				if( flag )
					*flag = index<0? LIST_FLAG_ADD_NEXT :
						LIST_FLAG_ADD_NEXT|LIST_FLAG_INVALID_INDEX;
				return headprev;
			}
			headprev = head;
			head = head->next;
		}

		if( flag )
			*flag = LIST_FLAG_ADD_NEXT;
		return head;
	}

	if( flag )
		*flag = LIST_FLAG_ADD_PREV;
	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lexist.c" )
/* listExist */

_bool __API
listExist
	P_B(( head, node ))
	P_1		struct _LIST __F* head
	P_N		struct _LIST __F* node
	P_E
{
	while( head ) {
		if( head==node ) return 1;
		head = head->next;
	}

	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "lfndtail.c" )

/* listFindTail */

struct _LIST __F* __API
listFindTail
	P_B(( head ))
	P_1		struct _LIST __F* head
	P_E
{
	if( head ) {
		register struct _LIST __F* headnext;
		while( (headnext=head->next)!=NULL )
			head = head->next;
	}
	return head;
}
__MODULEEND__



__MODULEBEGIN__( "lfindprv.c" )
/* listFindPrev */

struct _LIST __F* __API
listFindPrev
	P_B(( head, what ))
	P_1		struct _LIST __F* head
	P_N		struct _LIST __F* what
	P_E
{
	if( head==NULL || what==head ) return NULL;

	while( head->next!=what )
		head = head->next;

	return head;
}
__MODULEEND__



__MODULEBEGIN__( "lbitfind.c" )
/* listFindBit */

struct _LIST __F* __API
listFindBit
	P_B(( head, mask, offset, size, flag ))
	P_1		struct _LIST __F* head
	P_N		unsigned long mask
	P_N		int offset
	P_N		unsigned size
	P_N		unsigned flag
	P_E
{
	while( head ) {

		unsigned long masked =
			mask & memRestoreInt((_unsigned1 __F*)head+offset,size);

		if( flag&LIST_FLAG_ALL )
			masked = mask==masked;

		if( flag&LIST_FLAG_NOT ) {
			if( !masked ) return head;
		}
		else {
			if( masked ) return head;
		}

		head = head->next;

	} /* end while( head ) */

	return head;
}



__MODULEBEGIN__( "lfndi1.c" )
/* listFindInt1 */

#define __TEMPLATE__
#define F_listFindInt listFindInt1
#define T_value _integer1
#include <coralsrc/list.c>

__MODULEEND__



__MODULEBEGIN__( "lfndi2.c" )
/* listFindInt2 */

#define __TEMPLATE__
#define F_listFindInt listFindInt2
#define T_value _integer2
#include <coralsrc/list.c>

__MODULEEND__



__MODULEBEGIN__( "lfndi4.c" )
/* listFindInt4 */

#define __TEMPLATE__
#define F_listFindInt listFindInt4
#define T_value _integer4
#include <coralsrc/list.c>

__MODULEEND__



__MODULEBEGIN__( "lfndstr.c" )
/* listFindStr
 2000.07.19 */

struct _LIST __F* __API
listFindStr
	P_B(( head, s, offset, flag ))
	P_1		struct _LIST __F* head
	P_N		const char __F* s
	P_N		int offset
	P_N		unsigned flag
	P_E
{
	struct _str_compare_class_str str;
	str.s = s;
	str.offset = offset;
	str.flag = flag;

	while( head ) {

		if( str_compare_class_str(head,&str) )
			break;

		/* next node */
		head = head->next;
	}

	return head;
}

__MODULEEND__



__MODULEBEGIN__( "lmaxi1" )
/* listMaxInt1 */

#define __TEMPLATE__
#define F_listMaxInt listMaxInt1
#define T_value _integer1
#include <coralsrc/list.c>

__MODULEEND__



__MODULEBEGIN__( "lmaxi" )
/* listMaxInt */

#define __TEMPLATE__
#define F_listMaxInt listMaxInt2
#define T_value _integer2
#include <coralsrc/list.c>

__MODULEEND__



__MODULEBEGIN__( "lmaxi" )
/* listMaxInt */

#define __TEMPLATE__
#define F_listMaxInt listMaxInt4
#define T_value _integer4
#include <coralsrc/list.c>

__MODULEEND__



/*------------------------------------------------------------------------*
	 Circular List Iterations
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "lccount.c" )
/* listCNCount */

unsigned long __API
listCNCount
	P_B(( head ))
	P_1		struct _LIST __F* head
	P_E
{
	if( head ) {
		register unsigned long count;
		LIST_BEGIN_LINEAR(head);
		count = listCount( head );
		LIST_END_LINEAR(head);
		return count;
	}

	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "lcnode.c" )
/* listCNNode */

struct _LIST __F* __API
listCNNode
	P_B(( head, index ))
	P_1		struct _LIST __F* head
	P_N		long index
	P_E
{
	if( head ) {
		register struct _LIST __F* node;
		LIST_BEGIN_LINEAR(head);
		node = listNode( head, index );
		LIST_END_LINEAR(head);
		return node;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcindex.c" )

long __API
listCNIndex
	P_B(( head, node ))
	P_1		struct _LIST __F* head
	P_N		struct _LIST __F* node
	P_E
{
	if( head ) {
		long index;
		LIST_BEGIN_LINEAR(head);
		index = listIndex( head, node );
		LIST_END_LINEAR(head);
		return index;
	}
	return -1;
}
__MODULEEND__



__MODULEBEGIN__( "lcinsnod.c" )

struct _LIST __F* __API
listCNInsertionNode
	P_B(( head, index, flag ))
	P_1		struct _LIST __F* head
	P_N		long index
	P_N		unsigned __F* flag
	P_E
{
	if( head ) {
		register struct _LIST __F* node;

		if( index<0 ) {
			*flag = LIST_FLAG_ADD_NEXT;
			return head->prev;
		}

		LIST_BEGIN_LINEAR(head);
		node = listInsertionNode( head, index, flag );
		LIST_END_LINEAR(head);
		return node;
	}
	*flag = LIST_FLAG_ADD_PREV;
	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcexist.c" )
/* listCNExist */

_bool __API
listCNExist
	P_B(( head, node ))
	P_1		struct _LIST __F* head
	P_N		struct _LIST __F* node
	P_E
{
	if( head ) {
		register _bool exist;
		LIST_BEGIN_LINEAR(head);
		exist = listExist( head, node );
		LIST_END_LINEAR(head);
		return exist;
	}

	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "lcfndbit.c" )
/* listCNFindBit */

struct _LIST __F* __API
listCNFindBit
	P_B(( head, mask, offset, size, flag ))
	P_1		struct _LIST __F* head
	P_N		unsigned long mask
	P_N		int offset
	P_N		unsigned size
	P_N		unsigned flag
	P_E
{
	if( head ) {
		register struct _LIST __F* found;
		LIST_BEGIN_LINEAR( head );
		found = listFindBit( head, mask, offset, size, flag );
		LIST_END_LINEAR( head );
		return found;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcfndi1.c" )
/* listCNFindInt1 */

struct _LIST __F* __API
listCNFindInt1
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		_integer1 value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* found;
		LIST_BEGIN_LINEAR( head );
		found = listFindShort( head, value, offset );
		LIST_END_LINEAR( head );
		return found;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcfndi2.c" )
/* listCNFindInt2 */

struct _LIST __F* __API
listCNFindInt2
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		_integer2 value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* found;
		LIST_BEGIN_LINEAR( head );
		found = listFindLong( head, value, offset );
		LIST_END_LINEAR( head );
		return found;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcfndi4.c" )
/* listCNFindInt4 */

struct _LIST __F* __API
listCNFindInt4
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		_integer4 value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* found;
		LIST_BEGIN_LINEAR( head );
		found = listFindLong( head, value, offset );
		LIST_END_LINEAR( head );
		return found;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcfndstr.c" )

/* listCNFindStr */

struct _LIST __F* __API
listCNFindStr
	P_B(( head, s, offset, flag ))
	P_1		struct _LIST __F* head
	P_N		const char __F* s
	P_N		int offset
	P_N		unsigned flag
	P_E
{
	if( head ) {
		struct _LIST __F* found;
		LIST_BEGIN_LINEAR( head );
		found = listFindStr( head, s, offset, flag );
		LIST_END_LINEAR( head );
		return found;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcmaxi1.c" )
/* listCNMaxInt1 */

struct _LIST __F* __API
listCNMaxInt1
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		_integer1 __F* value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* maxnode;
		LIST_BEGIN_LINEAR( head );
		maxnode = listMaxInt1( head, value, offset );
		LIST_END_LINEAR( head );
		return maxnode;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcmaxi2.c" )
/* listCNMaxInt2 */

struct _LIST __F* __API
listCNMaxInt2
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		_integer2 __F* value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* maxnode;
		LIST_BEGIN_LINEAR( head );
		maxnode = listMaxInt2( head, value, offset );
		LIST_END_LINEAR( head );
		return maxnode;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcmaxi4.c" )
/* listCNMaxInt4 */

struct _LIST __F* __API
listCNMaxInt4
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		_integer4 __F* value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* maxnode;
		head->prev->next = NULL;
		maxnode = listMaxInt4( head, value, offset );
		head->prev->next = head;
		return maxnode;
	}

	return NULL;
}
__MODULEEND__



/*------------------------------------------------------------------------*
	Architecture specific routines
 *------------------------------------------------------------------------*/



#if __ARCH__==64



__MODULEBEGIN__( "lfindl.c" )
/* listFindLong */

#define __TEMPLATE__
#define F_listFindInt listFindLong
#define T_value long
#include <coralsrc/list.c>

__MODULEEND__



__MODULEBEGIN__( "lmaxl.c" )
/* listMaxLong */

#define __TEMPLATE__
#define F_listMaxInt listMaxLong
#define T_value long
#include <coralsrc/list.c>



__MODULEBEGIN__( "lcfindl" )
/* listCNFindLong */

struct _LIST __F* __API
listCNFindLong
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		long value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* found;
		LIST_BEGIN_LINEAR( head );
		found = listFindLong( head, value, offset );
		LIST_END_LINEAR( head );
		return found;
	}

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "lcmaxl.c" )
/* listCNMaxLong */

struct _LIST __F* __API
listCNMaxLong
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		long __F* value
	P_N		int offset
	P_E
{
	if( head ) {
		register struct _LIST __F* maxnode;
		LIST_BEGIN_LINEAR( head );
		maxnode = listMaxLong( head, value, offset );
		LIST_END_LINEAR( head );
		return maxnode;
	}

	return NULL;
}
__MODULEEND__



#endif /* __ARCH__==64 */



#else /* __TEMPLATE__ */



/*------------------------------------------------------------------------*
	Templates
 *------------------------------------------------------------------------*/



#ifdef F_listFindInt
/* listFindInt
 Template parameters
 F_listFindInt : the function name
 T_value : the value type  */

struct _LIST __F* __API
F_listFindInt
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		T_value value
	P_N		int offset
	P_E
{
	while( head ) {

		if( *(T_value __F*)( (_integer1 __F*)head + offset )==value )
			break;

		/* the next node */
		head = head->next;
	}
	return head;
}

#undef F_listFindInt
#undef T_value
#endif /* F_listFindInt */



#ifdef F_listMaxInt
/* listMaxInt
 Template parameters
 F_listMaxInt : the function name
 T_value : the value type  */

struct _LIST __F* __API
F_listMaxInt
	P_B(( head, value, offset ))
	P_1		struct _LIST __F* head
	P_N		T_value __F* value
	P_N		int offset
	P_E
{
	/* following initialization is moved here to avoid unintialized usage
	 warning for gcc - actually uninitialized case is never occurred
	 but compiler can generate that warning */
	struct _LIST __F* maxnode = NULL;
	T_value maxval;

	if( head ) {
		/* code reduction not to write duplicated calculation of offset-
		 translated pointer for initial value of maxid, just use arbitrary
		 value of integer ( minimum value ) - for the first initialization of
		 maxnode, evaluation of id==maxid is ( must be ) considered and
		 the first iteration step always initializes maxnode, because any
		 integer value in head is greater than or equal to OP_MININT(T_value)
		 : reviewed on 1998.03, 2001.03.27 */
		maxval = OP_MININT(T_value);
		while( head ) {
			T_value current = *(T_value __F*)((_integer1 __F*)head+offset);
			if(  current>=maxval ) {
				maxnode = head,
				maxval = current;
			}
			head = head->next;
		}
	}
	else {
		/* never consider to return OP_MININT(T_value) for failure case but zero must
			be returned for valid index management for ML solid modeler ( 1998.04.30 ) */
		maxval = 0;
	}

	if( value )
		*value = maxval;

	return maxnode;
}
__MODULEEND__

#undef F_listMaxInt
#undef T_value
#endif /* F_listMaxInt */



#undef __TEMPLATE__



#endif /* !__TEMPLATE__ */



/* end of list.c */

