/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Coral Memory Operations

 1996.01.   Linked list generalization
 1997.01.   Portable memory management functions
 1998.03.02 Reviewed
 2001.05.21 Reviewed for LGPL release

	Some redefinition of memory management functions is needed to support
 variable platforms. Some platform doesn't have functions such memCopy or
 memSet ( Especially 16 bit platforms : standard c functions like memcpy
 and memset can dispatch only 64KB memory blocks at a time. ) */



#ifdef __DEBUG__
#	define _MEM_DEBUG
#endif


#undef __DEBUG__
/* #define __DEBUG__ */


#include <coralsrc/list.h>
#include <coralsrc/memory.h>
#include <coralsrc/bintree.h>
#include <coral/inline.h>
#include <coral/compat/string.h>



/*------------------------------------------------------------------------*

	 Platform dependent memory routines

 *------------------------------------------------------------------------*/



#ifdef __WIN16__



__MODULEBEGIN__( "mwiter.c" )
/* mem_win_iter

 1997.10.
 Long iteration for huge memory on WIN16.
 See memLongIter to get some comments. */

long __API mem_win_iter
P_B(( cp, size ))
P_1  void __F*__F* cp
P_N  long __F* size
P_E
{
	long toread;

	/* instead of OFFSETOF, OP_LOW4 is used because they are equivalent */
	*size -= ( toread = 65536L-OP_LOW4(*cp) );
	if( *size<0 ) {
		toread += *size;
	}
	else {
		((_unsigned2 __F*)cp)[1]+=8; /* new selector */
		((_unsigned2 __F*)cp)[0]=0;  /* new offset */
	}

	return toread;
}
__MODULEEND__



#endif /* __WIN16__ */



__MODULEBEGIN__( "mliter.c" )
/* mem_long_iter

 16 bit platforms oriented  but needed for 32 bit platform to maintain
 ioEnhance long iterator for memory opeartion enhancement.
 1997. 3

 * declare current pointer(cp) and initialize it with destination pointer.
 cp = p

 * iteration constraint will be
	while( size>0 )
	{
	 ...
	}
 'size' is requested size.

 * now execution of memLongIter with requested size 'size' and dumping unit
 'toread' will returns valid operation size.

 opsize = memLongIter( &cp, &size, toread );

 * memLongIter shifts current pointer and stores it to 'cp'. So,
 p = cp

*/

_unsigned2 __API
mem_long_iter
P_B(( cp, size, toread ))
P_1  char __H* __F* cp
P_N  long __F* size
P_N  _unsigned2 toread
P_E
{
	/* win16 version is supported for sdk, especially to support ioEnhance :
	 memWinIter is more efficient for WIN16 2^16 threshold because following
	 code makes two iteration sequence over 65536 bytes operation, i.e.
	 65535 and 1 byte combination. */

#ifdef __WIN16__
	/* instead of OFFSETOF, OP_LOW4 is used because they are equivalent */
	if( 65536L-OP_LOW4(*cp) < toread )
		toread = (_unsigned2)(65536L-OP_LOW4(*cp));
#endif /* __WIN16__ */

	*size -= toread;
	if( *size<0 )
		toread+=(_unsigned2)*size;
	else
		*cp += toread;
	return toread;
}
__MODULEEND__



/*------------------------------------------------------------------------*

	Memory allocators

 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "mpalloc.c" )



/* memAlloc */



#ifndef mem_alloc

#ifdef _MSC_VER
static void __F* mem_alloc
P_B(( size ))
P_1  long size
P_E
{
	long __F* p;

	if( size>OP_MAXINT(size_t) ) {
		p = (long __F*)halloc( size, 1 );
		if( !p ) return NULL;
		p[0] = size; /* halloc is used */
	}
	else {
		p = (long __F*)_fmalloc( size );
		if( !p ) return NULL;
		p[0] = -1; /* _fmalloc is used */
	}

	return &p[1];
}
#else
/* #error */ @error "mem_alloc was not defined"
#endif /* _MSC_VER */

#endif /* mem_alloc */



#ifdef _MEM_DEBUG

struct _MEMDEBUG __F* mem_root;
_memmagic             mem_magic=1, __F* mem_magic_table;
unsigned mem_magic_count;
void (__API* mem_magic_callback) P_A(( _memmagic found ));


static _memmagic
mem_disable_debug P_A(( void ))
{
	_memmagic magic = mem_magic;
	if( magic==OP_MAXUINT(_memmagic) ) return 0;
	mem_magic = 0;
	return magic;
}


static int __API mem_compare_magic P_A(( const void __F* p, const void __F* key ));
static int __API
mem_compare_magic( p, key )
	const void __F* p;
	const void __F* key;
{
	_memmagic a = *(_memmagic __F*)p, b = *(_memmagic __F*)key;
	return a==b? 0 : a>b? 1: -1;
}



static void mem_add_debug P_A(( void __F* p, _memmagic magic, unsigned long size ));

static void
mem_add_debug( p, magic, size )
	void __F* p;
	_memmagic magic;
	unsigned long size;
{
	if( p ) {
		struct _MEMDEBUG __F* d;

		if( mem_magic_table ) {
			_memmagic __F* found = (_memmagic __F*)memBinarySearch
				( mem_magic_table, mem_magic_count,
					sizeof(_memmagic), &magic, mem_compare_magic, NULL );
			if( found ) mem_magic_callback( *found );
		}

		d = (struct _MEMDEBUG __F*)memGetExtra( p );
		listDNAdd( (struct _LIST __F*__F*)&mem_root,
			NULL, (struct _LIST __F*)d,
			LIST_FLAG_CIRCULAR|LIST_FLAG_ADD_PREV|LIST_FLAG_QUEUE );
		d->magic = magic++;
		d->size = size;
	}
	mem_magic = magic;
}



/* memDebug */

void __API
memEnableDebug P_A(( void ))
{
	if( mem_magic==0 ) mem_magic = 1;
}



/* memGetDebug */

struct _MEMDEBUG __F* __API
memGetDebug( struct _MEMDEBUG __F* current )
{
	if( current ) {
		current = current->next;
		return current==mem_root ? NULL : current;
	}
	return mem_root;
}



/* memGetMagic */

unsigned long __API
memGetMagic P_A(( void ))
{
	return mem_magic;
}



/* memSetMagic */

void __API
memSetMagic( table, n, callback )
	_memmagic __F* table;
	unsigned n;
	void (__API* callback) P_A(( _memmagic found ));
{
	mem_magic_table = table;
	mem_magic_count = n;
	mem_magic_callback = callback;
}



#endif /* _MEM_DEBUG */



void __F* __API
memAlloc( size )
	long size;
{

#ifdef _MEM_DEBUG

	if( mem_magic>0 ) {
		void __F* p;
		_memmagic magic = mem_disable_debug();
		if( magic==0 ) return NULL;
		p = memAllocExtra( size, sizeof(struct _MEMDEBUG) );
		mem_add_debug( p, magic, size );
		return p;
	}

#endif /* _MEM_DEBUG */

	return mem_alloc( size );

}

__MODULEEND__



__MODULEBEGIN__( "mresize.c" )
/* memResize */



#ifndef mem_realloc

static void __F* mem_realloc
P_B(( p, long size ))
P_1  void __F* p
P_N  long size
P_E
{
	long __F* newp=NULL;

	if( p ) {

		/* the standard spec. for realloc frees the pointer if size==0 */
		if( size>0 ) {

#ifdef __PALMOS__
			/* try to reduce or expand the memory size for the fixed pointer */
			if( MemPtrResize( p, size )==0 ) {
				return p;
			}
# else /* !__PALMOS__ */
# ifdef _MSC_VER
			long oldsize = *( (long __F*)p - 1 );
			if( oldsize<0 && size<=65535L ) {
				/* old pointer was allocated by _fmalloc and new size is smaller
					 than 64 kb */
				newp = _frealloc( (long __F*)p - 1, (size_t)size );
				if( newp )
					newp[0] = -1;
				return newp;
			}
# endif /* _MSC_VER */
# endif /* __PALMOS */
			else {

				newp = (long __F*)mem_alloc( size );

				if( newp ) {
# ifdef __PALMOS__
					long oldsize = MemPtrSize( p );
# endif
					/* reduce the size to be copied if the new memory block
						 cannot accept all of the old contents */
					if( oldsize>size ) oldsize = size;

					memCopy( newp, p, oldsize );

				} /* end if( newp ) */
				else {
					/* if allocation was failed, the given memory block is
						 unchanged */
					return NULL;
				}
			}

		} /* end if( size>0 ) */

		mem_free( p );

	} /* end if( p ) */

	return newp;
}

#endif /* mem_realloc */



void __F* __API memResize
P_B(( p, size ))
P_1  void __F* p
P_N  long size
P_E
{
#ifdef _MEM_DEBUG

	if( p==NULL ) return memAlloc( size );
	if( mem_magic>0 ) {
		_memmagic magic = mem_disable_debug();
		if( magic==0 ) return NULL;
		listDNCut( (struct _LIST __F*__F*)&mem_root,
			(struct _LIST __F*)memGetExtra(p) );
		p = memResizeExtra( p, size );
		mem_add_debug( p, magic, size );
		return p;
	}

#endif

	return mem_realloc( p, size );

}

__MODULEEND__



__MODULEBEGIN__( "mpcalloc.c" )
/* memCAlloc */

void __F* __API memCAlloc
P_B(( size ))
P_1  long size
P_E
{
	void __F* p = memAlloc( size );
	if( p )
		memSet( p, 0, size );
	return p;
}
__MODULEEND__



__MODULEBEGIN__( "mpfree.c" )
/* memFree */

#ifndef mem_free

#ifdef _MSC_VER
static void mem_free
P_B(( p ))
P_1  void __F* p
P_E
{
	if( p ) {
		long size = *(long __F*) ( p = (long __F*)p - 1 );
		if( size>=0 )
			hfree( p );
		else
			_ffree( p );
	} /* end if( p ) */
}
#else /* !_MSC_VER */

/* #error */ @error "mem_free was not defined"

#endif

#endif /* mem_free */



void __API memFree
P_B(( p ))
P_1  void __F* p
P_E
{

#ifdef _MEM_DEBUG

/*	if( p ) {
		long __F* actual = (long __F*) ( (_integer1 __F*)p - _DEBUG_MEM_CLEAR );
		memSet( p, 0, *actual );
		mem_free( actual );
	}*/
	if( p==NULL ) return;
	if( mem_magic>0 ) {
		struct _MEMDEBUG __F* d = (struct _MEMDEBUG __F*)memGetExtra( p );
		listDNCut( (struct _LIST __F*__F*)&mem_root, (struct _LIST __F*)d );
		memSet( p, 0xcd, d->size );
		mem_free( d );
		return;
	}
#endif

	mem_free( p );

}

__MODULEEND__



/*---------------------------------------------------------------------------*
	Extra memory
 *---------------------------------------------------------------------------*/



__MODULEBEGIN__( "mextra.c" )



/* mem_get_extra_size */

static unsigned mem_get_extra_size P_A(( void __F* p ));

static unsigned
mem_get_extra_size( p )
	void __F* p;
{
	unsigned extra;
	mem_short_copy( &extra,
		(_integer1 __F*)p-sizeof(unsigned), sizeof(unsigned) );
	return extra;
}



/* mem_init_extra */

static void __F* mem_init_extra P_A(( _integer1 __F* m, unsigned aligned ));

static void __F*
mem_init_extra( m, aligned )
	_integer1 __F* m;
	unsigned aligned;
{
	if( m ) {
		m += aligned;
		mem_short_copy( m-sizeof(unsigned), &aligned, sizeof(unsigned) );
	}
	return m;
}



/* memGetExtraSize */

unsigned __API
memGetExtraSize
P_B(( extra ))
P_1  unsigned extra
P_E
{
	return memAlign(extra+sizeof(unsigned),0);
}



/* memAllocExtra */

void __F* __API
memAllocExtra
P_B(( size, extrabytes ))
P_1  long size
P_N  unsigned extrabytes
P_E
{
	unsigned aligned = memGetExtraSize(extrabytes);
	return mem_init_extra
		( (_integer1 __F*)memAlloc(size+aligned), aligned );
}




/* memResizeExtra */

void __F* __API
memResizeExtra( p, size )
	void __F* p;
	long size;
{
	unsigned aligned = mem_get_extra_size(p);
	return mem_init_extra
		( memResize(memGetExtra(p),size+aligned), aligned );
}



/* memFreeExtra */

void __API
memFreeExtra
P_B(( p ))
P_1  void __F* p
P_E
{
	if( p ) memFree( memGetExtra(p) );
}



/* memGetExtra */

void __F* __API
memGetExtra
P_B(( p ))
P_1  void __F* p
P_E
{
	return (void __F*)( (_integer1 __F*)p - mem_get_extra_size(p) );
}

__MODULEEND__




/*---------------------------------------------------------------------------*
	Named memory
 *---------------------------------------------------------------------------*/

/* Named memory functions are not thread-safe */



__MODULEBEGIN__( "memname.c" )



#define _MEM_NAME_ITERPATH_COUNT 16 /* default bintree iterpath count */



static struct _mem_name __F* mem_name; /* the root of named memories */



/* mem_compare_name */

static int __API
mem_compare_name( struct _BINTREE __F* node, void __F* key )
{
	return strcmp( ((struct _mem_name __F*)node)->name, (const char __F*)key );
}



/* mem_get_name_extra */

static unsigned
mem_get_name_extra
P_B(( name ))
P_1  const char __F* name
P_E
{
	return sizeof(struct _mem_name) + strlen(name);
}



/* mem_init_name_iterpath */

static void
mem_init_name_iterpath
P_B(( iterpath, iteritem ))
P_1  struct _BINTREEITERPATH __F* iterpath
P_N  struct _BINTREEITERITEM __F* iteritem
P_E
{
	iterpath->path = iteritem;
	iterpath->last = _MEM_NAME_ITERPATH_COUNT-1;
	iterpath->flag = BINTREE_FLAG_ALLOW_REALLOC|BINTREE_FLAG_BALANCED;
}



/* memAllocName */

void __F* __API
memAllocName
P_B(( size, name, status ))
P_1  long size
P_N  const char __F* name
P_N  int __F* status
P_E
{
	int result = SUCCESS_EXIST;
	void __F* ptr = NULL;

	if( size==0 ) {
		ptr = memGetNamePtr( name );
		if( ptr )
			/* increase usage counter */
			((struct _mem_name __F*)memGetExtra( ptr ))->count++;
		else
			result = FAIL_ENTRY;
	}
	else {
		struct _mem_name __F* newname;
		struct _BINTREEITERITEM iteritem[_MEM_NAME_ITERPATH_COUNT];
		struct _BINTREEITERPATH iterpath;
		mem_init_name_iterpath( &iterpath, iteritem );

		switch( bintreeIterate(
					(struct _BINTREE __F*__F*)&newname,
					(struct _BINTREE __F*)mem_name,
					(void __F*)name, mem_compare_name, &iterpath ) ) {
			case 0:
			newname->count++;
			ptr = (_integer1 __F*)newname +
				memGetExtraSize(mem_get_name_extra(newname->name));
			break;

			case FAIL_BUFFER:
			result = FAIL_BUFFER;
			break;

			default:
			result = 0;
			ptr = memAllocExtra( size, mem_get_name_extra(name) );
			if( ptr ) {
				newname = (struct _mem_name __F*)memGetExtra(ptr);
				newname->count = 1;
				strcpy( newname->name, name );
				bintreeLinkAVL(
					(struct _BINTREE __F*__F*)&mem_name,
					(struct _BINTREE __F*)newname,
					&iterpath );
			}
		}

		if( iterpath.flag&BINTREE_FLAG_ALLOCATED )
			memFree( iterpath.path );
	}

	if( status ) *status = result;

	return ptr;
}



void __API
memFreeName( const char __F* name )
{
	void __F* p = memGetNamePtr(name);
	if( p ) {
		struct _BINTREEITERITEM iteritem[_MEM_NAME_ITERPATH_COUNT];
		struct _BINTREEITERPATH iterpath;

		if( OP_FINALIZE
				( ((struct _mem_name __F*)memGetExtra(p))->count ) ) {
			mem_init_name_iterpath( &iterpath, iteritem );

			bintreeRemove(
				(struct _BINTREE __F*__F*)&mem_name,
				(void __F*)name,
				mem_compare_name,
				&iterpath );
		}
	}
}



void __F* __API
memGetNamePtr
P_B(( name ))
P_1  const char __F* name
P_E
{
	_integer1 __F* p = (_integer1 __F*) bintreeSearch
		( (struct _BINTREE __F*)mem_name, (void __F*)name, mem_compare_name );
	return p ? p+memGetExtraSize(mem_get_name_extra(name)) : NULL;
}



const char __F* __API
memGetName
P_B(( p ))
P_1  void __F* p
P_E
{
	return ((struct _mem_name __F*)memGetExtra(p))->name;
}



__MODULEEND__



/*---------------------------------------------------------------------------*
		Memory Operators
 *---------------------------------------------------------------------------*/



/* memSet */


#ifndef memSet

__MODULEBEGIN__( "mset.c" )
/* memSet
 Written for win16 and dos16 */

void __API memSet
P_B(( dest, c, n ))
P_1  void __F* dest
P_N  _unsigned1 c
P_N  long n
P_E
{
	void __F* cp = dest;

#ifdef __WIN16__

	while( n>0 ) {
		long opsize = mem_win_iter( &cp, &n );
		if( opsize==65536L ) {
			*(_integer1 __F*)dest = c;
			mem_short_set( (void __F*)((_unsigned4)dest+1), c, 65535U );
		}
		else {
			mem_short_set( dest, c, (_unsigned2)opsize );
		}

		dest = cp;
	}

#else /* !__WIN16__ */

	while( size>0 ) {
		mem_short_set( dest, val,
		 mem_long_iter((char __H*__F*)&cp,&size,_MEM_LONGITER) );
		dest = cp;
	}

#endif /* __WIN16__ */

}
__MODULEEND__

#endif /* memSet */



/* memSetExt

param : an array which contains min cache counter, max cache counter, threshold counter */

void __API
memSetExt
P_B(( p, n, size, param ))
P_1  void __F* p
P_N  unsigned long n
P_N  unsigned size
P_N  unsigned __F* param
P_E
{
	unsigned long totalsize, cache, iter, i;

	register union {
		_integer1 i1;
		_integer2 i2;
		_integer4 i4;
		_integer8 i8;
	} value;

	register union {
		_integer2 __H* p2;
		_integer4 __H* p4;
		_integer8 __H* p8;
	} ptr;

	char __H* dest;

	if( size<=1 ) { /* cache is not needed - very simple */
		memSet( p, *(_integer1 __F*)p, n );
		return;
	}

	/* evaluate cache size */
	totalsize = n * size;
	cache = ( param? totalsize<param[2]? param[0] : param[1] : totalsize<32768U? 4096U : 16384U ) / size;

	iter = n/cache;
	if( n%cache>0 ) iter++;

	i = n/iter; /* actual cache count */
	cache = i*size; /* actual cache size */

	switch( size ) {

		case 2:
		value.i2 = *(_integer2 __F*)p;
		ptr.p2 = (_integer2 __H*)p;
		while( i-->0 ) *ptr.p2++ = value.i2;
		break;

		case 4:
		value.i4 = *(_integer4 __F*)p;
		ptr.p4 = (_integer4 __H*)p;
		while( i-->0 ) *ptr.p4++ = value.i4;
		break;

		case 8:
//XXX generalized _integer8 op routines
#ifdef __INTEGER8__
		value.i8 = *(_integer8 __F*)p;
		ptr.p8 = (_integer8 __H*)p;
		while( i-->0 ) *ptr.p8++ = value.i8;
#else
		value.i8[0] = ((*_integer4 __F*)val)[0];
		value.i8[1] = ((*_integer4 __F*)val)[1];
		ptr.p4 = (_integer4 __H*)p;
		while( i-->0 ) {
			*ptr.p4++ = value.i8[0];
			*ptr.p4++ = value.i8[1];
		}
#endif
		break;

		default: return; /* unavailable yet */
	}

	if( iter>1 ) {
		dest = (char __H*)p;
		totalsize -= cache;
		while( 1 ) {
			dest += cache;
			if( totalsize<=cache ) {
				memCopy( dest, p, totalsize );
				break;
			}
			memCopy( dest, p, cache );
			totalsize -= cache;
		}
	}
}



#ifndef memCopy

__MODULEBEGIN__( "mcopy.c" )
/* memCopy
 Written for dos16 */

void __API memCopy
P_B(( dest, src, size ))
P_1  void __F* dest
P_N  void __F* src
P_N  long size
P_E
{
	void __F* cp = src;
	while( size>0 ) {
		mem_short_copy( dest, src,
		 mem_long_iter((char __H*__F*)&cp,&size,_MEM_LONGITER) );
		src = cp,
		dest = (void __F*)( (char __H*)dest + _MEM_LONGITER );
	}
}
__MODULEEND__

#endif /* memCopy */



__MODULEBEGIN__( "mcvcopy.c" )
/* memConvCopy
 Conversion copy */

void __API memConvCopy
P_B(( destptr, srcptr, convert, n, destsize, srcsize ))
P_1  void __F* destptr
P_N  void __F* srcptr
P_N  void (__API* convert) P_A(( void __F* dest, void __F* src ))
P_N  long n
P_N  unsigned destsize
P_N  unsigned srcsize
P_E
{
	long i;
	char __H* dest = (char __H*)destptr;
	char __H* src = (char __H*)srcptr;

	for( i=0; i<n; i++ ) {
		/* convert an item */
		convert( dest, src );

		/* translate to the next item */
		dest+=destsize,
		src +=srcsize;
	}
}
__MODULEEND__



__MODULEBEGIN__( "minv.c" )
/* memInvert

	This method inverts the memory contents byte by byte. Available
 8086 instruction to optimize this method is xchg. */

void __API memInvert
P_B(( p, size ))
P_1  void __F* p
P_N  unsigned size
P_E
{
	register _unsigned1 __F* f=(_unsigned1 __F*)p, __F* b=f+size-1;
	size>>=1;
	while( size-- ) {
		register _unsigned1 temp = *f;
		*f++ = *b,
		*b-- = temp;
	}
}
__MODULEEND__



__MODULEBEGIN__( "mbinv.c" )
/* memInvertB

	B ( byte ) type inversion. This method inverts elements itself in an array
 byte by byte. */

void __API memInvertB
P_B(( src, n, unitsize ))
P_1  void __F* src
P_N  long n
P_N  unsigned unitsize
P_E
{
	register _integer1 __F* p = (_integer1 __F*)src;
	while( n-- ) {
		memInvert( p, unitsize );
		p+=unitsize;
	}
}

__MODULEEND__



__MODULEBEGIN__( "mainv.c" )
/* memInvertA

	A ( array ) type inversion. This method inverts the array element
 by element. */

void __API memInvertA
P_B(( p, unitbuf, n, unitsize ))
P_1  void __F* p
P_N  void __F* unitbuf
P_N  long n
P_N  unsigned unitsize
P_E
{
	register _unsigned1
	 __H* f=(_unsigned1 __H*)p,
	 __H* b=f+(n-1)*unitsize;

	n>>=1;
	while( n-- ) {
		memCopy( unitbuf, (void __F*)f, unitsize );
		memCopy( (void __F*)f, (void __F*)b, unitsize );
		memCopy( (void __F*)b, unitbuf, unitsize );
		f+=unitsize,
		b-=unitsize;
	}
}
__MODULEEND__



__MODULEBEGIN__( "mconv.c" )



/* lcm(dest_nbit,8)/dest_nbit */
static const char mem_src_align[] = { 0, 8, 4, 8, 2, 8, 4, 8, 1 };



/* memSourceTable */

void __API
memSourceTable( char __F* src_table, const char* dest_table, unsigned n )
{
	unsigned i;
	memset( src_table, MEM_SPECIAL_CHAR, 256 );
	for( i=0; i<n; i++ )
		src_table[(unsigned char)dest_table[i]] = i;
}



/* memSourceAlignSize */


unsigned __API
memSourceAlignSize
	P_B(( src_nbit ))
	P_1		unsigned nbit
	P_E
{
	return mem_src_align[nbit&0xff];
}



/* memDestBufferSize */

unsigned __API
memDestBufferSize
	P_B(( dest_nbit, src_nbit, len ))
	P_1		unsigned dest_nbit
	P_N		unsigned src_nbit
	P_N		unsigned len
	P_E
{
	unsigned align = mem_src_align[dest_nbit];
	return OP_ALIGN(OP_BYTE(len*src_nbit,dest_nbit),align);
}



__MODULEBEGIN__( "mconv.c" )
/* memConvert */

int __API
memConvert
	P_B(( dest, src, len, dest_table, src_table, dest_nbit, src_nbit, flag ))
	P_1		unsigned char __F* dest
	P_N		const unsigned char __F* src
	P_N		unsigned __F* len
	P_N		const unsigned char __F* dest_table
	P_N		const unsigned char __F* src_table
	P_N		unsigned dest_nbit
	P_N		unsigned src_nbit
	P_N		unsigned flag
	P_E
{
	static const unsigned char mask[] =
		{ 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f };

	unsigned src_len = *len;
	if( src_len>0 ) {
		unsigned dest_bit, src_bit, dest_full, dest_empty, src_full, src_empty;
		unsigned  l=0;
		unsigned char dest_char=0, src_char=0; /* actually src_char need not to be initialized */
		if( flag&MEM_FLAG_LOW_TO_HIGH ) {
			dest_full = dest_nbit;
			dest_empty = 0;
			src_full = 0;
			src_empty = src_nbit;
		}
		else {
			dest_full = 0;
			dest_empty = dest_nbit;
			src_full = src_nbit;
			src_empty = 0;
		}

		dest_bit = dest_empty;
		src_bit = src_empty;

		while( 1 ) {
			unsigned needed, dest_shift, src_shift;

			if( src_bit==src_empty ) { /* no remained bits in the source */
				src_char = *src;
				if( src_table ) {
					src_char = src_table[src_char];
					if( src_char==MEM_SPECIAL_CHAR ) {

						if( flag&MEM_FLAG_SPECIAL_CHAR_INDEX )
							len[1] = *len - src_len;

						*len = l;
						return MEM_SPECIAL_CHAR_FOUND | *src;
					}
				}
				src++;
				src_len--;
				src_bit = src_full;
			}


			if( flag&MEM_FLAG_LOW_TO_HIGH ) {
				unsigned dest_remained = dest_full - dest_bit;
				unsigned src_remained = src_empty - src_bit;
				needed = OP_MIN(dest_remained,src_remained);
				dest_shift = dest_bit;
				dest_bit += needed;
				src_shift = src_bit;
				src_bit += needed; /* ... (1) */
			}
			else {
				needed = OP_MIN(dest_bit,src_bit);
				dest_shift = dest_bit -= needed;
				src_shift = src_bit -= needed; /* ... (2) */
			}

			while( 1 ) {
				dest_char |= ((src_char>>src_shift)&mask[needed]) << dest_shift;

				if( dest_bit==dest_full ) { /* all destination bits were initialized */
					*dest++ = dest_table?
						dest_table[dest_char] : dest_char;
					dest_char = 0;
					l++;
					if( src_len<=0 ) {
						if( src_bit!=src_empty
						    && !(flag&MEM_FLAG_ALLOW_REMAINED) ) {

							if( flag&MEM_FLAG_LOW_TO_HIGH ) {
								needed = src_empty - src_bit;
								dest_shift = 0;
								/* in the routine (1), 'src_shift' is updated
									before 'src_bit' is progressed, so update
									'src_shift' here */
								src_shift = src_bit; 
							}
							else {
								needed = src_bit;
								dest_shift = dest_nbit - src_bit;
								/* in the routine (2), 'src_shift' is updated
									after 'src_bit' is progressed, so it is
									not needed to update 'src_shift */
							}
							src_bit = src_empty;
							continue;
						}
						if( flag&MEM_FLAG_ALIGN ) {
							unsigned align;
							dest_char = (char)( flag&0xff );
							align = mem_src_align[dest_nbit];
							src_len = l % align;
							if( src_len>0 ) {
								src_len = align - src_len;
								l += src_len;
								while( src_len-->0 )
									*dest++ = dest_char;
							}
						}
						*len = l;
						return 0;
					}
					dest_bit = dest_empty;
				} /* end if( destbit==destfull ) */

				if( src_len<=0 ) {
					dest_bit = dest_full;
					continue;
				}
				break;
			} /* end while( 1 ) */
		} /* end while( 1 ) */
	}
	return 0;
}


#if 0
int __API
memConvert
	P_B(( dest, src, len, dest_table, src_table, dest_nbit, src_nbit, flag ))
	P_1		unsigned char __F* dest
	P_N		const unsigned char __F* src
	P_N		unsigned __F* len
	P_N		const unsigned char __F* dest_table
	P_N		const unsigned char __F* src_table
	P_N		unsigned dest_nbit
	P_N		unsigned src_nbit
	P_N		unsigned flag
	P_E
{
	static const unsigned char mask[] =
		{ 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f };

	unsigned src_len = *len;
	if( src_len>0 ) {
		unsigned char dest_char=0, src_char;
		unsigned dest_bit=dest_nbit, src_bit=0, l=0;

		while( src_len>0 ) {
			unsigned needed, shift;

			if( src_bit<=0 ) { /* no remained bits in the source */
				src_char = *src;
				if( src_table ) {
					src_char = src_table[src_char];
					if( src_char==MEM_SPECIAL_CHAR ) {

						if( flag&MEM_FLAG_SPECIAL_CHAR_INDEX )
							len[1] = *len - src_len;

						*len = l;
						return MEM_SPECIAL_CHAR_FOUND | *src;
					}
				}
				src++;
				src_len--;
				src_bit = src_nbit;
			}

			needed = OP_MIN(dest_bit,src_bit);
			shift = dest_bit -= needed;
			src_bit -= needed;

			while( 1 ) {
				dest_char |= ((src_char>>src_bit)&mask[needed]) << shift;

				if( dest_bit<=0 ) { /* all destination bits were initialized */
					*dest++ = dest_table?
						dest_table[dest_char] : dest_char;
					dest_char = 0;
					l++;
					if( src_len<=0 ) {
						if( src_bit!=0 && !(flag&MEM_FLAG_ALLOW_REMAINED) ) {
							needed = src_bit;
							shift = dest_nbit - src_bit;
							src_bit = 0;
							continue;
						}
						if( flag&MEM_FLAG_ALIGN ) {
							unsigned align;
							dest_char = (char)( flag&0xff );
							align = mem_src_align[dest_nbit];
							src_len = l % align;
							if( src_len>0 ) {
								src_len = align - src_len;
								l += src_len;
								while( src_len-->0 )
									*dest++ = dest_char;
							}
						}
						*len = l;
						return 0;
					}
					dest_bit = dest_nbit;
				}
				break;
			} /* end while( 1 ) */
		} /* end while( src_len>0 ) */

		/* insufficient source buffer */

	}
	return FAIL_UNAVAILABLE;
}
#endif
__MODULEEND__



__MODULEBEGIN__( "malign.c" )
/* memAlign

	This function calculates aligned size of the given size using mrb value
 which means 'Most Restrictive Boundary'.  */

long __API memAlign
P_B(( size, mrb ))
P_1  long size
P_N  int mrb
P_E
{
	struct _alignl { _integer1 i1; _integer4 i4; };
	struct _alignd { _integer4 i4; double d; };
	struct _alignld { double d; long double ld; };
	if( mrb==0 ) {
		if( sizeof(struct _alignl)%4 !=0 ) return size; /* not aligned */
		if( sizeof(struct _alignd)%8 !=0 ) mrb = 4;
		else if( sizeof(struct _alignld)%sizeof(long double) !=0 ) mrb=sizeof(double);
		else mrb = sizeof(long double);
	}
	return ( size + (mrb-1) ) / mrb * mrb;
}

__MODULEEND__



__MODULEBEGIN__( "mborder.c" )
/* memByteOrder

 Gets machine byte order.
 Returns FLAG_LITTLEENDIAN or FLAG_BIGENDIAN */

unsigned __API memByteOrder
P_V
{
	register union { _unsigned4 u4; _unsigned1 u1[4]; } val;
	val.u4 = 0x01020304L;

	switch( val.u1[0] ) {
		case 4 :
		return FLAG_LITTLE_ENDIAN;

		case 1 :
		return FLAG_BIG_ENDIAN;
	}

	return FLAG_PDP_ENDIAN;
}
__MODULEEND__



__MODULEBEGIN__( "mborderh.c" )
/* memByteOrderHint

	Pre-evaluation of byte ordering flag. This function returns FLAG_DEFAULT
 or FLAG_EXCLUSIVE. Caution not to supply combinated size to 'flag' ! */

unsigned __API memByteOrderHint
P_B(( flag ))
P_1  unsigned flag
P_E
{
	register union { _unsigned4 u4; _unsigned1 u1[4]; } val;
	val.u4 = 0x01020304;

	switch( val.u1[0] ) {
		case 4: /* machine is little endian */
		return flag==FLAG_BIG_ENDIAN ? MEM_INVERT :
					 flag==FLAG_PDP_ENDIAN ? MEM_X1 : 0;

		case 1: /* machine is big endian */
		return flag==FLAG_BIG_ENDIAN ? 0 :
					 flag==FLAG_PDP_ENDIAN ? MEM_X2 : MEM_INVERT;
	}

	/* machine is pdp endian */
	return flag==FLAG_BIG_ENDIAN ? MEM_X2 :
				 flag==FLAG_PDP_ENDIAN ? 0 : MEM_X1;
}

__MODULEEND__



__MODULEBEGIN__( "mreorder.c" )
/* memReorder

	For valid byte ordering, 'src' must be an array of base classes.
 For example, the array of float. Not to be the array of float[3]
 and that case, 'n' need to be n*3 and 'basesize' need to be
 'basesize/3'. */

_bool __API memReorder
P_B(( src, n, byteorder ))
P_1  void __F* src
P_N  long n
P_N  unsigned byteorder
P_E
{
	unsigned size = byteorder&MASK_COMBINATED_SIZE;
	if( size>1 ) {
		switch( memByteOrderHint( byteorder&MASK_BYTE_ORDER ) ) {
			case MEM_INVERT:
			memInvertB( src, n, size );
			return 1;

			/* for pdp endians */
			case MEM_X1:
			return 1;
			case MEM_X2:
			return 1;
		}
	}
	return 0;
}

__MODULEEND__



/*
 *
 *  2D memory
 *
 */



__MODULEBEGIN__( "miter2.c" )
/* mem_iterate_2d
 calculates current column size of 2d memory instruction

 commented on 1999.07.12
	combinated size is the item size of given arrays, so currently it
 can be 1, 2, or 4 and maximum value is 128 */

long __API mem_iterate_2d
P_B(( itemsize, pcolrow, pcolcol, variable ))
P_1  unsigned itemsize
P_N  const void __F*__F* pcolrow
P_N  const void __F*__F* pcolcol
P_N  unsigned variable
P_E
{
	long total=1;
	unsigned short i=2;
	const unsigned colitemsize = variable&MASK_COMBINATED_SIZE;

	/* set first target to rows of columns */
	unsigned mask = FLAG_ROW_VARIABLE;
	const void __F*__F* pcol = pcolrow;

	while( i-- ) {
		register const void __F* col = *pcol;
		/* test 2d memory instruction was supplied for this column entry */
		if( col ) {
			long n;
			switch( colitemsize ) {
				case 1 :
				n = *(_integer1 __H*)col;
				break;

				case 2 :
				default:
				n = *(_integer2 __H*)col;
				break;

				case 4 :
				n = *(_integer4 __H*)col;
			}
			/* cumulate total size */
			total *= n;

			/* translate pointer if column size is variable */
			if( variable&mask )
				*pcol = (_integer1 __H*)col + colitemsize;
		}
		/* else : nothing to do, constant value for this column entry
			is regarded as 1 and need not to be multiplied */

		/* change target */
		pcol = pcolcol;
		mask = FLAG_COL_VARIABLE;
	}

	/* return current column size */
	return total * (long)itemsize;
}

__MODULEEND__



__MODULEBEGIN__( "msize2.c" )
/* memSize2D */

long __API memSize2D
P_B(( row, size, colrow, colcol, variable ))
P_1  long row
P_N  unsigned size
P_N  const void __F* colrow
P_N  const void __F* colcol
P_N  unsigned variable
P_E
{
	/* consider memor to store row pointers */
	long total = row * sizeof(void __F*);

	while( row-- )
		total += mem_iterate_2d( size, &colrow, &colcol, variable );

	return total;
}

__MODULEEND__



__MODULEBEGIN__( "minit2.c" )
/* memInit2D
 'column' parameter is reserved for memInitRect */

void __API memInit2D
P_B(( p, column, row, size, colrow, colcol, variable ))
P_1  void __F* p
P_N  void __F* column
P_N  long row
P_N  unsigned size
P_N  const void __F* colrow
P_N  const void __F* colcol
P_N  unsigned variable
P_E
{
	void __F* __H* pp = (void __F* __H*)p;
	/* reserve memory to fill row pointers if column memory block was
		not supplied */
	_integer1 __H* portion =
		column? (_integer1 __H*)column : (_integer1 __H*)p + row*sizeof(void __F*);

	while( row-- ) {
		*pp = (void __F*)portion;

		/* translate pointers */
		pp++;
		portion += mem_iterate_2d( size, &colrow, &colcol, variable );
	}
}

__MODULEEND__



__MODULEBEGIN__( "malloc2.c" )
/* memAlloc2D

	This function allocates a linear memory block and initialize pointers
 using its portions */

void __F* __API memAlloc2D
P_B(( row, size, colrow, colcol, variable ))
P_1  long row
P_N  unsigned size
P_N  const void __F* colrow
P_N  const void __F* colcol
P_N  unsigned variable
P_E
{
	void __F* p = memAlloc( memSize2D(row,size,colrow,colcol,variable) );

	if( p ) memInit2D( p, NULL, row, size, colrow, colcol, variable );

	return p;
}

__MODULEEND__



__MODULEBEGIN__( "mcopy2.c" )
/* memCopy2D
 2D copy operation
 1997. 5.
 1998. 3. Reviewed */

void __API memCopy2D
P_B(( dest, src, row, size, colrow, colcol, variable ))
P_1  void __F* dest
P_N  void __F* src
P_N  long row
P_N  unsigned size
P_N  const void __F* colrow
P_N  const void __F* colcol
P_N  unsigned variable
P_E
{
	register void __F* __H* destp = (void __F* __H*)dest;
	register void __F* __H* srcp = (void __F* __H*)src;
	while( row-- ) {
		memCopy( *destp++, *srcp++,
		 mem_iterate_2d(size,&colrow,&colcol,variable) );
	}
}
__MODULEEND__



__MODULEBEGIN__( "mset2.c" )
/* memSet2D */
/*
 2D memory initializer
 1997. 7.
 1998. 3. Reviewed
*/

void __API memSet2D
P_B(( dest, val, row, size, colrow, colcol, variable ))
P_1  void __F* dest
P_N  _unsigned1 val
P_N  long row
P_N  unsigned size
P_N  const void __F* colrow
P_N  const void __F* colcol
P_N  unsigned variable
P_E
{
	register void __F* __H* destp = (void __F* __H*)dest;
	while( row-- ) {
		memSet( *destp++, val,
		 mem_iterate_2d(size,&colrow,&colcol,variable ) );
	}
}
__MODULEEND__



__MODULEBEGIN__( "msizer.c" )
/* memSizeRect */

long __API memSizeRect
P_B(( row, colsize ))
P_1  long row
P_N  long colsize
P_E
{
	return memSize2D( row, 1, &colsize, NULL, sizeof(long) );
}

__MODULEEND__



__MODULEBEGIN__( "minitr.c" )
/* memInitRect
 'column' can be allocated independent to 'p'. This can be useful to
 initialize reference table for video memory */

void __API memInitRect
P_B(( p, column, row, colsize ))
P_1  void __F* p
P_N  void __F* column
P_N  long row
P_N  long colsize
P_E
{
	memInit2D( p, column, row, 1, &colsize, NULL, sizeof(long) );
}

__MODULEEND__



__MODULEBEGIN__( "mallocr.c" )
/* memAllocRect */

void __F* __API memAllocRect
P_B(( row, colsize ))
P_1  long row
P_N  long colsize
P_E
{
	return memAlloc2D( row, 1, &colsize, NULL, sizeof(long) );
}

__MODULEEND__



/*
 *
 *  Free column 2D allocation
 *
 */

__MODULEBEGIN__( "malloc2e.c" )



/* memAlloc2DExt */
/*
 Free column 2D allocator.
 1997. 5.
 1998. 3. Reviewed

 Allocates LxMxN matrix and MxN can be variable.
 This function allocates L size row vector and row iteration allocates
 MxN matrices which were sized by colrow and colcol arrays. MxN sub-
 arrays are not 2 dimensional but proper sized linear array is allocated.

 * For example,
 ID colrow = 1, colcol = 5;
 memAlloc2DExt( 5, sizeof(REAL), &colrow, &colcol, 0, 0 );
 will creates 5x1x5 real matrix.
*/

void __F* __API memAlloc2DExt
P_B(( row, size, colrow, colcol, variable ))
P_1  long row
P_N  unsigned size
P_N  const void __F* colrow
P_N  const void __F* colcol
P_N  unsigned variable
P_E
{
	/* allocate row */
	void __F* __H* pp = (void __F* __H*) memAlloc( (long)row*sizeof(void __F*) );

	if( pp ) {
		long irow;

		/* allocate columns */
		for( irow=0; irow<row; irow++ ) {
			if( ( pp[irow]=memAlloc
				( mem_iterate_2d(size,&colrow,&colcol,variable) ) )==NULL ) {
				memFree2DExt( pp, irow );
				return NULL;
			}
		}
	}

	return pp;
}



/* memFree2DExt */
/*
 2D memory deallocator
 1997. 5.
*/

void __API memFree2DExt
P_B(( mb, row ))
P_1  void __F* mb
P_N  long row
P_E
{
	if( mb ) {
		long irow; /* Row iterator */

		for( irow=0; irow<row; irow++ )
			memFree( ((void __F* __H*)mb)[irow] );

		memFree( mb );
	}
}


__MODULEEND__




/*
 *
 *  Search
 *
 */



__MODULEBEGIN__( "mlsrch.c" )
/* memLinearSearch

 Added on 2000.02.16 */

void __F* __API
memLinearSearch
P_B(( p, n, size, key, compare, index ))
P_1  const void __F* p
P_N  unsigned n
P_N  unsigned size
P_N  const void __F* key
P_N  int (__API* compare) P_A(( const void __F* p, const void __F* key ))
P_N  unsigned __F* index
P_E
{
	unsigned i;
	register _integer1 __F* current = (_integer1 __F*)p;

	for( i=0; i<n; i++ ) {
		if( compare(current,key)==0 ) {
			if( index ) *index = i;
			return current;
		}
		current += size;
	}

	return NULL;
}
__MODULEEND__





__MODULEBEGIN__( "mbsrch.c" )
/* memBinarySearch

	Prototpye modified on 1999.04.21 to support more flexible
 comparison procedure */



void __F* __API
memBinarySearch
P_B(( p, n, size, key, compare, index ))
P_1  const void __F* p
P_N  unsigned n
P_N  unsigned size
P_N  const void __F* key
P_N  int (__API* compare) P_A(( const void __F* p, const void __F* key ))
P_N  unsigned __F* index
P_E
{
	void __F* result = NULL;
	MEM_BEGIN_BSEARCH( n )
		MEM_ENTER_BSEARCH_LOOP( n )
			void __F* current = MEM_GET_CURRENT_BSEARCH_ELEM( p, size );
			register int cl = compare( current, key );
			if( cl==0 ) {
				result = current;
				break;
			}
			else if( cl<0 ) {
				MEM_BSEARCH_LESS
			}
			else {
				MEM_BSEARCH_GREATER(n)
			}
		MEM_LEAVE_BSEARCH_LOOP
		if( index ) *index = m;
	MEM_END_BSEARCH

	return result;
#if 0
	if( n>0 ) {
		void __F* current;
		unsigned l=0, m=0;
		n--;
		while( l<=n ) {
			register int cl;
			m = (l+n)>>1; /* middle position of l and n */
			current = (char __F*)p + m*size; /* get current item */
			cl = compare( current, key );

			/* check result and determine to continue iteration or not */
			if( cl<0 ) {
				l = m+1;
			}
			else if( cl>0 ) {
				if( m<=0 ) break;
				n = m-1;
			}
			else {
				result = current;
				break;
			}
		} /* end while( l<=n ) */

		if( index ) *index = m;
	} /* end if( n>0 ) */
#endif

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "mcntid.c" )
/* memCountValue */



unsigned __API memCountIdentity
P_B(( p, n, width, value ))
P_1  const void __F* p
P_N  unsigned n
P_N  unsigned width
P_N  long value
P_E
{
	register _integer1 __F* current = (_integer1 __F*)p;
	unsigned count=0;

	while( n-- ) {
		if( value==memRestoreInt(current,width) )
			count++;

		current += width;
	}

	return count;
}
__MODULEEND__



__MODULEBEGIN__( "mcntide.c" )
/* memCountIdentity */

unsigned __API memCountIdentityExt
P_B(( key, p, n, width, compare ))
P_1  const void __F* key
P_N  const void __F* p
P_N  unsigned n
P_N  unsigned width
P_N  int (__API* compare) P_A(( const void __F* a, const void __F* b ))
P_E
{
	register _integer1 __F* current = (_integer1 __F*)p;
	unsigned count=0;

	while( n-- ) {
		if( compare(key,current)==0 )
			count++;

		current += width;
	}

	return count;
}
__MODULEEND__



/*                      *\

	 Integer Size Control

\*                      */



__MODULEBEGIN__( "mreducei.c" )
/* memReduceInt

 Added to support ioWriteInt.
 Routines in ioWriteString is replaced to this method.
 2000.06.05 */

void __API memReduceInt
P_B(( value, size ))
P_1  long __F* value
P_N  unsigned size
P_E
{
	switch( size ) {
		case 1:
		*(_unsigned1 __F*)value = (_unsigned1)*value;
		break;

		case 2:
		*(_unsigned2 __F*)value = (_unsigned2)*value;

	/* for lp64 model */
# if __ARCH__==64

		case 0:
		case 4:
		*(_unsigned4 __F*)value = (_unsigned4)*value;

# elif __ARCH__>64

 /*  #error the current architecture was not considered */

# endif /* __ARCH__==64 */

	}
}
__MODULEEND__



__MODULEBEGIN__( "mrestori.c" )
/* memRestoreInt

 Added to support ioReadInt.
 Routines in ioReadString is replaced to this methid.
 2000.06.05 */

long __API memRestoreInt
P_B(( value, size ))
P_1  void __F* value
P_N  unsigned size
P_E
{

	switch( size ) {
		case 1:
		return *(_unsigned1 __F*)value;

		case 2:
		return *(_unsigned2 __F*)value;

# if __ARCH__==64 /* for lp64 model */

		case 0:
		case 4:
		return *(_unsigned4 __F*)value;

# elif __ARCH__>64

		/* #error the current architecture was not considered */

# endif /* __ARCH__==64 */
	}

	return *(long __F*)value;
}
__MODULEEND__



/* end of mem.c */
